//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        hgprot.h
//
// Purpose:     class HyperGProt (Hyper-G protocol handling class) 
//
// Created:     05 Dez 94    Thomas Dietinger
//
// Modified:    04 Apr 95    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>

#ifndef HG_HYPERG_HGPROT_H
#define HG_HYPERG_HGPROT_H

#include <apps/baseprot.h>
#include <fulltext/common/doclist.h>
#include <winsock.h>

class ServerConnection;
class CSocket;

/////////////////////////////////////////////////////////////////
// class HGListFD: Hyper-G file descriptor for retrieving lists
class HGListFD : public BaseFD
{
public:
    HGListFD()  {}

    ObjIdArray fdData_;
    RString helpStr_;       // helper string may be used or not
};

/////////////////////////////////////////////////////////////////
// class HGFTListFD: Hyper-G file descriptor for retrieving lists
//                 for fulltext queries
class HGFTListFD : public BaseFD
{
public:
    HGFTListFD()  {}

    FtDocList fdData_;
    RString helpStr_;       // helper string may be used or not
};

/////////////////////////////////////////////////////////////////
// class HGDataFD: Hyper-G file descriptor for getting/putting data
class HGDataFD : public BaseFD
{
public:
    HGDataFD()  { fdData_ = 0; }
    virtual ~HGDataFD();

    CSocket* fdData_;     // file descriptor representation
};

/////////////////////////////////////////////////////////////////
class HyperGProt : public BaseProt
{
public:
    HyperGProt(RString connectStr, boolean connect = true);  // usual way
    virtual ~HyperGProt();

    // get object information by destination ID/URL
    virtual Object getObjectInfo(RString dest);

    // check to see whether we can use this connection type with this protocol
    virtual boolean     useConnection(RString connectStr);

    virtual boolean     identify(const RString& name, 
                                 const RString& passwd = "", 
                                 boolean encrypt = false);
    virtual boolean     changePassword(const RString& oldPassword, 
                                       const RString& newPassword);

    virtual BaseFD*     open(ObjectPtr objPtr,           // pointer to object of interest
                             OpenMode mode,              // opening mode
                             ObjectList* objList = 0,    // pointer to additional list (e.g. list of activated collections)
                             RString helpStr = "");      // additional string (e.g. query string)
    virtual boolean     close(BaseFD* fd);

    virtual long        getData(BaseFD* fd, 
                                char* dataBuf, 
                                long buflen);
    virtual long        putData(BaseFD* fd, 
                                char* dataBuf, 
                                long buflen);

    virtual ObjectList* getList(BaseFD* fd, 
                                long count,
                                ObjectList* oldList = 0);
//    virtual boolean     putList(BaseFD* fd, 
//                                   ObjectList* newObjs);

    virtual boolean     insert(ObjectPtr object, 
                               ObjectPtr toColl);

    virtual boolean     erase(ObjectPtr);
    virtual boolean     moveTo(ObjectPtr, 
                               ObjectPtr fromColl, 
                               ObjectPtr toColl);
    virtual boolean     copyTo(ObjectPtr,
                               ObjectPtr toColl);
    virtual boolean     unlinkFrom(ObjectPtr, 
                                   ObjectPtr fromColl);

    virtual HGObject*   lock(ObjectPtr);
    virtual boolean     unlock(ObjectPtr);

    virtual boolean     modifyObjAttribute(ObjectPtr objPtr,    // referred object
                                           RString oldAttrLine, // old attribute
                                           RString newAttrLine  // new attribute
                                           );

    HgStub* serverPtr() const           { return (HgStub*)hgConnPtr_; }

protected:

    RString enCrypt(const RString& clear);
    boolean validatePassword(const RString& clear, 
                             const RString& encr);

    void receiveObjects(HGListFD* fd,
                        ObjectList* result,
                        boolean parents = false,
                        long count = -1) const;

    void receiveObjects(HGFTListFD* fd,
                        ObjectList* result,
                        boolean parents = false,
                        long count = -1) const;

    // protocol attributes (may be used or not)
    RString subType_;       // protocol subtype, if a Hyper-G server is used as a proxy server
    ServerConnection* hgConnPtr_;
    boolean binary_;
    boolean distributed_;
    boolean compressed_;

    boolean directDC_;

    Object userObj_;
};


#endif
