//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        protptr.h
//
// Purpose:     class ProtPtr (Pointer to a class (instance) derived from 
//              the basic protocol handling class) 
//
// Created:     10 Oct 94    Thomas Dietinger
//
// Modified:    02 Dez 94    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>

#ifndef HG_HYPERG_PROTPTR_H
#define HG_HYPERG_PROTPTR_H

#include <utils/list.h>

class BaseProt;

/////////////////////////////////////////////////////////////////
// class ProtPtr: Pointer to BaseProt (with resource management)

class ProtPtr 
{
public:
    ProtPtr(BaseProt* prot = 0);
    ProtPtr(const ProtPtr&);
    ~ProtPtr();

    ProtPtr& operator = (const ProtPtr&);
    ProtPtr& operator = (BaseProt*);
    BaseProt* protocol() const             { return prot_; }

private:
    friend inline int operator ==(const ProtPtr&, const ProtPtr&);
    friend inline int operator !=(const ProtPtr&, const ProtPtr&);
    friend inline int operator ==(const ProtPtr&, BaseProt*);
    friend inline int operator !=(const ProtPtr&, BaseProt*);
    BaseProt* prot_;
};

inline int operator ==(const ProtPtr& a, const ProtPtr& b) {
    return a.protocol() == b.protocol();
}

inline int operator !=(const ProtPtr& a, const ProtPtr& b) {
    return a.protocol() != b.protocol();
}

inline int operator ==(const ProtPtr& a, BaseProt* prot) {
    return a.protocol() == prot;
}

inline int operator !=(const ProtPtr& a, BaseProt* prot) {
    return a.protocol() != prot;
}


/////////////////////////////////////////////////////////////
// class ProtocolNode
// ----------------------------------------------------------
//
class ProtocolNode : public DLListNode 
{
public:
    ProtocolNode (const ProtPtr&);

    ProtPtr protPtr() const     { return protPtr_; }
    BaseProt* protocol() const  { return protPtr_.protocol(); }

private:
    ProtPtr protPtr_;
};

inline ProtocolNode::ProtocolNode(const ProtPtr& protPtr) : DLListNode() 
{
    protPtr_ = protPtr;
}

/////////////////////////////////////////////////////////////
// class ProtocolList
// ----------------------------------------------------------
//

DLListdeclare(ProtocolList, ProtocolNode)



#endif
