// This may look like C code, but it is really -*- C++ -*-

// **********************************************************
// File cursesviewer.C
// ----------------------------------------------------------
// Created: F. Kappe
// Addapted for PC: T. Dietinger
// **********************************************************


#include "stdafx.h"
#include <hgapp.h>
#include <pcutil/pcstuff.h>
#include <direct.h>
#include <viewer/text/viewdoc.h>
#include <viewer/text/viewvw.h>

//#ifdef LOCAL_DB_20
//#include <hyperg/l_server/editstub.h>
//#else
//#include <hyperg/server/editstub.h>
//#endif

#include "hglinks.h"
#include "baseprot.h"
#include "hgprot.h"
#include "cursesviewer.h"
#include "viewerlist.h"

#include "conf.h"
#include "treaps.h"
#include "utils.h"
#include "language.h"
#include <pcutil/debug.h>

#if defined(_MSC_VER) & defined(_DEBUG)
#define new DEBUG_NEW
#endif

extern StringKeyRefTree Seen;     // Documents we have already seen

/////////////////////////////////////////////////////////////////////////////

boolean ReadHTMLtoHTF(WWWDocument* current,
                      RString& title,
                      RString& text,
                      LinkList* linkList);

/////////////////////////////////////////////////////////////////////////////
// class TextViewer
// ~~~~~~~~~~~~~~~~
//
// Used to display TextDocuments or derived Documents. 
//

TextViewer::TextViewer(boolean html) : Viewer() 
{
//    _text = nil;
//    _dest = nil;
    _line = 0;
    _offset = -1;
    _html_mode = html;
    _numannotations = 0;
    tmpTextName_ = "";
    langcode_ = -1;
    if (html)
        sgmlCnf_ = "HTML";
    else
        sgmlCnf_ = "HTF";

    dataBase_ = 0;
    document_ = 0;

    lastChar_ = '\0';
}

TextViewer::~TextViewer() 
{
//    delete _dest;
    unlink(tmpTextName_);
}


boolean TextViewer::Load(Document* d, Anchor* a, boolean dohyphen) 
{
    dataBase_ = 0;
    document_ = d;

    _text = ((TextDocument*)document_)->text();

    if (d->DocType() == "Remote")
    {
        _html_mode = true;
        sgmlCnf_ = "HTML";
    }
    else
    {
        // check if there are annotations:
        _numannotations = 0;
        ObjectList* references = d->references();
        if (references)
            for (ObjectNode* n = references->getFirst(); n ; n = references->getNext(n))
                if (((Anchor*)n->ObjPtr().Obj())->LinkType() == "Annotation")
                    _numannotations++;
    }

    if (a) 
    {
        // compute position to jump to:
        ObjectID off((const char*)(a->Position()));
        _offset = off.id();
    }
    else
        _offset = -1;

    _title = d->Title();

    return true;
}
/*
boolean TextViewer::Load(HgStub* db, Document* d, Anchor* a, boolean dohyphen) 
{
    dataBase_ = db;
    document_ = d;

    _text = ((TextDocument*)document_)->text();

    if (d->DocType() == "Remote")
    {
        _html_mode = true;
        sgmlCnf_ = "HTML";
    }
    else
    {
        // check if there are annotations:
        _numannotations = 0;
        ObjectList* linksFrom = d->linksFrom();
        if (linksFrom)
            for (ObjectNode* n = linksFrom->getFirst(); n ; n = linksFrom->getNext(n))
                if (((Anchor*)n->ObjPtr().Obj())->LinkType() == "Annotation")
                    _numannotations++;
    }

    if (a) 
    {
        // compute position to jump to:
        ObjectID off((const char*)(a->Position()));
        _offset = off.id();
    }
    else
        _offset = -1;

    _title = d->Title();

    return true;
}
*/
//
// Browse and evaluate anchor klicks
//
char TextViewer::Browse(ObjectPtr& next) 
{
    tmpTextName_ = RString(tempnam(getenv("TEMP"), "HgTxt-")) + RString(".htf");

    // Important !!!!!!
    // => reset old hgObj_ pointer ! 
    // Now this time the destructor of hgObj_.Obj() will be called !
    // (Else it would be called in 'initDocument' and might erase the file 'tmpTextName_' !!!!)
    theApp.getTextDoc()->hgObj_ = ObjectPtr(0);

    FILE *f = fopen(tmpTextName_, "w");
    if (f)
    {
        // start download of inline images
        RStringList inlNameList;

        RString dest;
        if (document_)
        for (TextLinkNode* n = (TextLinkNode*)document_->linkList()->getFirst(); 
            n; 
            n = (TextLinkNode*)document_->linkList()->getNext(n)) 
        {
            if (n->getLinkType() == "inline")
            {
                dest = n->getDestId();

                // inline link, execute it immediately !!
                CommandStruct* newCmd = nil;
                HGObject* hgo = nil;

                int pos = dest.index(':');      // test for protocol specifier (eg. "http:", "gopher:", ..)
                if (pos != -1)
                {
                    if (inlNameList.find(dest)) // inline already in list, so ignore it !
                        continue;

                    hgo = makeInlineObjectById(dest, n->getSrcId());
                    if (!hgo)
                        continue;

                    RString tempName = getTempName((Document*)hgo);
                    boolean forceLoad = false;
                    if ( tempName.length() && getCompleteFileName(tempName, forceLoad))
                    {
                        unlink(tempName);       // delete file from cache, to be sure that it's the right one !
                    }

                    if (theApp.m_InlineOnDemand_)
                    {
                        delete hgo;
                        continue;
                    }

                    inlNameList.addTail(new RStringNode(dest));
                }
                else
                {
                    if (inlNameList.find(dest)) // inline already in list, so ignore it !
                        continue;

                    hgo = makeInlineObjectById(dest);

                    if (hgo)
                    {
                        if (!theApp.m_pTextInfo_ || !theApp.m_pTextInfo_->m_pView_ || 
                            !theApp.getTextView()->pictureList_.existID(n->getDestId()))
                        {
                            RString tempName = getTempName((Document*)hgo);
                            boolean forceLoad = false;
                            if ( tempName.length() && 
                                !getCompleteFileName(tempName, forceLoad) &&
                                theApp.m_InlineOnDemand_)
                            {
                                // inline object is NOT already in local cache !
                                // and we must NOT download it now (only on demand) !
                                delete hgo;
                                continue;
                            }
                        }

                        inlNameList.addTail(new RStringNode(dest));
                    }
                    else
                        continue;   // couldn't create inline object !
                }

                if (hgo)
                {
                    ObjectPtr objPtr(hgo);
                    newCmd = new CommandStruct(COMM_INLINE, objPtr);
                    theApp.cmdList_.addTail(newCmd);
//                    TRACE("TextViewer::Browse: schedule inline to download: %s\n", (const char*)dest);
                }
            }
        }
        theApp.getTextDoc()->setSgmlConfig(sgmlCnf_, langcode_);

//        if (document_)
//            theApp.getTextDoc()->setLinkList(*(document_->linkList()));

        Output(f);
        fclose(f);

        theApp.getTextDoc()->initDocument(tmpTextName_, _text, _offset, document_);

        ((CMDIChildWnd*)theApp.getTextInfo()->m_pFrameWnd_)->MDIActivate();
        theApp.getTextDoc()->SetTitle(_title);
    }
    else
        ErrorMsg("Can't write temporary HTF-text file !");

    return DO_GUICMD;

}

//
// Browse and evaluate anchor klicks
//
/*
char TextViewer::Browse(HgStub* db, ObjectPtr& next) 
{
    tmpTextName_ = RString(tempnam(getenv("TEMP"), "HgTxt-")) + RString(".htf");

    // Important !!!!!!
    // => reset old hgObj_ pointer ! 
    // Now this time the destructor of hgObj_.Obj() will be called !
    // (Else it would be called in 'initDocument' and might erase the file 'tmpTextName_' !!!!)
    theApp.getTextDoc()->hgObj_ = ObjectPtr(0);

    FILE *f = fopen(tmpTextName_, "w");
    if (f)
    {
        // start download of inline images
        RStringList inlNameList;

        RString dest;
        if (document_)
        for (TextLinkNode* n = (TextLinkNode*)document_->linkList()->getFirst(); 
            n; 
            n = (TextLinkNode*)document_->linkList()->getNext(n)) 
        {
            if (n->getLinkType() == "inline")
            {
                dest = n->getDestId();

                // inline link, execute it immediately !!
                CommandStruct* newCmd = nil;
                HGObject* hgo = nil;

                int pos = dest.index(':');      // test for protocol specifier (eg. "http:", "gopher:", ..)
                if (pos != -1)
                {
                    if (inlNameList.find(dest)) // inline already in list, so ignore it !
                        continue;

                    hgo = makeInlineObjectById(dest, n->getSrcId());

                    RString tempName = getTempName((Document*)hgo);
                    boolean forceLoad = false;
                    if ( tempName.length() && getCompleteFileName(tempName, forceLoad))
                    {
                        unlink(tempName);       // delete file from cache, to be sure that it's the right one !
                    }

                    if (theApp.m_InlineOnDemand_)
                    {
                        delete hgo;
                        continue;
                    }

                    inlNameList.addTail(new RStringNode(dest));
                }
                else
                {
                    if (inlNameList.find(dest)) // inline already in list, so ignore it !
                        continue;

                    hgo = makeInlineObjectById(dest);

                    if (hgo)
                    {
                        if (!theApp.m_pTextInfo_ || !theApp.m_pTextInfo_->m_pView_ || 
                            !theApp.getTextView()->pictureList_.existID(n->getDestId()))
                        {
                            RString tempName = getTempName((Document*)hgo);
                            boolean forceLoad = false;
                            if ( tempName.length() && 
                                !getCompleteFileName(tempName, forceLoad) &&
                                theApp.m_InlineOnDemand_)
                            {
                                // inline object is NOT already in local cache !
                                // and we must NOT download it now (only on demand) !
                                delete hgo;
                                continue;
                            }
                        }

                        inlNameList.addTail(new RStringNode(dest));
                    }
                    else
                        continue;   // couldn't create inline object !
                }

                if (hgo)
                {
                    ObjectPtr objPtr(hgo);
                    newCmd = new CommandStruct(COMM_INLINE, objPtr);
                    theApp.cmdList_.addTail(newCmd);
//                    TRACE("TextViewer::Browse: schedule inline to download: %s\n", (const char*)dest);
                }
            }
        }

        theApp.getTextDoc()->setSgmlConfig(sgmlCnf_, langcode_);

//        if (document_)
//            theApp.getTextDoc()->setLinkList(*(document_->linkList()));

        Output(db, f);
        fclose(f);

        theApp.getTextDoc()->initDocument(tmpTextName_, _text, _offset, dataBase_, document_);

        ((CMDIChildWnd*)theApp.getTextInfo()->m_pFrameWnd_)->MDIActivate();
        theApp.getTextDoc()->SetTitle(_title);
    }
    else
        ErrorMsg("Can't write temporary HTF-text file !");

    return DO_GUICMD;

}
*/

//
// Output to file descriptor.
//
void TextViewer::Output(FILE* f) 
{
    // Initialize state of NextChar()
//    _nextstart = _start.getFirst(); // next start anchor

    endAnchList_ = new LinkList();

    if (document_)
    {
        // sort links after end anchors (not start anchors)
        for (TextLinkNode* n = (TextLinkNode*)document_->linkList()->getFirst(); 
            n; 
            n = (TextLinkNode*)document_->linkList()->getNext(n)) 
        {
            for (BaseLinkNode* t = endAnchList_->getFirst(); 
                 t && ((TextLinkData*)t->getLinkData())->endOffset_ < ((TextLinkData*)n->getLinkData())->endOffset_;
                 t = endAnchList_->getNext(t)) ;
            if (t) 
                endAnchList_->insertBefore(new TextLinkNode(n), t);
            else
                endAnchList_->addTail(new TextLinkNode(n));
        }

        nextStart_ = (TextLinkNode*)document_->linkList()->getFirst();
        _startoffset = nextStart_ ? ((TextLinkData*)nextStart_->getLinkData())->startOffset_ : -1; 
        nextEnd_ = (TextLinkNode*)endAnchList_->getFirst();
        _endoffset = nextEnd_ ? ((TextLinkData*)nextEnd_->getLinkData())->endOffset_ : -1; 
    }
    else
    {
        _startoffset = -1;
        _endoffset = -1;
    }
//    _nextend = _end.getFirst();   // next end anchor    
//    _endoffset = _nextend ? _nextend->offset().id() : -1;
    _textptr = 
    _insertptr = 
    _insertlen = 0;
    _textlen = _text.length();
    _anchor = 0;
    lastChar_ = '\0';
    verbatim_ = false;

    char c;
    while ((c = NextChar()) != -1) 
        putc(c, f);

    delete endAnchList_;
    endAnchList_ = 0;
}

//
// Output to file descriptor.
//
/*
void TextViewer::Output(HgStub* db, FILE* f) 
{
    // Initialize state of NextChar()
//    _nextstart = _start.getFirst(); // next start anchor

    endAnchList_ = new LinkList();

    if (document_)
    {
        // sort links after end anchors (not start anchors)
        for (TextLinkNode* n = (TextLinkNode*)document_->linkList()->getFirst(); 
            n; 
            n = (TextLinkNode*)document_->linkList()->getNext(n)) 
        {
            for (BaseLinkNode* t = endAnchList_->getFirst(); 
                 t && ((TextLinkData*)t->getLinkData())->endOffset_ < ((TextLinkData*)n->getLinkData())->endOffset_;
                 t = endAnchList_->getNext(t)) ;
            if (t) 
                endAnchList_->insertBefore(new TextLinkNode(n), t);
            else
                endAnchList_->addTail(new TextLinkNode(n));
        }

        nextStart_ = (TextLinkNode*)document_->linkList()->getFirst();
        _startoffset = nextStart_ ? ((TextLinkData*)nextStart_->getLinkData())->startOffset_ : -1; 
        nextEnd_ = (TextLinkNode*)endAnchList_->getFirst();
        _endoffset = nextEnd_ ? ((TextLinkData*)nextEnd_->getLinkData())->endOffset_ : -1; 
    }
    else
    {
        _startoffset = -1;
        _endoffset = -1;
    }
//    _nextend = _end.getFirst();   // next end anchor    
//    _endoffset = _nextend ? _nextend->offset().id() : -1;
    _textptr = 
    _insertptr = 
    _insertlen = 0;
    _textlen = _text.length();
    _anchor = 0;
    lastChar_ = '\0';
    verbatim_ = false;

    char c;
    while ((c = NextChar(db)) != -1) 
        putc(c, f);

    delete endAnchList_;
    endAnchList_ = 0;
}
*/

//
// This Routine returns the next character to the parser; it has to insert
// the anchors on the fly and also determines the line to jump to if a 
// destination anchor was given.
//
/*
int TextViewer::NextChar() 
{
    static long oldStartOffset;
    static boolean tagOpen;
    static RString lastTag;

    if (_insertptr < _insertlen)               // we have to insert a string
        return _insert[_insertptr++];
    else 
    {                                     // interpret standard text
        if (_textptr == _startoffset && !_html_mode) 
        {          // insert start-anchor tag
            RString dest = nextStart_->getDestId();
            RString id(nextStart_->getSrcId());
            if (_html_mode)                             // we produce HTML (WWW)
            {
                if (nextStart_->getLinkType() == "inline")
                    sprintf(_insert, "<IMG SRC=\"A%s\"><SYNC OFFSET=\"%d\">", (const char*)id, _textptr);
                else
                    sprintf(_insert, "<A HREF=\"A%s\"><SYNC OFFSET=\"%d\">", (const char*)id, _textptr);
            }
            else                                        // we produce HTF
            {
                RString ltype;
                if (nextStart_->getLinkType() == "inline")
                {
                    sprintf(_insert, "<SYNC OFFSET=\"%d\"><INLINE ID=\"%s\" DHINT=\"%s\">",
                            _textptr,
                            (const char*)id,
                            (const char*)dest.GetSubstrIndex(0,49));
                }
                else
                    sprintf(_insert, "<AS ID=\"%s\" DHINT=\"%s\"><SYNC OFFSET=\"%d\">",
                            (const char*)id,
                            (const char*)dest.GetSubstrIndex(0,49), 
                            _textptr);

                ASSERT(id.length());
                ASSERT(dest.GetSubstrIndex(0,49).length());
            } // HTF

            _insertlen = strlen(_insert);
            ASSERT(_insertlen);
            _insertptr = 1;
            nextStart_ = (TextLinkNode*)document_->linkList()->getNext(nextStart_);
            oldStartOffset = _startoffset;
            _startoffset = nextStart_ ? ((TextLinkData*)nextStart_->getLinkData())->startOffset_ : -1; 
            return *_insert;
        }
        else 
        if (_textptr == _endoffset && !_html_mode) 
        {       // insert end-anchor tag
            if (_html_mode)                        // we produce HTML (WWW)
            {
                char fillchar[10];
                if (oldStartOffset == _endoffset)
                    strcpy(fillchar, "[]");      // workaround for HTML, if there is a link 
                else                            // without a link text !:add a fillchar
                    strcpy(fillchar, "");

                sprintf(_insert, "%s", fillchar);

            }
            else                                   // we produce HTF      
            {
                RString dest = nextEnd_->getDestId();
                RString id(nextEnd_->getSrcId());
                sprintf(_insert, "<AE ID=\"%s\"><SYNC OFFSET=\"%d\">", (const char*)id, _textptr);
                ASSERT(id.length());
            }

            ++_anchor;
            _insertlen = strlen(_insert);
            ASSERT(_insertlen);
            _insertptr = 1;
            nextEnd_ = (TextLinkNode*)endAnchList_->getNext(nextEnd_);
            _endoffset = nextEnd_ ? ((TextLinkData*)nextEnd_->getLinkData())->endOffset_ : -1; 
            return *_insert;
        }
        else 
        {                                   // insert standard text
            if (_textptr < _textlen) 
            {
                char ch = (_text)[(int)_textptr++];

                if (ch == '<')
                {
                    tagOpen = true;
                    lastTag = "";
                }
                else
                if (ch == '>')
                {
                    tagOpen = false;
//                    TRACE("SGML tag found: <%s>\n", lastTag.string());
                    if (lastTag.toupper() == "VERBATIM")
                        verbatim_ = true;
                    else
                    if (lastTag.toupper() == "/VERBATIM")
                        verbatim_ = false;
                }
                else
                if (tagOpen)
                {
                    char chr[2];
                    chr[0] = ch;
                    chr[1] = 0;
                    lastTag += chr;
                }

                if (!verbatim_ && (ch == '>' || ch == ';'))
                {
                    sprintf(_insert, "%c<SYNC OFFSET=\"%d\">", ch, _textptr);
                    _insertlen = strlen(_insert);
                    ASSERT(_insertlen);
                    _insertptr = 1;
                    lastChar_ = ch;
                    return *_insert;
                }
                else
                if (!verbatim_ && lastChar_ == '\n' && ch != '\n' && !tagOpen)
                {
                    sprintf(_insert, "<SYNC OFFSET=\"%d\">%c", _textptr-1, ch);
                    _insertlen = strlen(_insert);
                    ASSERT(_insertlen);
                    _insertptr = 1;
                    lastChar_ = ch;
                    return *_insert;
                }
                else
                {
                    ASSERT(ch);
                    lastChar_ = ch;
                    return ch;              // return the original char
                }
            }
            else
                return -1;                        // EOF
        }
    }
}
*/
//
// This Routine returns the next character to the parser; it has to insert
// the anchors on the fly and also determines the line to jump to if a 
// destination anchor was given.
//
int TextViewer::NextChar() 
{
    static long oldStartOffset;
    static boolean tagOpen;
    static RString lastTag;

    if (_insertptr < _insertlen)               // we have to insert a string
        return _insert[_insertptr++];
    else 
    {                                     // interpret standard text
        if (_textptr == _startoffset && !_html_mode) 
        {          // insert start-anchor tag
            RString dest = nextStart_->getDestId();
            RString id(nextStart_->getSrcId());
            if (_html_mode)                             // we produce HTML (WWW)
            {
                if (nextStart_->getLinkType() == "inline")
                    sprintf(_insert, "<IMG SRC=\"A%s\"><SYNC OFFSET=\"%d\">", (const char*)id, _textptr);
                else
                    sprintf(_insert, "<A HREF=\"A%s\"><SYNC OFFSET=\"%d\">", (const char*)id, _textptr);
            }
            else                                        // we produce HTF
            {
                RString ltype;
                if (nextStart_->getLinkType() == "inline")
                {
                    sprintf(_insert, "<SYNC OFFSET=\"%d\"><INLINE ID=\"%s\" DHINT=\"%s\">",
                            _textptr,
                            (const char*)id,
                            (const char*)dest.GetSubstrIndex(0,49));
                }
                else
                    sprintf(_insert, "<AS ID=\"%s\" DHINT=\"%s\"><SYNC OFFSET=\"%d\">",
                            (const char*)id,
                            (const char*)dest.GetSubstrIndex(0,49), 
                            _textptr);

                ASSERT(id.length());
                ASSERT(dest.GetSubstrIndex(0,49).length());
            } // HTF

            _insertlen = strlen(_insert);
            ASSERT(_insertlen);
            _insertptr = 1;
            nextStart_ = (TextLinkNode*)document_->linkList()->getNext(nextStart_);
            oldStartOffset = _startoffset;
            _startoffset = nextStart_ ? ((TextLinkData*)nextStart_->getLinkData())->startOffset_ : -1; 
            return *_insert;
        }
        else 
        if (_textptr == _endoffset && !_html_mode) 
        {       // insert end-anchor tag
            if (_html_mode)                        // we produce HTML (WWW)
            {
                char fillchar[10];
                if (oldStartOffset == _endoffset)
                    strcpy(fillchar, "[]");      // workaround for HTML, if there is a link 
                else                            // without a link text !:add a fillchar
                    strcpy(fillchar, "");

                sprintf(_insert, "%s", fillchar);

            }
            else                                   // we produce HTF      
            {
                RString dest = nextEnd_->getDestId();
                RString id(nextEnd_->getSrcId());
                sprintf(_insert, "<AE ID=\"%s\"><SYNC OFFSET=\"%d\">", (const char*)id, _textptr);
                ASSERT(id.length());
            }

            ++_anchor;
            _insertlen = strlen(_insert);
            ASSERT(_insertlen);
            _insertptr = 1;
            nextEnd_ = (TextLinkNode*)endAnchList_->getNext(nextEnd_);
            _endoffset = nextEnd_ ? ((TextLinkData*)nextEnd_->getLinkData())->endOffset_ : -1; 
            return *_insert;
        }
        else 
        {                                   // insert standard text
            if (_textptr < _textlen) 
            {
                char ch = (_text)[(int)_textptr++];

                if (ch == '<')
                {
                    tagOpen = true;
                    lastTag = "";
                }
                else
                if (ch == '>')
                {
                    tagOpen = false;
//                    TRACE("SGML tag found: <%s>\n", lastTag.string());
                    if (lastTag.toupper() == "VERBATIM")
                        verbatim_ = true;
                    else
                    if (lastTag.toupper() == "/VERBATIM")
                        verbatim_ = false;
                }
                else
                if (tagOpen)
                {
                    char chr[2];
                    chr[0] = ch;
                    chr[1] = 0;
                    lastTag += chr;
                }

                if (!verbatim_ && (ch == '>' || ch == ';'))
                {
                    sprintf(_insert, "%c<SYNC OFFSET=\"%d\">", ch, _textptr);
                    _insertlen = strlen(_insert);
                    ASSERT(_insertlen);
                    _insertptr = 1;
                    lastChar_ = ch;
                    return *_insert;
                }
                else
                if (!verbatim_ && lastChar_ == '\n' && ch != '\n' && !tagOpen)
                {
                    sprintf(_insert, "<SYNC OFFSET=\"%d\">%c", _textptr-1, ch);
                    _insertlen = strlen(_insert);
                    ASSERT(_insertlen);
                    _insertptr = 1;
                    lastChar_ = ch;
                    return *_insert;
                }
                else
                {
                    ASSERT(ch);
                    lastChar_ = ch;
                    return ch;              // return the original char
                }
            }
            else
                return -1;                        // EOF
        }
    }
}


/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// class InfoViewer
// ~~~~~~~~~~~~~~~~
//
// Used to display the Object Record of a Hyper-G Object and
// other things as text objects
//

//
// Show statistics and who is online
//
char InfoViewer::ShowStatus() 
{
    ProtPtr pPtr;

    // look for Hyper-G protocol in protocol list
    for (ProtocolNode* protNode = HGObject::conOrProtList_.getFirst();
         protNode && protNode->protocol() && protNode->protocol()->getProtType() != "hyperg";
         protNode = HGObject::conOrProtList_.getNext(protNode));

    if (protNode)
        pPtr = protNode->protPtr();
    else
    {
        ErrorMsg(IDS_ONLYHG);
        return 0;
    }

    static char buf[1000];
    Object stat;
    ((HyperGProt*)pPtr.protocol())->serverPtr()->Command("stat",stat);
    RString ltime = stat.field("LocalTime=");
    if (ltime.length()) 
    {
        RString up = stat.field("UpSince=");

        //what a hack:
        int pos = stat.indexa("FObj=");
        ObjectID fetchall(stat.gSubstrIndex(pos     , pos +  9));
        ObjectID fetch60m(stat.gSubstrIndex(pos + 11, pos + 20));
        ObjectID fetch15m(stat.gSubstrIndex(pos + 22, pos + 31));
        ObjectID fetch01m(stat.gRight(pos +33));
        ObjectID upsecs(up.gSubstrIndex(18,up.length()));
    
        sprintf(buf, ServerStatStr, 
            (const char*)((HyperGProt*)pPtr.protocol())->serverPtr()->ServerString(), 
            (const char*)ltime,
            (const char*)up.GetSubstrIndex(0,16),
            upsecs.id()/86400,        // days
            upsecs.id()%86400 / 3600, // hours
            upsecs.id()%3600 / 60,    // minutes
            upsecs.id()%60,           // seconds
            fetch01m.id()/60.0,
            fetch15m.id()/900.0,
            fetch60m.id()/3600.0,
            fetchall.id()/(float)upsecs.id());
    }
    else
        strcpy(buf,"<TITLE>xxx<XMP>stat command failed\n\n");

    RString who;
//    ((HyperGProt*)pPtr.protocol())->serverPtr()->Command("who l",who);
    ((HyperGProt*)pPtr.protocol())->serverPtr()->Command("who",who);

    _text = RString(buf + who.GetSubstrIndex(19,who.length()) + "</XMP>");
//    delete _dest; _dest = new OidUStringTree();

    char t[100];
    sprintf(t, Copyrightstr, ((HyperGProt*)pPtr.protocol())->serverPtr()->version(), hgVersion);
    _title = t;
    ObjectPtr tmpPtr(0);
    return Browse(tmpPtr);
}

/////////////////////////////////////////////////////////////////////////////
// class WWWViewer
// ~~~~~~~~~~~~~~~
//
// Used to display the WordWideWeb Documents
//

boolean WWWViewer::Load(Document* d, Anchor*, boolean) 
{
//    delete _dest; _dest = new OidUStringTree();
//    _start.free();
//    _end.free();

//    return ReadHTMLtoHTF((WWWDocument*)d, _title, _text, _start, _end, *_dest);
//    dataBase_ = db;
    document_ = d;

    _text = ((TextDocument*)document_)->text();

    return ReadHTMLtoHTF((WWWDocument*)d, _title, _text, d->linkList());
}

//
// Used to (re-)display a text document
//

char WWWViewer::Browse(ObjectPtr& next) 
{
    // query ending of next->Title() for 
    // .tiff
    // .gif
    // .??? => then it is an image etc. !!!

    return TextViewer::Browse(next);
}

/////////////////////////////////////////////////////////////////////////////
// class GopherViewer
// ~~~~~~~~~~~~~~~~~~
//
// Used to display Gopher Text Documents
//

GopherViewer::~GopherViewer()
{
    if (viewerInfo_)
    {
        // check, whether document is still valid (then it must also be in the viewer list !)
        for (CInfoNode* iNode = theApp.openViews_.getFirst(); iNode; iNode = theApp.openViews_.getNext(iNode))
        {
            if (iNode->info_.m_pDocument_ == viewerInfo_->m_pDocument_)
                viewerInfo_->m_pDocument_->OnCloseDocument();   // delete viewer, if it is still open
        }
    }

    if (tempName_.length())
        unlink(tempName_);
}


boolean GopherViewer::Load(Document* d, Anchor*, boolean) 
{
//    delete _dest; _dest = new OidUStringTree();
//    _start.free();
//    _end.free();

    _title = d->Title();
    _gophertype = ((GopherDocument*)d)->GopherType();
    boolean ret = LoadGopher((GopherDocument*)d, _text);

    return ret;
}


char GopherViewer::Browse(ObjectPtr& next)
{
    return TextViewer::Browse(next);
}

char GopherViewer::ImageDisplay(ObjectPtr& objPtr) 
{
    if (_gophertype == "I" || _gophertype == "g" || _gophertype == ":")
    {
        tempName_ = RString(tempnam(getenv("TEMP"), "hggif"));
        tempName_ = tempName_.GetSubstrIndex(0, tempName_.length()-4);
        tempName_ += "gif";
       
        if (writeBinFile(_text, _text.length(), tempName_))
        {       
/*
            // set private profile path    
            char path[_MAX_PATH];
            getcwd(path, _MAX_PATH);     // get current working directory
            strcat(path, "\\");
            strcat(path, "docman.ini");  // append ini-filename
            char szT[_MAX_PATH];
        
            ::GetPrivateProfileString("Image", "gif", "viewer.exe", 
                              szT, sizeof(szT), path);
            
            WaitAppExec(RString(szT) + RString(" ") + tempName); // waits for command to execute
*/
            int showStyle = theApp.GetPrivateProfileInt("MainFrame", "MMViewerStyle", SW_SHOWMINIMIZED);
            viewerInfo_ = theApp.m_pImageTemplate_->OpenNewObject(tempName_, objPtr.Obj()->Title(), showStyle);
            if (viewerInfo_->m_pDocument_)
            {
                viewerInfo_->m_pDocument_->SetTitle(objPtr.Obj()->Title());
                theApp.openViews_.addTail(viewerInfo_, imageViewer);
            }
            theApp.tileWindows();
        }
        else
        {   
            tempName_ = "";
        }
    }
    else 
    if (_gophertype == "S" || _gophertype == "s" || _gophertype == "<")
        ErrorMsg(IDS_NOTIMPLEMENT);

    return DO_GUICMD;
}

/////////////////////////////////////////////////////////////////////////////
// class WaisViewer
// ~~~~~~~~~~~~~~~~
//
// Used to display Wais Text Documents
//

boolean WaisViewer::Load(Document* d, Anchor*, boolean) 
{
    _title = d->Title();
    _waistype = ((WaisDocument*)d)->WaisType();
//    return LoadWais(_text, (WaisDocument*)d) ;

    ErrorMsg(IDS_NOTIMPLEMENT);
    return false;
}


char WaisViewer::Browse(ObjectPtr& next)
{
    return TextViewer::Browse(next);
}

char WaisViewer::ImageDisplay(ObjectPtr&) 
{
    ErrorMsg(IDS_NOTIMPLEMENT);
    return DO_GUICMD;
}

