// This may look like C code, but it is really -*- C++ -*-

// **********************************************************
// File cursesviewer.h
// ----------------------------------------------------------
// Created: F. Kappe
// **********************************************************

#ifndef cursesviewer_h
#define cursesviewer_h

#include "hgobject.h"
//#include "offsetlist.h"
#include <stdio.h>

struct CMDIObjectInfo;

class TextLinkNode;
class LinkList;

// **********************************************************
// class Viewer
// ----------------------------------------------------------
// Abstract Viewer Class. Viewers display Hyper-G objects
// **********************************************************

class Viewer 
{
public:
    Viewer() {}
    virtual ~Viewer() {}

    virtual boolean Load(Document*, Anchor* = nil, boolean = true) = 0;
    virtual char Browse(ObjectPtr& next) = 0;
    virtual void Output(FILE* pipe) = 0;
};

// **********************************************************
// class TextViewer
// ----------------------------------------------------------
// Used to display TextDocuments or derived Documents. 
// **********************************************************

class OidUStringTree;

class TextViewer : public Viewer 
{
public:
    TextViewer(boolean html=false);
    virtual ~TextViewer();

    virtual boolean Load(Document* d, Anchor* a = nil, boolean dohyphen = true);
//    virtual boolean Load(HgStub*, Document*, Anchor* = nil, boolean = true);
    virtual char Browse(ObjectPtr& next); 
//    virtual char Browse(HgStub*, ObjectPtr& next);
    virtual void Output(FILE* f);
//    virtual void Output(HgStub*, FILE* pipe);

    RString* getData() const;

protected:
    int NextChar();
    int NextChar(HgStub*);

    long _offset;
    RString _text;
    RString _title;
    int _line;
    int _numannotations;
    RString tmpTextName_;
    int langcode_;
    RString sgmlCnf_;

    HgStub* dataBase_;
    Document* document_;
    LinkList* endAnchList_;   // linklist sorted after end anchors (not start anchors)

private: // variables used by NextChar():
    char _insert[1024];
    long _insertptr;
    long _insertlen;
    long _textptr;
    long _textlen;

    TextLinkNode* nextStart_;
    TextLinkNode* nextEnd_;
    long _startoffset;
    long _endoffset;
    int _anchor;
    boolean verbatim_;
    boolean _html_mode;
    char lastChar_;
};

inline RString* TextViewer::getData() const 
{
    return (RString*)&_text;
}

// **********************************************************
// class InfoViewer
// ----------------------------------------------------------
// Used to display the Object Record of a Hyper-G Object
// **********************************************************

class InfoViewer : public TextViewer 
{
public:
    InfoViewer() : TextViewer() {}
    char ShowStatus();
};

// **********************************************************
// class WWWViewer
// ----------------------------------------------------------
// Used to display the WordWideWeb Documents
// **********************************************************

class WWWViewer : public TextViewer 
{
public:
    WWWViewer() : TextViewer(true) {};
    boolean Load(Document*, Anchor* = nil, boolean = true);
    char Browse(ObjectPtr& next);
};

// **********************************************************
// class GopherViewer
// ----------------------------------------------------------
// Used to display the Gopher Text Documents
// **********************************************************

class GopherViewer : public TextViewer 
{
public:
    GopherViewer() : TextViewer() { viewerInfo_ = 0; }
    ~GopherViewer();

    boolean Load(Document*, Anchor* = nil, boolean = true);
    char Browse(ObjectPtr& next);
    char ImageDisplay(ObjectPtr& next);

private:
    RString _gophertype;
    RString tempName_;
    CMDIObjectInfo* viewerInfo_;    // internal viewer information 
};

// **********************************************************
// class WaisViewer
// ----------------------------------------------------------
// Used to display the WAIS Text Documents
// **********************************************************

class WaisViewer : public TextViewer 
{
public:
    boolean Load(Document*, Anchor* = nil, boolean = true);
    char Browse(ObjectPtr& next);
    char ImageDisplay(ObjectPtr& next);

private:
    RString _waistype;
};

#endif




