//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        hgbaseobj.cpp
//
// Purpose:     Basic Hyper-G object classes: class HGObject and derived Classes
//
// Created:     18 Jan 92    Frank Kappe (hgobject.C for HGC-T)
//                           Thomas Dietinger (modified for use with Amadeus)
//
// Modified:    31 Aug 95    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>


#include <stdafx.h>

#ifndef _CONSOLE
#include <hgapp.h>
#include <mainfrm.h>
#else
#include <pcutil/consutil.h>
#endif

#include <resource.h>

#include <fstream.h>
#include <direct.h>
#include <dos.h>
#include <fcntl.h>

#include <pcutil/pcstuff.h>
#include <pcutil/socket.h>

#include "hgobject.h"
#include "hglinks.h"

#include "baseprot.h"

#include <Dispatch/rpcbuf.h>
#include <Dispatch/rpcservice.h>
#include <server/hgstub.h>
#include "hgprot.h"
#include <l_server/s_ldbprt.hpp>

#include <utils/environ.h>
#include <hyperg/hgheader.h>
#include <hyperg/dcsender.h>

#include "objectlist.h"
#include "conf.h"
#include "language.h"
#include "utils.h"
#include "treaps.h"

#ifndef _CONSOLE
#include "cursesviewer.h"
#include "viewerlist.h"
#endif


#ifdef __AMADEUS__	// G. Nagler 14.3.95
#include <viewer/text/viewdoc.h>
#include <viewer/text/viewvw.h>
#endif

#if defined(_MSC_VER) & defined(_DEBUG)
#define new DEBUG_NEW
#endif

// global variables:
// ~~~~~~~~~~~~~~~~~
extern StringKeyRefTree Seen;     // Documents we have already seen
extern RString SortOrder;         // contains 'P' after search
#ifndef _CONSOLE
extern ViewerList TheViewerList;  // The list of active Viewers
#endif

int MaxFullTextDocs = 40;

const int MAXPATH = 132;


// HGObject statics:
// ~~~~~~~~~~~~~~~~~
ProtocolList HGObject::conOrProtList_;
ProtPtr HGObject::homeProtPtr_ = ProtPtr();
HGLanguages *HGObject::_hglang = HGLanguages::DefTable();  
ObjectListList HGObject::activeObjects_;  // The list of active collections

boolean Cluster::isInCluster_ = false;

const RString hgProtStr = "hyperg";
extern RString ldbProtStr;


extern BOOL findFileType(   const char *magicFileName,          // magic file
                            const char *testFileBuf, int bLen,  // first bLen bytes of file to examine
                            char *outBuf, int outBufLen         // output buffer 
                        );


#if defined(_MSC_VER) & defined(_DEBUG)
#define new DEBUG_NEW
#endif


///////////////////////////////////////////////////////////////////
// class CursesSender: Display progress indicator while sending
// (Helper class for putDocument)

class CursesSender : public DcFileSender 
{
public:
//    CursesSender(ifstream* s, int length);
    CursesSender(const char* filename, int length);
    ~CursesSender();
private:
    virtual int read_();
    long length_;
    long bytesread_;
};

CursesSender::CursesSender(const char* filename, int length) : DcFileSender(filename) 
{
    length_ = length; 
    bytesread_ = 0;
}

CursesSender::~CursesSender() 
{
}

int CursesSender::read_() 
{
    int b_read = -1;
    if (!guiControls()->interrupted())
    {
        b_read = DcFileSender::read_() ;
        if (b_read >= 0) 
        {
            bytesread_ += b_read;
            guiControls()->setProgIndicator(bytesread_, length_);
        }
    }
    return b_read;
}

/////////////////////////////////////////////////////////////////
// Class ObjectPtr: Pointer to HGObject (with resource management)

ObjectPtr::ObjectPtr(HGObject* o) 
{
    _obj = o;
    Resource::ref(_obj);
}

ObjectPtr::ObjectPtr(const ObjectPtr& p) 
{
    _obj = p.Obj();
    Resource::ref(_obj);
}

ObjectPtr::~ObjectPtr() 
{
    Resource::unref(_obj);
}

ObjectPtr& ObjectPtr::operator = (const ObjectPtr& p) 
{
    Resource::unref(_obj);
    _obj = p.Obj();
    Resource::ref(_obj);
    return *this;
}

ObjectPtr& ObjectPtr::operator = (HGObject* obj) 
{
    Resource::unref(_obj);
    _obj = obj;
    Resource::ref(_obj);
    return *this;
}


/////////////////////////////////////////////////////////////////
// Class HGObject (Hyper-G Object); abstract:

HGObject::HGObject(const Object& o, ProtPtr pPtr)
{
    activated_ = false;
    objType_ = HGObjectType;
    protPtr_ = pPtr;
    updateAttributes(o);
}


boolean HGObject::updateAttributes(Object o)
{
    _id = o.ID();
    _goid = o.field("GOid="); 
    _type = o.type();
    _author = o.author();
    _created = o.created();
    _modified = o.modified();
    _expires = o.expire();
    _opens  = o.open();
    _rights = o.field("Rights=");
    _parent = o.field("Parent=");
//    _remoteid = o.field("RemoteID=");
    _serv = o.field("Serv=");
    layout_ = o.field("Layout=");
    _sequence = o.field("Sequence=");
    if (!_sequence.length() && o.indexa("Sequence=") != -1)
        _sequence = " ";    // dummy inserted !
    Score(o.field("Score="));

    _keywords.free();
    o.fields("Keyword=", _keywords);
    _userattr.free();
    o.fields("UserAttr=", _userattr);
    _titles.free();
    o.fields("Title=", _titles);
    _title = Title();
    _language = _hglang->BestLanguage;
    if (_title.length())
        _title = _language + ":" + _title;

    _dbName = o.field("DBName=");

    return true;
}

boolean HGObject::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        if (objInfo.length())
            return updateAttributes(objInfo);
    }

    return false;
}


void HGObject::Score(const RString& s) 
{
    if (s != RString()) 
    {
        char buf[10];
        sprintf(buf, "%04d", atoi(s));
        _score = buf;
    }
}

void HGObject::Score(int s) 
{
    char buf[10];
    sprintf(buf, "%04d", s);
    _score = buf;
}


const RString HGObject::SGMLTitle()
{
    // try to find a title in our current user interface language
    RString title;

    title = _hglang->GetBestTitle((RStringArray&)_titles);
    _language = _hglang->BestLanguage;  // this is the object's language that fits our needs best

    return title;
}

const RString HGObject::Title()
{
    RString t = SGMLTitle();
    return t.SGMLtoISO();
}

RString HGObject::Info() const 
{
    RString tmp = "ObjectID=" + ID().IDString();
    if (GOID() != RString()) 
        tmp += "\nGOid=" + GOID();

    tmp += "\nType=" + Type();

    int i;
    for (i=0; i < _titles.count(); i++)
        tmp += "\nTitle=" + _titles[i];

    for (i=0; i < _keywords.count(); i++)
        tmp += "\nKeyword=" + _keywords[i];

    if (Author() != RString()) tmp += "\nAuthor=" + Author();
    if (Created()!= RString()) tmp += "\nTimeCreated=" + Created();
    if (Modified()!=RString()) tmp += "\nTimeModified=" + Modified();
    if (Opens()  != RString()) tmp += "\nTimeOpen=" + Opens();
    if (Expires()!= RString()) tmp += "\nTimeExpire=" + Expires();
    if (Serv()   != RString()) tmp += "\nServ=" + Serv(); 
    if (Sequence()!=RString()) 
        tmp += "\nSequence=" + Sequence();  // show sequence in info !
    if (_parent  != RString()) tmp += "\nParent=" + _parent;
    if (_rights  != RString()) tmp += "\nRights=" + _rights;
    if (Score()  != RString()) tmp += "\nScore=" + Score();
    if (Layout()  != RString()) tmp += "\nLayout=" + Layout();
    for (i=0; i < _userattr.count(); i++)
        tmp += "\nUserAttr=" + _userattr[i];
    return tmp;
}

RString HGObject::ClientAttributes() const 
{
    return "\nTitle\nKeyword\nAuthor\nTimeCreated\nTimeOpen\nTimeExpire\nRights\nSequence\nUserAttr\nLayout\n";
}

ObjectList* HGObject::references(ObjectList* oldList, BaseFD* fd) 
{ 
    if (protocol())
    {
        fd = protocol()->open(ObjectPtr(this), getReferencesMode);
        if (fd)
        {
            ObjectList* tmpList = protocol()->getList(fd, fd->dataLen_, oldList);
            protocol()->close(fd);
            return tmpList;
        }
        return 0;
    }

    return 0;
}


void HGObject::activate(boolean activate) 
{
    ObjectNode* isActive = 0;
    RString protType = protocol()->getProtType();

    // search for same protocol type (lists are sorted after protocol types!)
    ObjectListNode* n = 0;
    for (ObjectListNode* n2 = activeObjects_.getFirst(); 
         n2 && n2->firstObj() && n2->firstObj()->protocol()->getProtType() == protocol()->getProtType();
         n2 = activeObjects_.getNext(n2))
    {
         n = n2;    // store last correct value!
    }

    // list for protocol not found?
    if (!n)
    {
        n = new ObjectListNode();       // make new list of objects (for this protocol type)
        activeObjects_.addTail(n);      // and add it to global list
    }
    else
        isActive = n->list()->find(IDString()); // search in this list for the object

    if (isActive)
        activated_ = true;

    if (activate && !isActive)
    {
        n->list()->addTail(new ObjectNode(this));
        activated_ = true;
    }
    else
    if (!activate && isActive)
    {
        n->list()->remove(isActive);    // remove object from list
        delete isActive;
        if (!n->list()->count())        // last object in list?
        {
            activeObjects_.remove(n);   // yes, remove also this list from global list
            delete n;
        }
        activated_ = false;
    }
}
//////////////////////////////////////////////////////////////////////////////
// Helper routines for normalized query string generation
//

RString subst(const RString& in,const RString& from, const RString& to, int ignorecase=1)
{
    RString comparestring = in;
    RString fromstring = from;

    if (ignorecase)
    {
        comparestring.tolower();
        fromstring.tolower();
    }

    int index = 0;
    int where;
    RString tmp;
    int fromlength = from.length();

    while ((where = comparestring.indexa(index,fromstring)) != -1)
    {
        tmp += in.gSubstrIndex(index, where - fromlength - 1) + to;
        index = where;
    }
    if (index <= in.length())
    {
        tmp += in.gRight(index);
    }
    return tmp;
}

RString subst(const RString& in,const RString& from, const RString& to,
  const RString& delim, int ignorecase=1)
{
  RString comparestring=in;
  RString fromstring=from;
  RString delimstring=delim;
  if (ignorecase)
  {
    comparestring.tolower();
    fromstring.tolower();
    delimstring.tolower();
  }
  int index=0;
  int where;
  RString tmp;
  int fromlength=from.length();
  while ((where=comparestring.indexa(index,fromstring))!=-1)
  {
    if (((where-fromlength==0)||
      (delimstring.index(comparestring[where-fromlength-1])!=-1)) &&
      (((where==in.length()) || (delim.index(comparestring[where])!=-1))))
    {// test if trailing char is a delim
      tmp+=in.gSubstrIndex(index,where-fromlength-1)+to;
    }
    else
    {
      tmp+=in.gSubstrIndex(index,where-1);
    }
    index=where;
  }
  if (index<=in.length())
  {
    tmp+=in.gRight(index);
  }
  return tmp;
}

RString compressSpaces(const RString& text)
{
  RString tmp;
  RString tmptext;
  int index =text.length()-1;
  int from=0;
  int oldcharwasspace=0;
  while (text[from]==' ')from++;
  while (text[index]==' ')index--;
  tmptext=text.gSubstrIndex(from,index);  // remove trailing spaces and leading spaces
  // begin compression
  index=0;
  from=0;
  while (index<=tmptext.length())
  {
    if ((tmptext[index]==' ') && (!oldcharwasspace) || (tmptext[index]=='\0'))
    {
      tmp+=tmptext.gSubstrIndex(from,index-1);
      oldcharwasspace=1;
      from = index;
    }
    else if((tmptext[index]!=' ') && oldcharwasspace)
    { // this is to replace
      tmp+=" ";
      oldcharwasspace=0;
      from=index;
    }
    index++;
  }
  return tmp;
}

RString prepareForBoolQuery(const RString& text)
{
  RString tmp=text;
  tmp = subst(tmp,"AND","&&"," ()&|!");
  tmp = subst(tmp,"OR","||"," ()&|!");
  tmp = subst(tmp,"NOT","!"," ()&|!");
  tmp = subst(tmp,"ANDNOT","&!"," ()&|!");
  tmp = subst(tmp,"||","|"); // set all || to | dont change |
  tmp = subst(tmp,"|","||"); // set all | to ||
  tmp = subst(tmp,"&!","!");
  tmp = subst(tmp,"&&","&");
  tmp = subst(tmp,"&","&&");
  tmp = subst(tmp,"!","&!");
  tmp = subst(tmp,"\"","");
  return tmp;
}


//----------------- composeSimpleTitleKeywordBoolQuery  ---------------
RString composeSimpleTitleKeywordBoolQuery(const RString& simple)
{
    RString tmp;
    int i;
    for (i=0;i < HGObject::_hglang->Count();i++)
    {
        if (tmp=="")
	        tmp = tmp + "(Title=" + HGObject::_hglang->GetLangByOrder(i) + ":" + simple;
        else
	        tmp = tmp + "||Title=" + HGObject::_hglang->GetLangByOrder(i) + ":" + simple;
    }

    if (tmp =="")
        tmp = tmp + "(Keyword=" + simple;
    else
        tmp = tmp + "||Keyword=" + simple;

    tmp += ")";
    return tmp;
}

RString composeBoolQuery(RString searchText, RString authorText)
{
    RString tmp("");

    searchText = prepareForBoolQuery(searchText);
    authorText = prepareForBoolQuery(authorText);

    // first compose the title keyword query
    RString splitter("!&(|)"); 
    int index = 0;
    while (index < searchText.length())
    {
        if (splitter.index(searchText[index]) != -1)
        { // the character is one of the splitters
            tmp += searchText.gSubstrIndex(index,index);
            index++;
        }
        else
        { // it is the beginning of a word
            int from = index;
            while ((index < searchText.length()) && 
                   (splitter.index(searchText[index])==-1))
                index++;

            RString word = searchText.gSubstrIndex(from,index-1);
            tmp += composeSimpleTitleKeywordBoolQuery(word);
        }
    }

    if (authorText.length())
    {  // build the author query
        RString query2;  
        int index = 0;
        while (index < authorText.length())
        {
            if (splitter.index(authorText[index]) != -1)
            { // the character is one of the splitters
	            query2 += authorText.gSubstrIndex(index, index);
	            index++;
            }
            else
            { // it is the beginning of a word
	            int from = index;
	            while ((index < authorText.length()) && 
	                   (splitter.index(authorText[index]) == -1))
	                index++;
	            RString word = authorText.gSubstrIndex(from, index-1);
	            query2 = query2 + "Author=" + word;
            }
        }
        if (query2 != "")
        {
            if (tmp != "")
	            tmp = RString("(") + tmp + ")&&(" + query2 + ")";
            else 
                tmp = query2;
        }
    }

    return tmp;
}

//----------------- composeFTQuery ------------------------------------
RString composeFTQuery(RString tmp)
{
    RString searchText = compressSpaces(tmp);

    searchText = prepareForBoolQuery(searchText);
    searchText = subst(searchText," ","&",
                       "abcdefghijklmnopqrstuvwxyz0123456789-/");

    return searchText;
}

ObjectList* HGObject::keyQuery(RString& searchstr, 
                               boolean wholeDB, 
                               boolean score,
                               ObjectList* addList) 
{ 
    searchstr.SGMLISOtolowerASCII();
    RString query = composeBoolQuery(searchstr, "");

    long len = 0;

    if (wholeDB)
    {
        // search over all connection oriented protocols!
        for (ProtocolNode* protNode = HGObject::conOrProtList_.getFirst();
             protNode && protNode->protocol();
             protNode = HGObject::conOrProtList_.getNext(protNode))
        {
            BaseFD* fd = protNode->protocol()->open(ObjectPtr(0), keyQueryMode, 0, query);
            while (fd && fd->pos_ < fd->dataLen_)
            {
                len += fd->dataLen_;
                addList = protNode->protocol()->getList(fd, 10/*fd->dataLen_*/, addList);
                guiControls()->setProgIndicator(fd->pos_, len+10);
                if (guiControls()->interrupted())
                {
                    protNode->protocol()->close(fd);
                    guiControls()->clearProgIndicator();
                    return addList;
                }
            }

            protNode->protocol()->close(fd);
        } // for all protocols ...
        guiControls()->clearProgIndicator();
        return addList;
    } // whole DB
    else
    {
        // search for all activated protocols ...
        for (ObjectListNode* n = activeObjects_.getFirst(); 
             n && n->firstObj(); 
             n = activeObjects_.getNext(n))
        {
            BaseFD* fd = n->firstObj()->protocol()->open(ObjectPtr(0), keyQueryMode, n->list(), query);

            if (fd)
            {
                len += fd->dataLen_;
                while (fd->pos_ < fd->dataLen_)
                {
                    addList = n->firstObj()->protocol()->getList(fd, 10/*fd->dataLen_*/, addList);
                    guiControls()->setProgIndicator(fd->pos_, len+10);  // schummel, schummel ...
                }
            }

            n->firstObj()->protocol()->close(fd);
        }

        guiControls()->clearProgIndicator();
        return addList;
    }

    return 0;
}

ObjectList* HGObject::fulltextQuery(RString& searchstr, 
                                    boolean wholeDB,
                                    ObjectList* addList) 
{ 
    searchstr.SGMLISOtolowerASCII() ;
    RString ftQuery = composeFTQuery(searchstr);

    long len = 0;

    if (wholeDB)
    {
        // search over all connection oriented protocols!
        for (ProtocolNode* protNode = HGObject::conOrProtList_.getFirst();
             protNode && protNode->protocol();
             protNode = HGObject::conOrProtList_.getNext(protNode))
        {
            // search over all languages!
            for (int i = 0; i < HGObject::_hglang->Count(); i++)
            {
                BaseFD* fd = protNode->protocol()->open(ObjectPtr(0), fullTextQueryMode, 0, _hglang->GetLangByOrder(i) + ":" + ftQuery);

                if (fd)
                {
                    while ((fd->pos_ < fd->dataLen_) && ((fd->pos_+len) < MaxFullTextDocs))
                    {
                        addList = protNode->protocol()->getList(fd, 10/*fd->dataLen_*/, addList);
                        guiControls()->setProgIndicator(fd->pos_, len+fd->dataLen_+10);
                        if (guiControls()->interrupted())
                        {
                            protNode->protocol()->close(fd);
                            guiControls()->clearProgIndicator();
                            return addList;
                        }
                    }
                    len += fd->dataLen_;
                }
                protNode->protocol()->close(fd);
            } // for all languages ...
        } // for all protocols ...
        guiControls()->clearProgIndicator();
        return addList;
    } // whole DB
    else
    {
        for (ObjectListNode* n = activeObjects_.getFirst(); 
             n && n->firstObj(); 
             n = activeObjects_.getNext(n))
        {
            // search over all languages!
            for (int i = 0; i < HGObject::_hglang->Count(); i++)
            {
                BaseFD* fd = n->firstObj()->protocol()->open(ObjectPtr(0), fullTextQueryMode, n->list(), _hglang->GetLangByOrder(i) + ":" + ftQuery);

                if (fd)
                {
                    while ((fd->pos_ < fd->dataLen_) && ((fd->pos_+len) < MaxFullTextDocs))
                    {
                        addList = n->firstObj()->protocol()->getList(fd, 10/*fd->dataLen_*/, addList);
                        guiControls()->setProgIndicator(fd->pos_, len+10+fd->dataLen_);  // schummel, schummel ...
                        if (guiControls()->interrupted())
                        {
                            n->firstObj()->protocol()->close(fd);
                            guiControls()->clearProgIndicator();
                            return addList;
                        }
                    }
                    len += fd->dataLen_;
                }
                n->firstObj()->protocol()->close(fd);
            }
        }

        guiControls()->clearProgIndicator();
        return addList;
    }

    return 0;
}

ObjectList* HGObject::findObj(RString& searchstr, 
                              RString& author, 
                              RString& date, 
                              boolean wholeDB, 
                              boolean score,
                              ObjectList* addList)
{ 
    searchstr.SGMLISOtolowerASCII();
    author.SGMLISOtolowerASCII();

    RString query = composeBoolQuery(searchstr, author);

    if (date != "")
    {
        if (query.length())
            query += RString("&&") + "TimeCreated=" + date + "^";
        else
            query = "TimeCreated=" + date + "^";
    }

    return keyQuery(query, wholeDB, score, addList);
}


boolean HGObject::modifyAttribute(RString oldAttrLine,
                                  RString newAttrLine)
{
    if (protocol())
        return protocol()->modifyObjAttribute(ObjectPtr(this), oldAttrLine, newAttrLine);
    else
        return false;
}


boolean HGObject::lock()
{
    if (protocol())
        return !protocol()->lock(ObjectPtr(this)) ? false : true;
    else
        return false;
}

boolean HGObject::unlock()
{
    if (protocol())
        return protocol()->unlock(ObjectPtr(this));
    else
        return false;
}

RString HGObject::user()
{
    if (protocol())
        return protocol()->getUser();
    else
        return "anonymous";
}


ObjectPtr HGObject::getRecentParent()
{
    if (parentObjKey_.length())
        return ObjectPtr(makeObjectById(parentObjKey_));
    else
        return ObjectPtr(0);
}


boolean HGObject::readOnly()
{
    if (user() == "anonymous")
        return true;
    else
        return false;
}


/////////////////////////////////////////////////////////////////
// Class Anchor:

Anchor::Anchor(const Object& o) : HGObject(o) 
{
    objType_ = AnchorType;
    if (Type() == "Anchor") 
    {
        updateAttributes(o);
    }
    else 
    {
        ErrorMsg("Fatal Error: Anchor with Type=", Type());
        abort();
    }
}


boolean Anchor::updateAttributes(Object o)
{
    _position = o.position();
    _dest = o.field("Dest=");
    _gdest = o.field("GDest=");
    _src = o.field("Src=");
    _hint = o.field("Hint=");
    _linktype = o.field("LinkType=");
    _issrc = (o.field("TAnchor=") == "Src");
    _state = o.field("State=");
    getTarget();

    return true;
}


boolean Anchor::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            HGObject::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}


RString Anchor::Info() const 
{
    RString tmp = HGObject::Info();
    tmp += RString("\nTAnchor=") + (IsSrc() ? "Src" : "Dest");
    if (LinkType() != RString()) 
        tmp += "\nLinkType=" + LinkType();
    if (Position() != RString()) 
        tmp += "\nPosition=" + Position();
    if (Dest() != 0) 
        tmp += "\nDest=" + Dest();
    if (GDest() != RString()) 
        tmp += "\nGDest=" + GDest();
    if (Hint() != RString()) 
        tmp += "\nHint=" + Hint();
    if (State() != RString()) 
        tmp += "\nState=" + State();
    return tmp;
}

RString Anchor::ClientAttributes() const 
{
    return HGObject::ClientAttributes() + "Position\nLinkType\n";
}

/*
boolean Anchor::GetTarget(HgStub* db) 
{
    if (!db || !IDString().length())
        return false;

    if (!_target.Obj()) 
    {
        ObjectID targetid;
        if (db->getDocByAnchor(ID(), targetid)) 
            _target = ObjectPtr(makeObjectById(targetid.IDString()));
    }
    return true;  // everything ok
}
*/
boolean Anchor::getTarget() 
{ 
    // quick-hack ! ######
    if (protocol() && protocol()->getProtType() == hgProtStr)
    {
        if (!IDString().length())
            return false;

        if (!_target.Obj()) 
        {
            ObjectID targetid;
            if (((HyperGProt*)protocol())->serverPtr()->getDocByAnchor(ID(), targetid)) 
                _target = ObjectPtr(makeObjectById(targetid.IDString()));
        }
        return true;  // everything ok
    }

    // steve
    if (protocol() && protocol()->getProtType() == "hglocal")
    {
        if (!IDString().length())
            return false;

        if (!_target.Obj()  &&  !IsSrc()) 
        {
            ObjectID targetid = _src;
            _target = ObjectPtr(makeObjectById(targetid.IDString()));
        }
        return true;  // everything ok
    }

    return false;
//    return GetTarget(HG); 
}

char Anchor::browse(ObjectPtr& next, Anchor* a) 
{ 
    // To do ! ######
    // Work over this function !! ('seen' problem) ...
    if (getTarget())  // target could be retrieved
    {

        // mark destination as seen also
        KeyRefPtr dummy;
        if (!Seen.find( _target.Obj()->ID().IDString(),dummy))
            Seen.insert( _target.Obj()->ID().IDString(),dummy);
    
        next = _target;
        return _target.Obj()->browse(next, this);
    }
    return '\0'; // error occured
}

ObjectList* Anchor::myParents(ObjectList* oldList, BaseFD* fd) const 
{
    // quick-hack ! #####
    if (protocol() && protocol()->getProtType() == hgProtStr)
    {
        ObjIdArray parentids;
        if (!((HyperGProt*)protocol())->serverPtr()->GetParentsDocColl(ID(), parentids))
        {
            TRACE("Anchor::GetTarget:db->GetParentsDocColl failed !\n");
            ASSERT(0);
            return nil;
        }

        ObjectList* result = oldList;
        if (!result)
            result = new ObjectList();
        ReceiveObjects(*((HyperGProt*)protocol())->serverPtr(), parentids, nil, result);

        return result;
    }

    return 0;
} // Anchor::myParents


boolean Anchor::insert(
                       ObjectPtr srcOrDestDoc   // source (if it's a source anchor) or
                                                // destination (if it's a dest. anchor) DOCUMENT
                                                // Note: this is NOT the destination anchor !
                      ) 
{
    if (protocol())
        return protocol()->insert(ObjectPtr(this), srcOrDestDoc);

    return false;
}

boolean Anchor::erase()
{ 
    if (protocol())
        return protocol()->erase(ObjectPtr(this));

    return false;
}

boolean Anchor::moveTo(ObjectPtr fromColl, ObjectPtr toColl) 
{ 
    if (protocol())
        return protocol()->moveTo(ObjectPtr(this), fromColl, toColl);

    return false;
}

boolean Anchor::copyTo(ObjectPtr toColl) 
{ 
    if (protocol())
        return protocol()->copyTo(ObjectPtr(this), toColl);

    return false;
}

boolean Anchor::unlinkFrom(ObjectPtr fromColl) 
{ 
    if (protocol())
        return protocol()->unlinkFrom(ObjectPtr(this), fromColl);

    return false;
}

/////////////////////////////////////////////////////////////////
// Class User:

User::User(const Object& o) : HGObject(o) 
{
    objType_ = UserType;
    if (Type() == "User") 
    {
        updateAttributes(o);
    }
    else 
        ASSERT(0);
}

boolean User::updateAttributes(Object o)
{
    uName_ = o.field("UName=");
    home_ = o.field("Home=");
    desc_ = o.field("Description=");

    groups_.free();
    o.fields("Group=", groups_);
    passwds_.free();
    o.fields("Passwd=", passwds_);
    hosts_.free();
    o.fields("Host=", hosts_);

    return true;
}


boolean User::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            HGObject::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}

RString User::Info() const 
{
    RString tmp = HGObject::Info();

    tmp += RString("\nUName=") + uName();
    if (home() != RString()) 
        tmp += "\nHome=" + home();
    if (desc() != RString()) 
        tmp += "\nDescription=" + desc();

    int i;
    for (i=0; i < groups_.count(); i++)
        tmp += "\nGroup=" + groups_[i];

    for (i=0; i < hosts_.count(); i++)
        tmp += "\nHost=" + hosts_[i];

    for (i=0; i < passwds_.count(); i++)
        tmp += "\nPasswd=" + passwds_[i];

    return tmp;
}

RString User::ClientAttributes() const 
{
    return HGObject::ClientAttributes() + "Home\nDescription\nGroup\nHost\nPasswd\n";
}

ObjectList* User::getGroups()
{
    // ToDo !! #####
    return 0;
}

/////////////////////////////////////////////////////////////////
// Class Group:

Group::Group(const Object& o) : HGObject(o) 
{
    objType_ = GroupType;
    if (Type() == "Group") 
    {
        updateAttributes(o);
    }
    else 
        ASSERT(0);
}

boolean Group::updateAttributes(Object o)
{
    uGroup_ = o.field("UGroup=");
    desc_ = o.field("Description=");

    groups_.free();
    o.fields("Group=", groups_);

    return true;
}

boolean Group::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            HGObject::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}

RString Group::Info() const 
{
    RString tmp = HGObject::Info();

    tmp += RString("\nUGroup=") + uGroup();

    int i;
    for (i=0; i < groups_.count(); i++)
        tmp += "\nGroup=" + groups_[i];

    if (desc() != RString()) 
        tmp += "\nDescription=" + desc();

    return tmp;
}

RString Group::ClientAttributes() const 
{
    return HGObject::ClientAttributes() + "Description\nGroup\n";
}

ObjectList* Group::getGroups()
{
    // ToDo !! #####

    // query mit 'UGroup=<name>'
    return 0;
}


/////////////////////////////////////////////////////////////////
// Class Document; abstract:

Document::Document(const Object& o) : HGObject(o) 
{
    objType_ = DocumentType;
    if (Type() == "Document") 
    {
        _src_anchors = 0;
        _dest_anchors = 0;
        updateAttributes(o);
    }
    else 
    {
        ErrorMsg("Fatal Error: Document with Type=", Type());
        abort();
    }
}

boolean Document::updateAttributes(Object o)
{
    _doctype = o.documenttype();
    _subtype = o.field("SubType=");
    _name = o.name();
    _path = o.path();
    _presentation = o.field("PresentationHints=");
    if (!_presentation.length() && o.indexa("PresentationHints=") != -1)
        _presentation = " ";    // dummy inserted !
    _loaded = false;
    _anchors_loaded = false;
    if (!_src_anchors)
        _src_anchors = new ObjectList();
    if (!_dest_anchors)
        _dest_anchors = new ObjectList();
    _srcs_of_dest = nil;
    linkList_ = nil;

    viewer_ = nil;
    inlineID_ = o.field("InlineID=");

    if (DocType() == "Remote") 
    {
        _protocol = o.protocol();
        _host = o.host();
        _port = o.port();
        remoteDocType_ = o.field("RemoteDocType=");
    }

    return true;
}


boolean Document::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            HGObject::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}


Document::~Document() 
{
//    TRACE("Document::~Document(): %s\n", (const char*)Info());

    delete _src_anchors;
    delete _dest_anchors;

    delete linkList_;

    delete _srcs_of_dest;

#ifndef _CONSOLE
    if (viewer_)                                // we generated a viewer, try to delete it
    {
        boolean isInList = false;               // but first look whether it has not been deleted elsewhere !!
        ViewerNode* nextNode;
        for (ViewerNode* n = TheViewerList.getFirst(); n; )
        {
            nextNode = TheViewerList.getNext(n);
            if (n == viewer_)
            {
                TheViewerList.remove(viewer_);
                isInList = true;                // it has not been deleted elsewhere !!
                break;
            }
            n = nextNode;
        }

        if (isInList)
        {
            delete viewer_;

            // also delete related files !
            if (tempName_.length())
            {

                LPOFSTRUCT pTmpOf = new OFSTRUCT;
                HFILE hTest;
                if ((hTest = OpenFile(tempName_, pTmpOf, OF_READ)) != HFILE_ERROR)
                {
                    _lclose(hTest);
                    TRACE("Document::~Document(): delete file: %s\n", tempName_);
                    unlink(tempName_);
                }
                delete pTmpOf;
            }
        }
    }
#endif

}

RString Document::Info() const 
{
    RString tmp = HGObject::Info() + "\nDocumentType=" + DocType();
    if (Path()  != RString()) 
        tmp += "\nPath="  + Path();

    if (Presentation() != RString()) 
        tmp += "\nPresentationHints=" + Presentation();

    if (DocType() == "Remote") 
    {
        tmp += "\nProtocol=" + Protocol()
               + "\nHost=" + Host()
               + "\nPort=" + Port()
               + "\nRemoteDocType=" + remoteDocType();
    }

    return tmp;
}

RString Document::ClientAttributes() const 
{
    
    RString tmpAttr = HGObject::ClientAttributes() + "PresentationHints\n";
    if (DocType() == "Remote") 
        tmpAttr += "Protocol\nHost\nPort\nPath\n";

    return tmpAttr;
}

RString Document::IDString() const 
{
    if (DocType() == "Remote") 
        return Protocol() + "://" + Host() + ":" + Port() + toUDI(Path());
    else
        return HGObject::IDString();
}

/*
 * DoesFileExist
 *
 * Purpose:
 *  Determines if a file path exists
 *
 * Parameters:
 *  lpszFile        LPTSTR - file name
 *
 * Return Value:
 *  BOOL            TRUE if file exists, else FALSE.
 *
 */
BOOL WINAPI DoesFileExist(LPTSTR lpszFile)
{
	static const TCHAR *arrIllegalNames[] = 
	{
		TEXT("LPT1"), TEXT("LPT2"), TEXT("LPT3"),
		TEXT("COM1"), TEXT("COM2"), TEXT("COM3"), TEXT("COM4"),
		TEXT("CON"),  TEXT("AUX"),  TEXT("PRN")
	};

	// check is the name is an illegal name (eg. the name of a device)
	for (int i = 0; i < (sizeof(arrIllegalNames)/sizeof(arrIllegalNames[0])); i++)
	{
		if (lstrcmpi(lpszFile, arrIllegalNames[i])==0)
			return FALSE;
	}

	// check the file's attributes
	DWORD dwAttrs = GetFileAttributes(lpszFile);
	if (dwAttrs == 0xFFFFFFFF)
	{
		// look in path for file
		TCHAR ch2[2];
		LPTSTR lpszFilePart;
		return SearchPath(NULL, lpszFile, NULL, 2, ch2, &lpszFilePart) != 0;
	}
	else if (dwAttrs & FILE_ATTRIBUTE_DIRECTORY)
	{
		return FALSE;
	}
	return TRUE;
}

char Document::browse(ObjectPtr& next, Anchor* a) 
{ 
#ifndef _CONSOLE    // browsing not supported in console apps currently !

    // Attention: special Treatment for "Remote" documents!!
    //  => MUST be browsed using the protocol specified in the URL, 
    //     if they are NOT handeld by a Hyper-G gateway !!!!!

    if (!viewer_)
    {
        for (ViewerNode* n = TheViewerList.getFirst(); n; n = TheViewerList.getNext(n))
        {
            if (n->docPtr_.Obj()->ID() == ID() && 
                ((Document*)n->docPtr_.Obj())->DocType() != "Remote")   // don't cache remote docs !
            {
                loadAnchors();                      // load anchors anyway !
                // viewer with this ID already in list !
                viewerApp_ = n->viewerApp_;
                tempName_ = n->tempName_;
                if (!DoesFileExist((char*)tempName_.string()))
                {
                    getData(true);  // try to reload file!
                    if (!DoesFileExist((char*)tempName_.string()))
                        return 0;   // give it up!
                }

                n->viewDocument(this);              // activate this viewer
                return '\0';         
            }
        }

        // ToDO !!!! #####
//        if (((Document*)n->docPtr_.Obj())->DocType() != "Remote")
//        {
//            => Do something special with remote documents !!
//        }
//        else
//
        if (getData(true))
        {   
            loadAnchors();                          // load anchors !
            viewer_ = new ViewerNode(this);         // create new Viewer
            TheViewerList.addTail( viewer_ ); 
        }
    }
    else
    {
//        TRACE("Document::Browse: viewerApp: %s\n", viewerApp_);
        viewer_->viewDocument(this);
    }
#endif

    return '\0';
}


//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// split 'command' into individual arguments (with parameter replacement)
//
RString Document::parseViewerCmd(const RString& cmd, RString tmpFName) 
{
    RString cmdLine;
    RString args[10];     // 1st: file name, 
                          // 2nd-10th: additional arguments of Generic document      
    RString doctype = DocType();
    if (doctype == "Generic") 
    {
        GenericDocument* gd = (GenericDocument*)this;
        doctype += "." + gd->SubType();
    
        int pos = 0;
        int count = 0;
        while (gd->Arguments().GetWord(pos, " \t", args[count++]))
            if (count > 9) break;

//        if (count==1 && tmpFName.length())
//            args[0] = tmpFName;
    }
  
    // store file name in argument 0 !
    args[0] = tmpFName;

    // now we have to split the line into individual arguments
    int pos = 0;
    RString arg;
    boolean noSpace = true;
    while (cmd.GetWord(pos, " \t$", arg)) 
    {
        if ( ((pos - arg.length() - 1) >= 0) && (cmd[pos - arg.length() - 1] == '$') )
        {
            arg = "$" + arg;
            if ( ((pos - arg.length() - 1) >= 0) && (cmd[pos - arg.length() - 1] != ' ') )
                noSpace = true;
            else
                noSpace = false;
        }
        else
            noSpace = false;

        if (arg == "$title")
            arg = Title();
        else
        if (arg == "$doctype")
            arg = DocType();
        else 
        if (arg == "$path")
            arg = Path();
        else 
        if (arg == "$author")
            arg = Author();
        else 
        if (arg == "$created")
            arg = Created();
        else 
        if (arg == "$language")
            arg = Language();
        else 
        if (arg == "$subtype")
        {
            if (DocType() == "Generic")
                arg = ((GenericDocument*)this)->SubType();
            else 
                arg = "";
        }
        else 
        if (arg == "$host")
        {
            if (DocType() == "Remote")
                arg = ((RemoteDocument*)this)->Host();
            else 
                arg = "";
        }
        else 
        if (arg == "$port")
        {
            if (DocType() == "Remote")
                arg = ((RemoteDocument*)this)->Port();
            else 
                arg = "";
        }
        else 
        if (arg == "$arg0")
            arg = args[0];
        else 
        if (arg == "$arg1")
            arg = args[1];
        else 
        if (arg == "$arg2")
            arg = args[2];
        else 
        if (arg == "$arg3")
            arg = args[3];
        else 
        if (arg == "$arg4")
            arg = args[4];
        else 
        if (arg == "$arg5")
            arg = args[5];
        else 
        if (arg == "$arg6")
            arg = args[6];
        else 
        if (arg == "$arg7")
            arg = args[7];
        else 
        if (arg == "$arg8")
            arg = args[8];
        else 
        if (arg == "$arg9")
            arg = args[9];
    
        if (!noSpace && cmdLine.length())
            cmdLine += " ";
        cmdLine += arg;
    }

    if ( cmd.indexa("internal") == -1 && 
         cmd.indexa("$arg0") == -1 && DocType() != "Generic") // file name not specified so far, so do that now
        cmdLine += tmpFName;

    return cmdLine;
}


/////////////////////////////////////////////////////////////////////////////
// 

boolean Document::loadDocument(BaseFD* fd, RString& data)
{                
    data = RString::lambda();      
#ifdef _CONSOLE

    // currently this doesn't work in console mode apps !

    return 0;

#else
    const int MAXBUF = 512;
    CString tempNameExt;       // extension of temporary file name
    CString docType;
    long totalBytes = 0;
    char myBuf[MAXBUF+1];
    int bytesRead;

    theApp.BeginBusyCursor(); 

    do {
        if (guiControls()->interrupted()) {   // current operation interrupted by user !
            bytesRead = -1;
            break;
        }

        bytesRead = protocol()->getData(fd, myBuf, MAXBUF);   
        if (bytesRead <= 0)
            break;                              // Error or nothing read !
        
        totalBytes += bytesRead;                // dump out byte statistic

        if (fd->dataLen_ > 0)
            guiControls()->setProgIndicator(totalBytes, fd->dataLen_);
        else
            guiControls()->setProgIndicator((unsigned int)totalBytes);

        myBuf[bytesRead] = 0;

        data += RString(myBuf);

    } while (bytesRead > 0);                 // Attention: bytesRead == -1 => error occured !!
    
    guiControls()->clearProgIndicator();
    theApp.EndBusyCursor(); 
       
    return totalBytes == data.length(); //fd->dataLen_;
#endif  // ! _CONSOLE
} // LoadDocument

boolean Document::loadDocument(BaseFD* fd, RString& cmdLine, RString& tempName, boolean forceLoad)
{                      
#ifdef _CONSOLE

    // currently this doesn't work in console mode apps !

    return 0;

#else
    const int MAXBUF = 512;
    CString tempNameExt;       // extension of temporary file name
    CString docType;

    int saveUnknownDocType = theApp.GetPrivateProfileInt("DocManager", "SaveUnknownDocType", 0);

    if (!tempName.length())     // tempName already specified ! (no generic document)
        tempName = getTempName(this);
    
    TRACE("Document::loadDocument:tempName=%s\n", (const char*)tempName);

    if ( tempName.length() && !getCompleteFileName(tempName, forceLoad) )
    {                                   
        // file does not exist, so fetch it from server
        
        char myBuf[2*MAXBUF];
        int bytesRead;
        int totalBytes = 0;

        // evaluate document file type (= file name extension)
        bytesRead = protocol()->getData(fd, myBuf, MAXBUF); // get first 512 bytes to evaluate file type
        if (DocType() == "PostScript")   // special treatment for PostScript, Magic-File-Numbers
        {
            tempNameExt = getTempNameExt( myBuf, MAXBUF );
            if (!tempNameExt.GetLength())
                tempNameExt = "PS";             // don't work reliable here !
        }
        else
            tempNameExt = getTempNameExt( myBuf, MAXBUF );

        ASSERT(tempName.length());

        if (tempName.index('*') != -1)
        {
            RString tempName2 = tempName.gSubstrIndex(0, tempName.length()-2) + (const char*)tempNameExt;  // add file name extension
            tempName = tempName2;
        }

        TRACE("Document::loadDocument:tempName=%s\n", (const char*)tempName);
        ASSERT(tempName.length());

        // evaluate viewer application string (if we don't know it yet)
        RString viewerType = DocType();
        if (viewerType == "Remote")
            viewerType = remoteDocType();

        if (!cmdLine.length() && tempNameExt.GetLength()) 
            cmdLine = theApp.GetPrivateProfileString(viewerType, tempNameExt);

        if (!cmdLine.length())
        {
            if (!saveUnknownDocType)
            {
                ErrorMsg(IDS_MSGUNKNOWNDOCTYPE);
                return false;
            }
            else
            {
//####### dummy code - replace it with true save dialog ! ########
                TRACE("Document::loadDocument: unknown document type ! (%s)\n", (const char* )viewerType);
                CString msgStr;
                msgStr.LoadString(IDS_MSGUNKNOWNDOCTYPE);
                msgStr += " (" + CString((const char*)viewerType) + ")"; 
                ErrorMsg((const char*)msgStr);
                return false;
//################################################################
            }

        }
    
        CFile myFile;

        ASSERT(tempName.length());

        if (tempName.index('*') != -1)          // skip wildcards !
            tempName = tempName.gSubstrDelim(0, '*');

        if ( myFile.Open(tempName, CFile::modeWrite | CFile::modeCreate) )
        {
            totalBytes += bytesRead;
            myFile.Write(myBuf, (int)bytesRead);
            theApp.BeginBusyCursor(); 

            do
            {
                if (guiControls()->interrupted())    // current operation interrupted by user !
                {
                    bytesRead = -1;
                    break;
                }

                // bytesRead = mySock->Read(myBuf, MAXBUF);   
                bytesRead = protocol()->getData(fd, myBuf, MAXBUF);   
                              
                if (bytesRead <= 0)
                    break;                              // Error or nothing read !
        
                totalBytes += bytesRead;                // dump out byte statistic

                if (fd->dataLen_ > 0)
                    guiControls()->setProgIndicator(totalBytes, fd->dataLen_);
                else
                    guiControls()->setProgIndicator((unsigned int)totalBytes);

                myFile.Write(myBuf, (int)bytesRead);

            } while (bytesRead > 0);                 // Attention: bytesRead == -1 => error occured !!
    
            myFile.Close();

            guiControls()->clearProgIndicator();
            theApp.EndBusyCursor(); 
        
        }
        else
        {
            ErrorMsg("Sorry, can't create temporary file - skipping multimedia document !");
            return false;
        }
            
        if (bytesRead == -1)
        {
            unlink(tempName);                           // delete faulty file
            return false;
        }
    }
    else // file exists, or error ....
    {
        if (!tempName.length())
        {
            ErrorMsg(IDS_MSGERRLOADDOC);
            return false;
        }

        if (!cmdLine.length())
        {
            CString tmpNam = tempName;
            tempNameExt = tmpNam.Mid( tmpNam.ReverseFind('.') +1 );
            cmdLine = (const char*)theApp.GetPrivateProfileString(DocType(), tempNameExt);
            if (!cmdLine)
            {
                ErrorMsg(IDS_MSGUNKNOWNDOCTYPE);
                return false;
            }
        }
    }
    
    return true;

#endif  // ! _CONSOLE
} // LoadDocument


/////////////////////////////////////////////////////////////////////////////
// 
/*
boolean Document::loadDocument(CSocket* mySock, RString& cmdLine, RString& tempName, boolean forceLoad)
{                      
#ifdef _CONSOLE

    // currently this doesn't work in console mode apps !

    return 0;

#else
    const int MAXBUF = 512;
    CString tempNameExt;       // extension of temporary file name
    CString docType;

    int saveUnknownDocType = theApp.GetPrivateProfileInt("DocManager", "SaveUnknownDocType", 0);

    if (!tempName.length())     // tempName already specified ! (no generic document)
        tempName = getTempName(this);
    
    TRACE("Document::loadDocument:tempName=%s\n", (const char*)tempName);

    if ( tempName.length() && !getCompleteFileName(tempName, forceLoad) )
    {                                   
        // file does not exist, so fetch it from server
        
        char myBuf[2*MAXBUF];
        int bytesRead;
        int totalBytes = 0;
        int headerLen = 0; 

        bytesRead = mySock->Read(myBuf, MAXBUF);     // get header or first 512 bytes to evaluate file type
//        TRACE("Document::loadDocument: socket-read, %d bytes read\n", bytesRead);
        if (bytesRead == -1 || !bytesRead)
        {
            mySock->linger();
            mySock->Close();
//            TRACE("Document::loadDocument: didn't get data from socket, file: %s\n", (const char*)tempName);
            if (bytesRead == SOCKET_ERROR)
                ErrorMsg("Socket error:", WSAGetLastError());
            ErrorMsg(IDS_ERRNOSOCKREAD);
            return false;
        }

        HgHeader fileHeader;
        int headerRet = fileHeader.read(myBuf, bytesRead);
//        TRACE("Document::loadDocument: file-header, %d bytes\n", headerRet);
        int bytesRead2 = 0;
        if (headerRet != -1 && headerRet == bytesRead)
        {
            headerLen = headerRet;
            bytesRead = mySock->Read(myBuf, MAXBUF);     // get first 512 bytes to evaluate file type
            ASSERT(bytesRead>0);
        }
        else
        if (headerRet != -1 && headerRet < bytesRead)
        {
            memmove(myBuf, myBuf + headerRet, bytesRead - headerRet);
            bytesRead2 = mySock->Read(myBuf+bytesRead-headerRet, MAXBUF-bytesRead + headerRet);
//            TRACE("Document::loadDocument: socket-read2, %d bytes read\n", bytesRead2);
            if (bytesRead2 != -1)
                bytesRead += bytesRead2-headerRet;
            else
            {
                mySock->linger();
                mySock->Close();
//                TRACE("Document::loadDocument: didn't get data from socket, file: %s\n", (const char*)tempName);
                if (bytesRead2 == SOCKET_ERROR)
                    ErrorMsg("Socket error:", WSAGetLastError());
                ErrorMsg(IDS_ERRNOSOCKREAD);
                return false;
            }
        }

        // evaluate document file type (= file name extension)
        if (DocType() == "PostScript")   // special treatment for PostScript, Magic-File-Numbers
        {
            tempNameExt = getTempNameExt( myBuf, MAXBUF );
            if (!tempNameExt.GetLength())
                tempNameExt = "PS";             // don't work reliable here !
        }
        else
            tempNameExt = getTempNameExt( myBuf, MAXBUF );

        ASSERT(tempName.length());

        if (tempName.index('*') != -1)
        {
            RString tempName2 = tempName.gSubstrIndex(0, tempName.length()-2) + (const char*)tempNameExt;  // add file name extension
            tempName = tempName2;
        }

        TRACE("Document::loadDocument:tempName=%s\n", (const char*)tempName);
        ASSERT(tempName.length());

        // evaluate viewer application string (if we don't know it yet)
        RString viewerType = DocType();
        if (viewerType == "Remote")
            viewerType = remoteDocType();

        if (!cmdLine.length() && tempNameExt.GetLength()) 
            cmdLine = theApp.GetPrivateProfileString(viewerType, tempNameExt);

        if (!cmdLine.length())
        {
            if (!saveUnknownDocType)
            {
                mySock->linger();                   // prevent hang at Close
                mySock->Close();
                ErrorMsg(IDS_MSGUNKNOWNDOCTYPE);
                return false;
            }
            else
            {
//####### dummy code - replace it with true save dialog ! ########
                mySock->linger();                   // prevent hang at Close
                mySock->Close();
                TRACE("Document::loadDocument: unknown document type ! (%s)\n", (const char* )viewerType);
                CString msgStr;
                msgStr.LoadString(IDS_MSGUNKNOWNDOCTYPE);
                msgStr += " (" + CString((const char*)viewerType) + ")"; 
                ErrorMsg((const char*)msgStr);
                return false;
//################################################################
            }

        }
    
        CFile myFile;

        ASSERT(tempName.length());

        if (tempName.index('*') != -1)          // skip wildcards !
            tempName = tempName.gSubstrDelim(0, '*');

        if ( myFile.Open(tempName, CFile::modeWrite | CFile::modeCreate) )
        {
            totalBytes += bytesRead;
            myFile.Write(myBuf, (int)bytesRead);
            theApp.BeginBusyCursor(); 

            do
            {
                if (guiControls()->interrupted())    // current operation interrupted by user !
                {
                    bytesRead = -1;
                    break;
                }

                bytesRead = mySock->Read(myBuf, MAXBUF);   
                if (bytesRead <= 0)
                    break;                              // Error or nothing read !
        
                totalBytes += bytesRead;                // dump out byte statistic

                if (headerRet != -1 && fileHeader.size() > 0)
                    guiControls()->setProgIndicator(totalBytes, fileHeader.size());
                else
                    guiControls()->setProgIndicator((unsigned int)totalBytes);

                myFile.Write(myBuf, (int)bytesRead);

            } while (bytesRead > 0);                 // Attention: bytesRead == -1 => error occured !!
    
            myFile.Close();

            guiControls()->clearProgIndicator();
            theApp.EndBusyCursor(); 
        
        }
        else
        {
            ErrorMsg("Sorry, can't create temporary file - skipping multimedia document !");
            mySock->linger();                   // prevent hang at Close
            mySock->Close();
            return false;
        }
            
        mySock->linger();                   // prevent hang at Close
        mySock->Close();

        if (bytesRead == -1)
        {
            unlink(tempName);                           // delete faulty file
            return false;
        }
    }
    else // file exists, or error ....
    {
        mySock->linger();                   // prevent hang at Close
        mySock->Close();

        if (!tempName.length())
        {
            ErrorMsg(IDS_MSGERRLOADDOC);
            return false;
        }

        if (!cmdLine.length())
        {
            CString tmpNam = tempName;
            tempNameExt = tmpNam.Mid( tmpNam.ReverseFind('.') +1 );
            cmdLine = (const char*)theApp.GetPrivateProfileString(DocType(), tempNameExt);
            if (!cmdLine)
            {
                ErrorMsg(IDS_MSGUNKNOWNDOCTYPE);
                return false;
            }
        }
    }
    
    return true;

#endif  // ! _CONSOLE
} // LoadDocument
*/

/////////////////////////////////////////////////////////////////////////////
// 
// get real document data -
// return application start-string or empty string on failure !
boolean Document::getData(boolean forceLoad, RString tempName, RString cmdLine) 
{ 
    viewerApp_ = "";
    tempName_ = tempName;

    if (!protocol())
        return false;
    
    BaseFD* dataFD = protocol()->open(this, getDataMode);
    BOOL bSuccess = loadDocument(dataFD, cmdLine, tempName_, forceLoad);
    protocol()->close(dataFD);
    if (bSuccess)
    {
        if (cmdLine.length())        // get assoc. application starting-string
        {
            viewerApp_ = parseViewerCmd(cmdLine, tempName_); 
            return true;
        }
        else 
        {
            ErrorMsg(IDS_MSGUNKNOWNDOCTYPE);
            return false;
        } 
    }

    TRACE("Document::getData: loadDocument NOT successful !(tempName: %s, cmdLine: %s)\n", (const char*)tempName_, (const char*)cmdLine);
    return false;
}


boolean Document::putData(RString fName, const char* dataBuf, int dataLen)
{
    if (!protocol() || protocol()->getProtType() != hgProtStr)
        return false;

    HgStub* hyperG = ((HyperGProt*)protocol())->serverPtr();

#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  HGBASEOBJ.CPP putData");
    return false;
#else
    if (fName.length())     // we have a file to upload !
    {
        ifstream file(fName);
        file.seekg(0, fstream::end);
        int length = (int)file.tellg();
        file.seekg(0, fstream::beg);

        CursesSender sender(fName, length);
        RString emptyStr;
        if (!hyperG->putDocument(Environ::hostAddress(), sender.port(), _id, _path, emptyStr))
            HGError(*hyperG);

        return true;
    }
    else
    {
        TRACE("Not implemented yet !");
        ASSERT(0);
        return false;
    }
#endif
}

//////////////////////////////////////////////////////////////////////////////
ObjectList* Document::myParents(ObjectList* oldList, BaseFD* fd) 
{
    if (protocol())
    {
        BaseFD* fd = protocol()->open(ObjectPtr(this), getParentsMode);
        if (fd)
        {
            ObjectList* tmpList = protocol()->getList(fd, fd->dataLen_, oldList);
            protocol()->close(fd);
            return tmpList;
        }
    }

    return 0;
} // Document::MyParents

ObjectList* Document::srcAnchors(ObjectList* oldList, BaseFD* fd) 
{ 
    loadAnchors();
    return _src_anchors;
}

ObjectList* Document::destAnchors(ObjectList* oldList, BaseFD* fd) 
{ 
    loadAnchors();
    return _dest_anchors;
}

ObjectList* Document::links(ObjectList* oldList, BaseFD* fd) 
{ 
    ObjectList* result = oldList;
    if (!result)
        result = new ObjectList();

    if (!_src_anchors || !_src_anchors->count()) 
    {
        loadAnchors();
    }
    if (_src_anchors) 
    {
        for (ObjectNode* n = _src_anchors->getFirst(); n ; n = _src_anchors->getNext(n)) 
        {
            HGObject* obj = makeObjectById(((Anchor*)(n->ObjPtr().Obj()))->Dest(), n->ObjPtr().Obj()->protPtr_);
            if (obj)
                result->addTail(new ObjectNode(obj));      
        }
    }

    result->sortAndTrashDuplicates();

    return result;
/*
        // quick-hack ! #####
        if (protocol() && protocol()->getProtType() == hgProtStr)
        {
            ObjIdArray tmp;
            for (ObjectNode* n = _src_anchors->getFirst(); n ; n = _src_anchors->getNext(n)) 
            {
                ObjectID dest(((Anchor*)(n->ObjPtr().Obj()))->Dest());
                ObjectID doc;
                if (((HyperGProt*)protocol())->serverPtr() &&
                    dest.id() && 
                    ((HyperGProt*)protocol())->serverPtr()->getDocByAnchor(dest, doc)) 
                {
                    tmp.insert( doc);
                }
                else
                {
                    HGObject* obj = makeObjectById(((Anchor*)(n->ObjPtr().Obj()))->Dest(), n->ObjPtr().Obj()->protPtr_);
                    if (obj)
                        result->addTail(new ObjectNode(obj));      
                }
            }
            // only for Hyper-G objects !
            if (tmp.count()) 
            {
                ReceiveObjects(*((HyperGProt*)protocol())->serverPtr(), tmp, nil, result);
            }
        }
    }
    return result;

//    return LinksTo(HG); 
*/
}


//
// Note: return value destroyed by document destructor! (Don't delete it yourself!)
//
ObjectList* Document::references(ObjectList* oldList, BaseFD* fd) 
{ 
    if (!_srcs_of_dest && protocol())   // not yet loaded
    {
        fd = protocol()->open(ObjectPtr(this), getReferencesMode);
        if (fd)
        {
            _srcs_of_dest = protocol()->getList(fd, fd->dataLen_, oldList);
            protocol()->close(fd);

            // now add internal destination anchors to list of references
            loadAnchors();
            if (_dest_anchors) 
            {
                for (ObjectNode* n = _dest_anchors->getFirst();
                     n ; n = _dest_anchors->getNext(n))
                {
                    fd = protocol()->open(n->ObjPtr(), getReferencesMode);
                    if (fd)
                    {
                        if (fd->dataLen_ > 0)
                            _srcs_of_dest = protocol()->getList(fd, fd->dataLen_, _srcs_of_dest);
                        protocol()->close(fd);
                    }
                }
            }
            return _srcs_of_dest;
        }
        return 0;
    }

    return _srcs_of_dest;
}

boolean Document::insert(ObjectPtr fatherPtr) 
{
    if (protocol())
        return protocol()->insert(ObjectPtr(this), fatherPtr);

    return false;
}

boolean Document::erase() 
{ 
    if (protocol())
        return protocol()->erase(ObjectPtr(this));

    return false;
}

boolean Document::moveTo(ObjectPtr fromColl, ObjectPtr toColl) 
{ 
    if (protocol())
        return protocol()->moveTo(ObjectPtr(this), fromColl, toColl);

    return false;
}

boolean Document::copyTo(ObjectPtr toColl) 
{ 
    if (protocol())
        return protocol()->copyTo(ObjectPtr(this), toColl);

    return false;
}

boolean Document::unlinkFrom(ObjectPtr fromColl) 
{ 
    if (protocol())
        return protocol()->unlinkFrom(ObjectPtr(this), fromColl);

    return false;
}


void Document::reLoadAnchors() 
{
    _anchors_loaded = false;

    if (_src_anchors)
        _src_anchors->free();
    if (_dest_anchors)
        _dest_anchors->free();

//    delete linkList_;
//    linkList_ = 0;

    delete _srcs_of_dest;
    _srcs_of_dest = 0;

    loadAnchors();
}


////////////////////// protected: ///////////////////////////////

void Document::loadAnchors() 
{ 
    if (!_anchors_loaded && protocol()) // && protocol()->getProtType() == ldbProtStr)
    {
        BaseFD* fd = protPtr_.protocol()->open(this, getLinksMode);
        if (fd)
        {
            delete linkList_;
            linkList_ = new LinkList;

            ObjectList* tmpList = protocol()->getList(fd, fd->dataLen_);
            protocol()->close(fd);

            if (tmpList)
                for (ObjectNode* n = tmpList->getFirst(); n ; n = tmpList->getNext(n)) 
                {
                    Anchor* a = (Anchor*)n->ObjPtr().Obj(); // we are sure this is an anchor
                    if (a->IsSrc()) 
                    {
                        if (a->Dest() != 0)                 // valid source anchors ?
                            _src_anchors->addTail(new ObjectNode(ObjectPtr(a)));
                    }
                    else                                    // destination anchor
                        _dest_anchors->addTail(new ObjectNode(ObjectPtr(a)));
                }

            delete tmpList;

            _anchors_loaded = true;

            if (_src_anchors)
                for (ObjectNode* n = _src_anchors->getFirst(); n ; n = _src_anchors->getNext(n)) 
                    linkList_->insert((Anchor*)n->ObjPtr().Obj(), this);

            if (_dest_anchors)
                for (ObjectNode* n = _dest_anchors->getFirst(); n ; n = _dest_anchors->getNext(n)) 
                    linkList_->insert((Anchor*)n->ObjPtr().Obj(), this);
            return;
        }
        return;
    }
}


/////////////////////////////////////////////////////////////////
// Class Collection:

Collection::Collection(const Object& o, ObjectList* objList) : Document(o) 
{
    objType_ = CollectionType;
    _members = objList;
    if (DocType() == "collection" || DocType() == "Remote") 
    {
        updateAttributes(o);
    }
    else 
    {
        ErrorMsg("Fatal Error: Collection with DocumentType=", DocType());
        abort();
    }
}


boolean Collection::updateAttributes(Object o)
{
    _colltype = o.collectiontype();
    _descriptions.free();
    o.fields("Description=", _descriptions);
    _sortorder = o.field("SortOrder=");
    _subdocs = o.field("Subdocs=");

    return true;
}


boolean Collection::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            Document::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}


Collection::~Collection()
{
    delete _members;
    _members = 0;
}

const RString Collection::Description() const 
{
    return _hglang->GetBestTitle((RStringArray&)_descriptions);
}


RString Collection::Info() const 
{
    RString tmp = Document::Info();
    if (Name()  != RString()) 
        tmp += "\nName="  + Name();
    if (CollType() != RString()) 
        tmp += "\nCollectionType=" + CollType();
    if (SortOrder() != RString()) 
        tmp += "\nSortOrder=" + SortOrder();
    if (SubDocs() != RString()) 
        tmp += "\nSubdocs=" + SubDocs();
    if (License() != RString()) 
        tmp += "\nLicense=" + License();
    for (int i=0; i < _descriptions.count(); i++)
        tmp += "\nDescription=" + _descriptions[i];
    return tmp;
}

RString Collection::ClientAttributes() const 
{
    return Document::ClientAttributes() + "Name\nCollectionType\nDescription\nSortOrder\nLicense\n";
}

ObjectList* Collection::myParents(ObjectList* oldList, BaseFD* fd)
{
    if (protocol())
    {
        BaseFD* fd = protPtr_.protocol()->open(ObjectPtr(this), getParentsMode);
        if (fd)
        {
            ObjectList* tmpList = protocol()->getList(fd, fd->dataLen_, oldList);
            protocol()->close(fd);
            return tmpList;
        }
    }

    return 0;
}

ObjectList* Collection::myChildren(ObjectList* oldList, BaseFD* fd) 
{ 
    if (protocol())
    {
        if (!fd)
            fd = protPtr_.protocol()->open(ObjectPtr(this), getChildrenMode);
        if (fd)
        {
            ObjectList* tmpList = protocol()->getList(fd, fd->dataLen_, oldList);
            protocol()->close(fd);
            return tmpList;
        }
    }

    return 0;
}


ObjectList* Collection::members(ObjectList* oldList, BaseFD* fd) 
{
    if (!_members)
        _members = myChildren(oldList, fd);
    return _members;
}


// clean up list of members
void Collection::freeMembers() 
{ 
    delete _members; 
    _members = 0; 
}

boolean Collection::insert(ObjectPtr fatherPtr) 
{
    if (protocol())
        return protocol()->insert(ObjectPtr(this), fatherPtr);

    return false;
}

boolean Collection::erase() 
{ 
    if (protocol())
        return protocol()->erase(ObjectPtr(this));

    return false;
}

boolean Collection::moveTo(ObjectPtr fromColl, ObjectPtr toColl) 
{ 
    if (protocol())
        return protocol()->moveTo(ObjectPtr(this), fromColl, toColl);

    return false;
}

boolean Collection::copyTo(ObjectPtr toColl) 
{ 
    if (protocol())
        return protocol()->copyTo(ObjectPtr(this), toColl);

    return false;
}

boolean Collection::unlinkFrom(ObjectPtr fromColl) 
{ 
    if (protocol())
        return protocol()->unlinkFrom(ObjectPtr(this), fromColl);

    return false;
}


/////////////////////////////////////////////////////////////////
// Class Cluster:

Cluster::Cluster(const Object& o, ObjectList* objList) : Collection(o, objList) 
{
    objType_ = ClusterType;
    if (CollType() == "Cluster") 
    {
    }
    else 
    {
        ErrorMsg("Fatal Error: Cluster with CollectionType=", CollType());
        abort();
    }
    isInClusterSet_ = false;
}

Cluster::~Cluster() {
}
char Cluster::browse(ObjectPtr& next, Anchor* a) 
{ 
    KeyRefPtr dummy;
    UObjectID id;

    if (!members() || !members()->count()) // empty cluster
        return '\0';            // error flag

#ifdef __AMADEUS__
    if (!isInCluster_)
    {
        // we are currently not in a recursivly called cluster, thus
        // remove all previously displayed multimedia documents !
        CInfoNode* nextNode;
        for (CInfoNode* iNode = theApp.openViews_.getFirst(); iNode; )
        {
            nextNode = theApp.openViews_.getNext(iNode);
            if ((iNode->type_ != mainframe) && (iNode->type_ != collBrowser) && (iNode->type_ != textViewer))
            {
                if (iNode->info_.m_pDocument_)
                    iNode->info_.m_pDocument_->OnCloseDocument();
            }
            iNode = nextNode;
        }

        // ok, this might be the first of a recursivly called cluster, so set flag
        isInCluster_ = true;
        isInClusterSet_ = true;
    }
#endif

    boolean havetext = false;

//
//## adjust this according to new language classes !!
//
    // only one language-dependent document is shown per document type:
//    RString chosenlanguage = lang_code[::Language];
    RString chosenlanguage = _hglang->GetLangByOrder(0);    // get best language

    RStringArray docType;       // remember document types we see
    RStringUOidTree chosenDoc;  // stores the OID of the chosen Document 
                                // for each document type
    long index = 0;
    ObjectNode* n;

    // first we see if we have something in our UI language:
    for (n = _members->getFirst(); n ; n = _members->getNext(n)) 
    {
        ObjectPtr obj = n->ObjPtr();

        if (obj.Obj()->Type() == "Document" && 
            ((Document*)obj.Obj())->Presentation().indexa("Hidden") == -1 && 
            obj.Obj()->Languages() == 1) 
        {
            RString dt = ((Document*)obj.Obj())->DocType();
            if (chosenDoc.find(dt, id)) 
            { // we already had a document of this type
                if (((Document*)obj.Obj())->Language() == chosenlanguage) 
                {
                    chosenDoc.remove(dt, id);
                    id = UObjectID(index); // replace with this index to _members;
                    chosenDoc.insert(dt, id);
                }
            }
            else 
            {                           // first document of this type: always insert it
                id = UObjectID(index);  // set to this index to _members;
                chosenDoc.insert(dt, id);
                docType.insert(dt);
            }
        }
        index++;
    }

    // now show the text document:
    boolean textFound = chosenDoc.find("text", id);
    if (textFound)
    {
        index = id.id();
        n = (*_members)[(int)index];
        ObjectPtr chosentext = n->ObjPtr();

        // mark chosen doc as seen (cluster is marked automatically)
        if (!Seen.find( chosentext.Obj()->ID().IDString(),dummy))
            Seen.insert( chosentext.Obj()->ID().IDString(),dummy);
  
        next = chosentext; // by default
        chosentext.Obj()->browse(next);
    }
  
    // then show non-text documents:
    for (int i=0; i<docType.count(); i++) 
    {
        if (docType[i] != "text") 
        {
            chosenDoc.find(docType[i], id);

            index = id.id();
            ObjectNode* n = (*_members)[(int)index];
            ObjectPtr obj = n->ObjPtr();

            if (((Document*)obj.Obj())->Presentation().indexa("Hidden") == -1)
            {
                if (!Seen.find( obj.Obj()->ID().IDString(),dummy))
                    Seen.insert( obj.Obj()->ID().IDString(),dummy);
                obj.Obj()->browse(next, a);
            }
        }      
    }

    // finally show language-independent documents 
    // (are shown regardless of type (except for texts)):
    for (n = _members->getFirst(); n ; n = _members->getNext(n)) 
    {
        ObjectPtr obj = n->ObjPtr();
        if (obj.Obj()->Type() == "Document" && 
            ((Document*)obj.Obj())->Presentation().indexa("Hidden") == -1)
            if (((Document*)obj.Obj())->DocType() != "text")
            {
                if (obj.Obj()->Languages() > 1) 
                { // is NOT a multi-language document
                    if (!Seen.find( obj.Obj()->ID().IDString(),dummy))
                        Seen.insert( obj.Obj()->ID().IDString(),dummy);
                    obj.Obj()->browse(next, a);
                }
            }
    }

#ifndef __PC__
    TheViewerList.free();  // remove all other viewers when cluster is left
#endif
    
    if (isInClusterSet_)    // we leave now the last of alle recursivly called clusters, so reset flag
    {
        isInCluster_ = false;
        isInClusterSet_ = false;
    }
//    return retval;
    return DO_GUICMD;
}


class UObjectPtr : public ObjectPtr 
{
public:
    UObjectPtr(HGObject* o = nil)   : ObjectPtr(o)      {};
    UObjectPtr(const ObjectPtr& ptr) : ObjectPtr(ptr)    {};

private:
    friend class ObjPtrUStringTree;
    static UObjectPtr& mergeDataList (const UObjectPtr& a, const UObjectPtr& b) 
    {
        return *(new UObjectPtr());
    };
};


Treapdeclare(ObjPtrUStringTree, RString, UObjectPtr)
Treapimplement(ObjPtrUStringTree, RString, UObjectPtr)


// returns only visible children (chosen by currently selected language settings)
boolean Cluster::getVisibleChildren(ObjectList& visObjList, boolean recursive)
{
    if (!members() || !members()->count()) // empty cluster
        return false;

    // collect all multilingual documents
    ObjectNode* n;
    for (n = _members->getFirst(); n ; n = _members->getNext(n)) 
    {
        ObjectPtr obj = n->ObjPtr();

        if (obj.Obj()->Type() == "Document" && 
            ((Document*)obj.Obj())->Presentation().indexa("Hidden") == -1 && 
            obj.Obj()->Languages() != 1) 
        {
            if (((Document*)obj.Obj())->DocType() == "text")    // text documents should be
                visObjList.addHead(new ObjectNode(obj));        // displayed at first !
            else
            if (recursive &&
                ((Document*)obj.Obj())->DocType() == "collection" &&
                ((Collection*)obj.Obj())->CollType() == "Cluster")
            {
                getVisibleChildren(visObjList, recursive);
            }
            else
                visObjList.addTail(new ObjectNode(obj));
        }
    }


    RStringList docTypeList;
    ObjPtrUStringTree chosenDoc;    // stores the ObjectPtr of the chosen Document 
                                    // for each document type
    long index = 0;

    UObjectPtr oldObj;

    // examine language specific documents
    for (n = _members->getFirst(); n ; n = _members->getNext(n)) 
    {
        ObjectPtr obj = n->ObjPtr();

        if (obj.Obj()->Type() == "Document" && 
            ((Document*)obj.Obj())->Presentation().indexa("Hidden") == -1 && 
            obj.Obj()->Languages() == 1) 
        {
            RString dt = ((Document*)obj.Obj())->DocType();
            if (chosenDoc.find(dt, oldObj))
            {   
                // we already had a document of this type
                // has the new object a 'better' language than the stored one ?
                if (_hglang->GetOrder(obj.Obj()->Language()) < _hglang->GetOrder(oldObj.Obj()->Language())) 
                {
                    chosenDoc.remove(dt, UObjectPtr(oldObj));
                    chosenDoc.insert(dt, obj);
                }
            }
            else 
            {
                // first document of this type: always insert it
                chosenDoc.insert(dt, obj);
                docTypeList.addTail(new RStringNode(dt));
            }
        }
    }

    // ...
    for (RStringNode* strNode = docTypeList.getFirst(); 
        strNode; 
        strNode = docTypeList.getNext(strNode)) 
    {
        if (chosenDoc.find(strNode->str_, oldObj))
        {
            if (((Document*)oldObj.Obj())->DocType() == "text")    // text documents should be
                visObjList.addHead(new ObjectNode(oldObj));        // displayed at first !
            else
            if (recursive &&
                ((Document*)oldObj.Obj())->DocType() == "collection" &&
                ((Collection*)oldObj.Obj())->CollType() == "Cluster")
            {
                getVisibleChildren(visObjList, recursive);
            }
            else
                visObjList.addTail(new ObjectNode(oldObj));
        }
    }
    return true;
}


/////////////////////////////////////////////////////////////////
// warning, kludge ahead:
//
// Class FakeObject (e.g. for displaying Command List with same routine
// as ordinary Hyper-G Objects

FakeObject::FakeObject(const Object& o) : HGObject(o) 
{
    objType_ = FakeObjectType;
}

