//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        hgdocobj.cpp
//
// Purpose:     Hyper-G document classes (from Document derived Classes)
//
// Created:     18 Jan 92    Frank Kappe (hgobject.C for HGC-T)
//                           Thomas Dietinger (modified for use with Amadeus)
//
// Modified:    31 Aug 95    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>


#include <stdafx.h>

#ifndef _CONSOLE
#include <hgapp.h>
#include <mainfrm.h>
#endif

#include <resource.h>

#include <fstream.h>
#include <direct.h>
#include <dos.h>
#include <fcntl.h>

#include <pcutil/pcstuff.h>
#include <pcutil/socket.h>

#include "hgobject.h"
#include "hglinks.h"
#include "baseprot.h"
#include <l_server/s_ldbprt.hpp>

#include <Dispatch/rpcbuf.h>
#include <Dispatch/rpcservice.h>
#include <server/hgstub.h>
#include "hgprot.h"

#include <utils/environ.h>
#include <hyperg/hgheader.h>

#include "objectlist.h"
#include "conf.h"
#include "language.h"
#include "utils.h"
#include "treaps.h"

#ifndef _CONSOLE
#include "cursesviewer.h"
#include "viewerlist.h"
#endif

#ifdef __AMADEUS__
#include <viewer/text/viewdoc.h>
#include <viewer/text/viewvw.h>
#endif

// global variables:
StringKeyRefTree ActiveIDStrings; // IDStrings of active remote databases
extern StringKeyRefTree Seen;     // Documents we have already seen

#if defined(_MSC_VER) & defined(_DEBUG)
#define new DEBUG_NEW
#endif

const RString hgProtStr = "hyperg";
extern RString ldbProtStr;

/////////////////////////////////////////////////////////////////
// Class TextDocument:

TextDocument::TextDocument(const Object& o) : Document(o) 
{
    objType_ = TextDocumentType;
    if (DocType() == "text" || DocType() == "Remote") 
    {
        _viewer = nil;
    } 
    else 
    {
        ErrorMsg("Fatal Error: TextDocument with DocumentType=", DocType());
        abort();
    }
}

TextDocument::~TextDocument() 
{
    delete _viewer;
}

char TextDocument::browse(ObjectPtr& next, Anchor* dest) 
{ 
    if (!protocol())
        return 0;
        
    if (!_viewer)
        _viewer = new TextViewer();

    if (!_loaded)
    {
        if (getData(true))
        {
            loadAnchors();                        // load anchors !
            _loaded = _viewer->Load(this, dest);
        }
    }

    return _loaded ? _viewer->Browse(next) : '\0';

    return 0;
}

boolean TextDocument::getData(boolean forceLoad, RString tempName, RString cmdLine) 
{ 
    if (!protocol())
        return false;

    viewerApp_ = "";
    tempName_ = "";
    {
        BaseFD* dataFD = protocol()->open(this, getDataMode);
        BOOL bSuccess = loadDocument(dataFD, text_);
        protocol()->close(dataFD);
        viewerApp_ = "internal";
        ASSERT(text_.length());

        // For our old SGML Parser: check for doctype definition and delete it !
        int pos = text_.indexa("<!DOCTYPE");
        if (pos == -1)
            pos = text_.indexa("<!doctype");
        if (pos == -1)
            pos = text_.indexa("<!DocType");
        if (pos > -1)
        {
            pos -= 10;   // substract length of "<!DOCTYPE"
            int pos2 = text_.indexa(pos, ">");
            RString temp = text_.gSubstrIndex(0, pos);
            temp += text_.gSubstrIndex(pos2, text_.length());
            text_ = temp;
        }
    
        return bSuccess;
    }
    return false;
}


////////////////////// protected: ///////////////////////////////
void TextDocument::loadAnchors() 
{ 
    if (!_anchors_loaded) 
    {
        if (DocType() == "Remote")
        {
            delete linkList_;
            linkList_ = new LinkList;

            _anchors_loaded = true;

            if (Protocol() == "http")  // http: load links seperately !
            {
                getHTMLLinks(this, RString(""), text_, linkList_);

                for (TextLinkNode* n = (TextLinkNode*)linkList_->getFirst(); 
                    n; 
                    n = (TextLinkNode*)linkList_->getNext(n)) 
                {
                    Object o = "ObjectID=" + n->getSrcId() + "\nType=Anchor\nAuthor=system";
                    o += "\nTimeCreated=" + TimeNow() + "\nTAnchor=Src\nDest=" + n->getDestId() + "\n";
                    HGObject* a = MakeObject(o);
                    _src_anchors->addTail(new ObjectNode(ObjectPtr(a)));
                }
            }
        }
        else
            Document::loadAnchors();
    }
}


/////////////////////////////////////////////////////////////////
// Class PICDocument:

PICDocument::PICDocument(const Object& o) : Document(o) 
{
    objType_ = PICDocumentType;
    if (DocType() != "PIC" || DocType() == "Remote") 
    {
        ErrorMsg("Fatal Error: PICDocument with DocumentType=", DocType());
        abort();
    }
}

/////////////////////////////////////////////////////////////////
// Class ImageDocument:

ImageDocument::ImageDocument(const Object& o) : Document(o) 
{
    objType_ = ImageDocumentType;
    if (DocType() != "Image" && DocType() != "Remote") 
    {
        ErrorMsg("Fatal Error: ImageDocument with DocumentType=", DocType());
        abort();
    }
}

/////////////////////////////////////////////////////////////////
// Class SoundDocument:

SoundDocument::SoundDocument(const Object& o) : Document(o) 
{
    objType_ = SoundDocumentType;
    if (DocType() != "Sound" && DocType() != "Remote") 
    {
        ErrorMsg("Fatal Error: SoundDocument with DocumentType=", DocType());
        abort();
    }
}

/////////////////////////////////////////////////////////////////
// Class MovieDocument:

MovieDocument::MovieDocument(const Object& o) : Document(o) 
{
    objType_ = MovieDocumentType;
    if (DocType() != "Movie" && DocType() != "Remote") 
    {
        ErrorMsg("Fatal Error: MovieDocument with DocumentType=", DocType());
        abort();
    }
}

/////////////////////////////////////////////////////////////////
// Class PostScriptDocument:

PostScriptDocument::PostScriptDocument(const Object& o) : Document(o) 
{
    objType_ = PostScriptDocumentType;
    if (DocType() != "PostScript" && DocType() != "Remote") 
    {
        ErrorMsg("Fatal Error: PostScriptDocument with DocumentType=", DocType());
        abort();
    }
}

/////////////////////////////////////////////////////////////////
// Class SceneDocument:

SceneDocument::SceneDocument(const Object& o) : Document(o) 
{
    objType_ = SceneDocumentType;
    if (DocType() != "Scene" && DocType() != "Remote") 
    {
        ErrorMsg("Fatal Error: SceneDocument with DocumentType=", DocType());
        abort();
    }
}

/////////////////////////////////////////////////////////////////
// Class GenericDocument:

GenericDocument::GenericDocument(const Object& o) : Document(o) 
{
    objType_ = GenericDocumentType;
    if (DocType() == "Generic" || DocType() == "Remote") 
    {
//        _subtype = o.field("SubType=");
        updateAttributes(o);
    }
    else 
    {
        ErrorMsg("Fatal Error: GenericDocument with DocumentType=", DocType());
        abort();
    }
}

boolean GenericDocument::updateAttributes(Object o)
{
    _arguments = o.field("Arguments=");

    return true;
}


boolean GenericDocument::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            Document::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}


RString GenericDocument::Info() const 
{
    RString tmp = Document::Info();
    if (SubType()   != RString()) 
        tmp += "\nSubType="  + SubType();
    if (Arguments() != RString()) 
        tmp += "\nArguments="  + Arguments();
    return tmp;
}

RString GenericDocument::ClientAttributes() const 
{
    return Document::ClientAttributes() + "SubType\nArguments\n";
}

boolean GenericDocument::loadDocument(BaseFD* fd, RString& cmdLine, RString& tempName, boolean forceLoad)
{                      
#ifdef _CONSOLE     // not supported in console apps currently !
    return 0;
#else
    // evaluate commandline and temp name for generic objects
    GenericDocument* gd = (GenericDocument*)this;

    int pos = 0;
    gd->Arguments().GetWord(pos, " \t", tempName);

    if (tempName.length())
    {
        RString tempDir = getenv("TEMP");
        if (tempDir.gRight(1) != '\\')
            tempDir += RString("\\");
        tempName = tempDir + tempName;
    }
    else
    {
        tempName = tempnam(getenv("TEMP"), "GenOb");    // no argument specified !
        tempName += ".";
    }

    cmdLine = theApp.GetPrivateProfileString(DocType(), gd->SubType());

    return Document::loadDocument(fd, cmdLine, tempName, forceLoad);
#endif
}

/////////////////////////////////////////////////////////////////
// Class RemoteDocument; abstract:

RemoteDocument::RemoteDocument(const Object& o) : Document(o) 
{
    objType_ = RemoteDocumentType;
}

/////////////////////////////////////////////////////////////////
// Class WWWDocument:

WWWDocument::WWWDocument(const Object& o) : RemoteDocument(o) 
{
    objType_ = WWWDocumentType;
    if (Protocol() == "http")
    {
        _viewer = nil;
        updateAttributes(o);
    }
    else 
    {
        ErrorMsg("Fatal Error: WWWDocument with Protocol =", Protocol());
        abort();
    }
}

boolean WWWDocument::updateAttributes(Object o)
{
    _wwwtype = o.field("\nWWWType=");

    return true;
}


boolean WWWDocument::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            Document::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}


WWWDocument::~WWWDocument() 
{
    delete _viewer;
}

RString WWWDocument::Info() const 
{
    return RemoteDocument::Info() + "\nWWWType=" + WWWType();
}

char WWWDocument::browse(ObjectPtr& next, Anchor* a) 
{ 
    if (remoteDocType() == "text" || !remoteDocType().length())
    {
        // currently we support only text documents !
        if (!_viewer)
            _viewer = new WWWViewer();
        if (!_loaded)
        {
            delete linkList_;
            linkList_ = new LinkList;

            _loaded = _viewer->Load(this);
        }

        return _loaded ? _viewer->Browse(next) : '\0';
    }
    else
    {
        TRACE("WWWDocument::Browse: remoteDocType: %s\n", remoteDocType().string());
        return 0;
    }
}

boolean WWWDocument::getData(boolean forceLoad, RString tempName, RString cmdLine) 
{ 
    return Document::getData(forceLoad, tempName, cmdLine);
}

ObjectList* WWWDocument::links(ObjectList* oldList, BaseFD* fd) 
{ 
    ObjectList* _result = new ObjectList();
    RString text, title; 
    if (remoteDocType() == "text" && 
        getHTMLLinks((WWWDocument*)this, title, text, linkList()))
    {
        for (TextLinkNode* n = (TextLinkNode*)linkList()->getFirst(); 
            n; 
            n = (TextLinkNode*)linkList()->getNext(n)) 
        {
            HGObject* obj = makeObjectById(n->getDestId());
            if (obj) 
            {
                ObjectNode* n = new ObjectNode(obj);
                _result->addTail( n);
            }
        }
    }
    return _result;
}

/////////////////////////////////////////////////////////////////
// Class GopherDocument; abstract:

GopherDocument::GopherDocument(const Object& o) : RemoteDocument(o) 
{
    objType_ = GopherDocumentType;
    if (Protocol() == "gopher") 
    {
        _viewer = nil;
        updateAttributes(o);
    }
    else 
    {
        ErrorMsg("Fatal Error: GopherDocument with Protocol:", Protocol());
        abort();
    }
}

boolean GopherDocument::updateAttributes(Object o)
{
    _gophertype = o.gophertype();

    return true;
}


boolean GopherDocument::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            Document::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}


GopherDocument::~GopherDocument() 
{
    delete _viewer;
}

RString GopherDocument::Info() const 
{
    return RemoteDocument::Info() + "\nGopherType=" + GopherType();
}

RString GopherDocument::ClientAttributes() const 
{
    return RemoteDocument::ClientAttributes() + "GopherType\n";
}

ObjectList* GopherDocument::Gophers(HgStub* db, const RString& text) const 
{
    return 0;
}

ObjectList* GopherDocument::gophers(const RString& text) const 
{ 
//    if (protocol() && protocol()->getProtType() == hgProtStr)
    {
//        HgStub* db = ((HyperGProt*)protocol())->serverPtr();

        ObjectList* l = new ObjectList();
        int pos = 0;
        int count=0;
        RString line;
        RString _r_n = "\r\n" ;
        while (text.GetWord(pos, _r_n, line) && line[0] != '.') 
        {
            RString title = line.GetSubstrDelim(1, '\t');
            RString path  = line.GetSubstrDelim(title.length()+2, '\t');
            RString host  = line.GetSubstrDelim(path.length()+title.length()+3, '\t');
            RString port  = line.GetRight(host.length()+path.length()+title.length()+4);
            Object o = "ObjectID=0x00000000\nType=Document\nDocumentType=Remote\n";
            char buf[5];
            count += 10;
            sprintf(buf, "%d", count);
            if (line[0] == '8') 
            { // make it a TelnetCommand
                if (port == "0") 
                    port = TELNETPORT;
                o += "Protocol=telnet\nHost=" + host + "\nPort=" + port + "\nPath=" + path + 
                     "\nTitle=??:" + title + "\nSequence=" +  RString(buf) + "\nAuthor=";
//                if (db)
//                    o += db->user().GetSubstrDelim(0,'(') +  "\n";
//                else
                    o += "system\n";
            }
            else 
            {
                o += "Protocol=gopher\nGopherType=" + line.GetSubstrIndex(0,0) +
                     "\nTitle=??:" + title + "\nPath=" + path + "\nHost=" + host +
                     "\nPort=" + port + "\nSequence=" + RString(buf) + "\nAuthor=";
//                if (db)
//                    o += db->user().GetSubstrDelim(0,'(') +  "\n";
//                else
                    o += "system\n";
            }
            HGObject* r = MakeObject(o);
            if (r) 
            {
                ObjectNode* n = new ObjectNode(r);
                l->addTail( n);
            }
        }
        return l;
    }

    return 0;
}


char GopherDocument::browse(ObjectPtr& next, Anchor* a) 
{ 
    if (!_viewer || !_loaded) 
    {
        if (!_viewer)
            _viewer = new GopherViewer();

        _loaded = _viewer->Load(this);

        if (!_loaded && _viewer->getData())     // user aborted loading, but we got some data
            return _viewer->Browse(next);   // -> try to display it. Next time reload whole doc. !
    }
    return _loaded ? _viewer->Browse(next) : '\0';
}


/////////////////////////////////////////////////////////////////
// Class GopherText:

GopherText::GopherText(const Object& o) : GopherDocument(o) 
{
    objType_ = GopherTextType;
}

/////////////////////////////////////////////////////////////////
// Class GopherImage:

GopherImage::GopherImage(const Object& o) : GopherDocument(o) 
{
    objType_ = GopherImageType;
    if (GopherType() != "I") // old
        if (GopherType() != "g") //old
            if (GopherType() != ":") 
            {
                ErrorMsg("Fatal Error: GopherImage with GopherType=", GopherType());
                abort();
            }
}

char GopherImage::browse(ObjectPtr& next, Anchor* a) 
{ 
    if (!_viewer || !_loaded) 
    {
        if (!_viewer)
            _viewer = new GopherViewer();

        _loaded = _viewer->Load(this);

        if (!_loaded && _viewer->getData())
            return _viewer->ImageDisplay(next);
    }
    return _loaded ? _viewer->ImageDisplay(next) : '\0';
}


/////////////////////////////////////////////////////////////////
// Class GopherSound:

GopherSound::GopherSound(const Object& o) : GopherDocument(o) 
{
    objType_ = GopherSoundType;
    if (GopherType() != "S") // old
        if (GopherType() != "s") //old
            if (GopherType() != "<") 
            {
                ErrorMsg("Fatal Error: GopherSound with GopherType=", GopherType());
                abort();
            }
}

char GopherSound::browse(ObjectPtr& next, Anchor* a) 
{ 
    if (!_viewer || !_loaded) 
    {
        if (!_viewer)
            _viewer = new GopherViewer();

        _loaded = _viewer->Load(this);

        if (!_loaded && _viewer->getData())
            return _viewer->ImageDisplay(next);
    }
    return _loaded ? _viewer->ImageDisplay(next) : '\0';
}


/////////////////////////////////////////////////////////////////
// Class GopherCollection:

GopherCollection::GopherCollection(const Object& o) : GopherDocument(o) 
{
    objType_ = GopherCollectionType;
    if (GopherType() != "1" && GopherType() != "11" ) 
    {
        ErrorMsg("Fatal Error: GopherCollection with GopherType=", GopherType());
        abort();
    }
}

ObjectList* GopherCollection::myChildren(ObjectList* oldList, BaseFD* fd)
{ 
    RString text;
    LoadGopher(this, text);
    return gophers(text);
}


/////////////////////////////////////////////////////////////////
// Class GopherIndex:

GopherIndex::GopherIndex(const Object& o) : GopherDocument(o) 
{
    objType_ = GopherIndexType;
    if (GopherType() != "7") 
    {
        ErrorMsg("Fatal Error: GopherIndex with GopherType=", GopherType());
        abort();
    }
}

/////////////////////////////////////////////////////////////////
// Class WaisDocument; abstract:

WaisDocument::WaisDocument(const Object& o) : RemoteDocument(o) 
{
    objType_ = WaisDocumentType;
    if (Protocol() == "wais") 
    {
        updateAttributes(o);
        _viewer = nil;
    }
    else 
    {
        ErrorMsg("Fatal Error: WaisDocument with Protocol=", Protocol());
        abort();
    }
}

boolean WaisDocument::updateAttributes(Object o)
{
    _size = o.field("Size=");
    _waistype = o.field("WaisType=");
    _docid = o.field("WaisDocID=");

    return true;
}


boolean WaisDocument::update() 
{
    if (protocol())
    {
        Object objInfo = protocol()->getObjectInfo(IDString());
        {
            Document::updateAttributes(objInfo);
            return updateAttributes(objInfo);
        }
    }

    return false;
}


RString WaisDocument::Info() const 
{
    RString tmp = RemoteDocument::Info() + "\nWaisType=" + WaisType();
    if (Size()  != RString()) tmp += "\nSize=" + Size();
//  if (DocID() != RString()) tmp += "\n" + BeautifulWaisDocID (DocID());   // in waisutils.C

    return tmp;
}

RString WaisDocument::ClientAttributes() const 
{
    return RemoteDocument::ClientAttributes() + "WaisType\n";
}

char WaisDocument::browse(ObjectPtr& next, Anchor* a) 
{ 
    return 0;
}

/////////////////////////////////////////////////////////////////
// Class WaisSource:

WaisSource::WaisSource(const Object& o) : WaisDocument(o) 
{
    objType_ = WaisSourceType;
    if (WaisType() != "WSRC") 
    {
        ErrorMsg("Fatal Error: WaisSource with WaisType=", WaisType());
        abort();
    }
}
char WaisSource::browse(ObjectPtr& next, Anchor* a) 
{ 
    ErrorMsg(IDS_NOTIMPLEMENT);
    return 0;
}


/////////////////////////////////////////////////////////////////
// Class WaisText:

WaisText::WaisText(const Object& o) : WaisDocument(o) 
{
    objType_ = WaisTextType;
}

/////////////////////////////////////////////////////////////////
// Class WaisImage:

WaisImage::WaisImage(const Object& o) : WaisDocument(o) 
{
    objType_ = WaisImageType;
    if (WaisType() != "GIF")
        if (WaisType() != "TIFF")
            if (WaisType() != "PICT") 
            {
                ErrorMsg("Fatal Error: WaisImage with WaisType=", WaisType());
                abort();
            }
}

char WaisImage::browse(ObjectPtr& next, Anchor* a) 
{ 
    return 0;
//    return Browse(HG, next, a); 
}


/////////////////////////////////////////////////////////////////
// Class TelnetSession:

TelnetSession::TelnetSession(const Object& o) : RemoteDocument(o) 
{
    objType_ = TelnetSessionType;
    if (Protocol() != "telnet" &&
        Protocol() != "tn3270") 
    {
        ErrorMsg("Fatal Error: TelnetSession with Protocol=", Protocol());
        abort();
    }
}

char TelnetSession::browse(ObjectPtr& next, Anchor* a) 
{ 
#ifdef __AMADEUS__
    RString command;
    if (Protocol() == "telnet")
        command = RString((const char*)theApp.telnetCmd_) + RString(" ") + Host() + RString(" ") + Port();
    else
        command = RString((const char*)theApp.tn3270Cmd_) + RString(" ") + Host() + RString(" ") + Port();

    if (Path() != "")
    {
       char buffer[1024];
       char login[80];

//       command += RString(" \"") + Path() +RString("\"");
        strcpy(login, Path()); 
        if (_hglang->GetLangByOrder(0)=="ge")
            sprintf(buffer,"Sie mssen nun '%s' als Login angeben !", login);
        else
            sprintf(buffer,"If your are prompted for login, please enter '%s' !", login);
        if (MessageBox(GetFocus(), buffer, "Info", MB_OKCANCEL | MB_ICONINFORMATION) == IDCANCEL)
            return DO_GUICMD;
    }

    RunProcess(command, true); // run in quiet mode: fkappe 02/12/92
#endif
    return '\b'; // undo + redo last command
}


/////////////////////////////////////////////////////////////////
// Class FTPDocument:

FTPDocument::FTPDocument(const Object& o) : RemoteDocument(o) 
{
    objType_ = FTPDocumentType;
    if (Protocol() != "ftp") 
    {
        ErrorMsg("Fatal Error: FTPDocument with Protocol=", Protocol());
        abort();
    }
}

char FTPDocument::browse(ObjectPtr& next, Anchor* a) 
{ 
#ifdef __AMADEUS__
    RString command = RString((const char*)theApp.ftpCmd_) + RString(" ftp://") + Host();
    if (Path().length() && Path().index(0) != '/')
        command += RString("/");
    command += Path();

    RunProcess(command, true);
#endif
    return '\0';
}

