//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        hgobject.h
//
// Purpose:     Class HGObject (Hyper-G Object) and derived Classes
//
// Created:     18 Jan 92    Frank Kappe (hgobject.C for HGC-T)
//                           Thomas Dietinger
//
// Modified:    06 Mar 95    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>

#ifndef HG_APPS_HGOBJECT_H
#define HG_APPS_HGOBJECT_H

extern int MaxFullTextDocs;

class BaseFD;
class CSocket;
class LinkList;

#include <apps/protptr.h>    
#include <database/objectid.h>
#include "language.h"
#include "langutl.h"

#include <Dispatch/enter-scope.h>
#include <InterViews/resource.h>

enum HGObjectEnum { HGObjectType, AnchorType, 
                    UserType, GroupType, DocumentType, 
                    CollectionType, ClusterType, TextDocumentType,
                    PICDocumentType, ImageDocumentType, SoundDocumentType, 
                    MovieDocumentType, PostScriptDocumentType, SceneDocumentType,
                    GenericDocumentType, RemoteDocumentType, WWWDocumentType,
                    GopherDocumentType, GopherTextType, GopherImageType, 
                    GopherSoundType, GopherCollectionType, GopherIndexType,
                    WaisDocumentType, WaisSourceType, WaisTextType,
                    WaisImageType, TelnetSessionType, FTPDocumentType,
                    FakeObjectType };

class HgStub;
class Anchor;
class Document;
class Collection;
class ObjectList;
class ObjectNode;
class ObjectListList;
class HGObject;
class ViewerNode;
class BaseProt;
class HyperGProt;

/////////////////////////////////////////////////////////////////
// Class ObjectPtr: Pointer to HGObject (with resource management)

class ObjectPtr 
{
public:
    ObjectPtr(HGObject* o = nil);
    ObjectPtr(const ObjectPtr&);
    ~ObjectPtr();

    ObjectPtr& operator = (const ObjectPtr&);
    ObjectPtr& operator = (HGObject*);
    HGObject* Obj() const             { return _obj; }

private:
    friend inline int operator ==(const ObjectPtr&, const ObjectPtr&);
    friend inline int operator !=(const ObjectPtr&, const ObjectPtr&);
    friend inline int operator ==(const ObjectPtr&, HGObject*);
    friend inline int operator !=(const ObjectPtr&, HGObject*);
    HGObject* _obj;
};

inline int operator ==(const ObjectPtr& a, const ObjectPtr& b) {
    return a.Obj() == b.Obj();
}

inline int operator !=(const ObjectPtr& a, const ObjectPtr& b) {
    return a.Obj() != b.Obj();
}

inline int operator ==(const ObjectPtr& a, HGObject* obj) {
    return a.Obj() == obj;
}

inline int operator !=(const ObjectPtr& a, HGObject* obj) {
    return a.Obj() != obj;
}

/////////////////////////////////////////////////////////////////
// Class HGObject (Hyper-G Object); abstract:

class HGObject : public Resource
{
public:

    static ProtocolList conOrProtList_;     // list of connection oriented protocols
    static ObjectListList activeObjects_;   // list of lists of (for search) activated collections (seperated 

    friend class BaseProt;
    friend class HyperGProt;

    HGObject(const Object& o, ProtPtr pPtr = ProtPtr() );
    virtual ~HGObject() { }             // is called by unref

    virtual boolean update();           // updates object attributes to current state in database

    // ID: old version - please use IDString instead!
    ObjectID ID() const                     { return _id; }

    const RString& GOID() const             { return _goid; }       // global object id
    const RString& Type() const             { return _type; }       // object type
    const RString& Author() const           { return _author; }     // author
    const RString& Rights() const           { return _rights; }     // access rights of object
    const RString& Created() const          { return _created; }    // creation date
    const RString& Modified() const         { return _modified; }   // modification date
    const RString& Expires() const          { return _expires; }    // expiration date
    const RString& Opens() const            { return _opens; }  
    const RString& Language() const         { return _language; }
    int            Languages() const        { return _titles.count(); }
    const RString& Parent() const           { return _parent; }     // parent collection
    const RString& Serv() const             { return _serv; }       // database server
    const RString& Sequence() const         { return _sequence; }   // sequence number
    const RString& Score() const            { return _score; }      // score (after fulltext query)
    void           Score(const RString&);
    void           Score(int);
    const RString  SGMLTitle();
    const RString  Title();
    void           Title(const RString& t)  { _title = t; }
    HGObjectEnum   getType() const          { return objType_; }
    const RString& DBName() const           { return _dbName; }
    const RString& Layout() const           { return layout_; }

    virtual RString IDString() const        { return _id.IDString(); }
    virtual RString Info() const;
    virtual RString ClientAttributes() const;

    virtual boolean active() const          { return activated_; }
    virtual void activate(boolean);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual boolean Active() const          { return false; }
//    virtual void Activate(boolean)          { }

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr& next, Anchor*  dest = nil) = 0;

    virtual ObjIdArray* MyParents(HgStub*) const  { return nil; }
    virtual ObjectList* MyChildren(HgStub*)       { return nil; }
    virtual ObjectList* LinksTo(HgStub*)          { return nil; }
    virtual ObjectList* LinksFrom(HgStub*);

    virtual ObjectList* KeyQuery(HgStub*, RString&, 
                                 boolean wholeDB = false, boolean score = false) const;
    virtual ObjectList* FulltextQuery(HgStub*, RString&, boolean wholeDB = false) const;
    virtual ObjectList* FindSimilar(HgStub*) const{ return nil; }
    virtual ObjectList* FindObj(HgStub*, RString&, RString&, 
                                RString&, boolean wholeDB = false, boolean score = false) const;

    virtual boolean Erase(HgStub*)                          { return false; }
    virtual boolean Insert(HgStub*, ObjectID, ObjectID&)    { return false; }
    virtual boolean MoveTo(HgStub*, ObjectID, ObjectID)     { return false; }
    virtual boolean CopyTo(HgStub*, ObjectID)               { return false; }
    virtual boolean UnlinkFrom(HgStub*, ObjectID)           { return false; }
*/
// These are the NEW functions, please use these instead of the one above !
    virtual boolean getData(boolean forceload, 
                            RString tempName = "", 
                            RString cmdLine = "")           { return false; }
    virtual boolean putData(RString fName, 
                            const char* dataBuf = 0, 
                            int dataLen = 0)                { return false; }

    virtual char browse(ObjectPtr& next, Anchor*  dest = nil) = 0;
    virtual ObjectList* myParents(ObjectList* oldList = 0, 
                                  BaseFD* fd = 0)           { return nil; }
    virtual ObjectList* myChildren(ObjectList* oldList = 0, 
                                   BaseFD* fd = 0)          { return nil; }
    virtual ObjectList* links(ObjectList* oldList = 0, 
                              BaseFD* fd = 0)             { return nil; }
    virtual ObjectList* references(ObjectList* oldList = 0, 
                                   BaseFD* fd = 0);

    static ObjectList* keyQuery(RString&, 
                                boolean wholeDB = false, 
                                boolean score = false,
                                ObjectList* oldList = 0);

    static ObjectList* fulltextQuery(RString&, 
                                     boolean wholeDB = false,
                                     ObjectList* oldList = 0);
//    virtual ObjectList* findSimilar()               { return nil; }
    static ObjectList* findObj(RString&, 
                               RString&, 
                               RString&, 
                               boolean wholeDB = false, 
                               boolean score = false,
                               ObjectList* oldList = 0);

    // delete object completely (=unlink from all collections !)
    virtual boolean erase()                         { return false; }
    
    virtual boolean insert(ObjectPtr toColl)        { return false; }
    virtual boolean moveTo(ObjectPtr fromColl, 
                           ObjectPtr toColl)        { return false; }
    virtual boolean copyTo(ObjectPtr toColl)        { return false; }
    virtual boolean unlinkFrom(ObjectPtr fromColl)  { return false; }

    virtual boolean modifyAttribute(RString oldAttrLine,
                                    RString newAttrLine);

    virtual boolean lock();
    virtual boolean unlock();

    virtual RString user();

    virtual ObjectPtr getRecentParent();

    virtual boolean readOnly();                     // are we allowed to modify this object in the database ?

    void setProt(ProtPtr pPtr)                      { protPtr_ = pPtr; }

    ProtPtr protPtr_;
    inline BaseProt* protocol() const               { return protPtr_.protocol(); }

    static ProtPtr homeProtPtr_;                    // pointer to Hyper-G Home Server 
                                                    // for very special functions !
    static BaseProt* getHomeServer()                { return homeProtPtr_.protocol(); }

	static HGLanguages *_hglang;  

    RString parentObjKey_;                          // recent (hierarchical) parent (id/URL)

protected:

    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

    HGObjectEnum objType_;
    ObjectID _id;
    boolean activated_;

private:
    RString _goid;
    RString _type;
    RString _author;
    RString _created;
    RString _modified;
    RString _expires;
    RString _opens;
    RString _title;
    RString _parent;
    RString _rights;
    RString _sequence;
    RString _score;
    RString _serv;
    RString _dbName;

    RStringArray _titles;
    RStringArray _keywords;
    RStringArray _userattr;
    RString _language;
    RString layout_;
};


/////////////////////////////////////////////////////////////////
// Class Anchor:

class Anchor : public HGObject 
{
public:
    Anchor(const Object&);

    virtual boolean update();           // updates object attributes to current state in database

    const RString& Position() const          { return _position; }
    const RString  Dest() const               { return _dest; }
    const RString& GDest() const             { return _gdest; }
    const RString  Src() const               { return _src; } // steve
    const RString  GSrc() const               { return _gsrc; } // steve
    const RString& Hint() const              { return _hint; }
    const RString& LinkType() const          { return _linktype; }
    const boolean  IsSrc() const             { return _issrc; }
    const RString& State() const             { return _state; }
    void    State(const RString& s)   { _state = s; }

    virtual RString Info() const;
    virtual RString ClientAttributes() const;

// These are the OLD functions, please use the new protocol-independent below !
/*
    ObjectPtr Target(HgStub* db)      { GetTarget(db); return _target; }
    virtual char Browse(HgStub*, ObjectPtr& next, Anchor* dest = nil);

    virtual ObjIdArray* MyParents(HgStub*) const;

    virtual boolean Erase(HgStub*);
    virtual boolean MoveTo(HgStub*, ObjectID, ObjectID);
    virtual boolean CopyTo(HgStub*, ObjectID);
    virtual boolean UnlinkFrom(HgStub*, ObjectID);
*/

// These are the NEW functions, please use these instead of the one above !
    ObjectPtr target()                      { getTarget(); return _target; }
    virtual char browse(ObjectPtr& next, Anchor* dest = nil);

    virtual ObjectList* myParents(ObjectList* oldList = 0, 
                                  BaseFD* fd = 0) const;

    virtual boolean erase();
    virtual boolean insert(ObjectPtr srcOrDestDoc);
    virtual boolean moveTo(ObjectPtr fromColl, 
                           ObjectPtr toColl);
    virtual boolean copyTo(ObjectPtr toColl);
    virtual boolean unlinkFrom(ObjectPtr fromColl);

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

    boolean GetTarget(HgStub*);

    boolean getTarget();

    RString _position;
    RString _dest;
    RString _gdest;
    RString _src; // steve
    RString _gsrc; // steve
    RString _hint;
    RString _linktype;
    boolean _issrc;
    RString _state;    // to be passed to viewers
    ObjectPtr _target;
};


/////////////////////////////////////////////////////////////////
// Class User:

class User : public HGObject 
{
public:
    User(const Object&);

    RString uName() const       { return uName_; }
    RString home() const        { return home_; }
    RString desc() const        { return desc_; }

    virtual boolean update();           // updates object attributes to current state in database
    virtual RString Info() const;
    virtual RString ClientAttributes() const;

    virtual ObjectList* getGroups();

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

    RString uName_;                     // unique user name
    RStringArray groups_;               // list of groups the user belongs to
    RString home_;                      // home collection
    RStringArray passwds_;               // list of passwords
    RStringArray hosts_;                 // list of hosts for auto-identification
    RString desc_;                      // description
};


/////////////////////////////////////////////////////////////////
// Class Group:

class Group : public HGObject 
{
public:
    Group(const Object&);

    RString uGroup() const      { return uGroup_; }
    RString desc() const        { return desc_; }

    virtual boolean update();           // updates object attributes to current state in database
    virtual RString Info() const;
    virtual RString ClientAttributes() const;

    virtual ObjectList* getGroups();

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

    RString uGroup_;                    // unique group name
    RStringArray groups_;
    RString desc_;
};


/////////////////////////////////////////////////////////////////
// Class Document; abstract:

class Document : public HGObject 
{
public:
    Document(const Object&);
    virtual ~Document();

    virtual boolean update();           // updates object attributes to current state in database

    const RString& DocType() const           { return _doctype; }
    const RString& SubType() const           { return _subtype; }
    const RString& Name() const              { return _name; }
    const RString& Path() const              { return _path; }
    const RString& Presentation() const      { return _presentation; }
    const RString& tempName() const          { return tempName_; }
    const RString& viewerApp() const         { return viewerApp_; }
    const RString& remoteDocType() const     { return remoteDocType_; }

// for remote documents:
    const RString& Protocol() const          { return _protocol; }
    const RString& Host() const              { return _host; }
    const RString& Port() const              { return _port; }

    LinkList* linkList()                     { return linkList_; }

    virtual RString Info() const;
    virtual RString ClientAttributes() const;
    virtual RString IDString() const;

//    virtual boolean loadDocument(CSocket* mySock, RString& cmdLine, RString& tempName, boolean forceLoad);
    virtual boolean loadDocument(BaseFD* fd, RString& data);
    virtual boolean loadDocument(BaseFD* fd, RString& cmdLine, RString& tempName, boolean forceLoad);

    RString parseViewerCmd(const RString& cmd, RString tmpFName);

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr& next, Anchor* dest = nil);

    virtual boolean getData(boolean forceload, HgStub*, ObjectPtr& next, RString tempName = "", RString cmdLine = "");
    virtual ObjIdArray* MyParents(HgStub*) const;

    virtual ObjectList* LinksTo(HgStub*);
    virtual ObjectList* LinksFrom(HgStub*);

    virtual boolean Erase(HgStub*);
    virtual boolean Insert(HgStub*, ObjectID, ObjectID&);
    virtual boolean MoveTo(HgStub*, ObjectID, ObjectID);
    virtual boolean CopyTo(HgStub*, ObjectID);
    virtual boolean UnlinkFrom(HgStub*, ObjectID);

    ObjectList* SrcAnchors(HgStub*);
    ObjectList* DestAnchors(HgStub*);
*/
// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr& next, Anchor* dest = nil);

    virtual boolean getData(boolean forceload, RString tempName = "", RString cmdLine = "");
    virtual boolean putData(RString fName, const char* dataBuf = 0, int dataLen = 0);
    virtual ObjectList* myParents(ObjectList* oldList = 0, BaseFD* fd = 0);
    virtual ObjectList* links(ObjectList* oldList = 0, BaseFD* fd = 0);
    virtual ObjectList* references(ObjectList* oldList = 0, BaseFD* fd = 0);

    virtual boolean erase();
    virtual boolean insert(ObjectPtr toColl);
    virtual boolean moveTo(ObjectPtr fromColl, 
                           ObjectPtr toColl);
    virtual boolean copyTo(ObjectPtr toColl);
    virtual boolean unlinkFrom(ObjectPtr fromColl);

    virtual void reLoadAnchors();

    ObjectList* srcAnchors(ObjectList* oldList = 0, BaseFD* fd = 0);
    ObjectList* destAnchors(ObjectList* oldList = 0, BaseFD* fd = 0);

    RString getInlineID() const               { return inlineID_; }

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

//    virtual void LoadAnchors(HgStub*);
    virtual void loadAnchors();

    boolean _loaded;
    boolean _anchors_loaded;
    ObjectList* _src_anchors;
    ObjectList* _dest_anchors;
    ObjectList* _srcs_of_dest;

    ViewerNode* viewer_;
    RString tempName_;
    RString viewerApp_;

    LinkList* linkList_;


private:
    RString _doctype;
    RString _subtype;
    RString _name;
    RString _path;
    RString _presentation;
    RString inlineID_;

// for remote documents:
    RString _protocol;
    RString _host;
    RString _port;
    RString remoteDocType_;
};


/////////////////////////////////////////////////////////////////
// Class Collection

class Collection : public Document // for non-obvious reasons
{ 
public:
    Collection(const Object&, ObjectList* objList = nil);
    ~Collection();

    virtual boolean update();           // updates object attributes to current state in database

    const RString& CollType() const          { return _colltype; }
    const RString& SortOrder() const         { return _sortorder; }
    const RString& License() const           { return _license; }
    const RString& SubDocs() const           { return _subdocs; }
    const RString  Description() const;

    virtual RString Info() const;
    virtual RString ClientAttributes() const;

//    virtual void activate(boolean);

//    virtual boolean Active() const;
//    virtual void Activate(boolean);

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil) { return '\0'; }

    virtual ObjIdArray* MyParents(HgStub*) const;

    virtual ObjectList* MyChildren(HgStub*);
    ObjectList* Members(HgStub*);       // returns stored list of members or gets the first one

    virtual boolean Erase(HgStub*);
    virtual boolean Insert(HgStub*, ObjectID, ObjectID&);
    virtual boolean MoveTo(HgStub*, ObjectID, ObjectID);
    virtual boolean CopyTo(HgStub*, ObjectID);
    virtual boolean UnlinkFrom(HgStub*, ObjectID);
*/
// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil)     { return '\0'; }

    virtual ObjectList* myParents(ObjectList* oldList = 0, BaseFD* fd = 0);
    virtual ObjectList* myChildren(ObjectList* oldList = 0, BaseFD* fd = 0);
    ObjectList* members(ObjectList* oldList = 0, BaseFD* fd = 0);   // returns stored list of members or gets the first one
    virtual void freeMembers();             // cleans up list of members

    virtual boolean erase();
    virtual boolean insert(ObjectPtr toColl);
    virtual boolean moveTo(ObjectPtr fromColl, 
                           ObjectPtr toColl);
    virtual boolean copyTo(ObjectPtr toColl);
    virtual boolean unlinkFrom(ObjectPtr fromColl);

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

    ObjectList* _members;

private:
    RString _colltype;
    RString _sortorder;
    RString _subdocs;
    RString _license;

    RStringArray _descriptions;
};

/////////////////////////////////////////////////////////////////
// Class Cluster

class Cluster : public Collection 
{
public:
    Cluster(const Object&, ObjectList* objList = nil);
    ~Cluster();

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);

    // returns only visible children (chosen by currently selected language settings)
    virtual boolean getVisibleChildren(ObjectList& visObjList, boolean recursive = true);

private:
    static boolean isInCluster_;
    boolean isInClusterSet_;
};

/////////////////////////////////////////////////////////////////
// Class TextDocument:
class TextViewer;
class TextDocument : public Document 
{
public:
    TextDocument(const Object&);
    ~TextDocument();

    //  virtual RString ClientAttributes() const;

// These are the OLD functions, please use the new protocol-independent below !
//    virtual boolean getData(boolean forceload, HgStub*, ObjectPtr& next, RString tempName = "", RString cmdLine = "");

//    virtual char Browse(HgStub*, ObjectPtr& next, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual boolean getData(boolean forceload, RString tempName = "", RString cmdLine = "");

    virtual char browse(ObjectPtr& next, Anchor* dest = nil);

    RString& text() { return text_; }

protected:
//    virtual void LoadAnchors(HgStub*);
    virtual void loadAnchors();

    RString text_;

private:
    TextViewer* _viewer;
};

/////////////////////////////////////////////////////////////////
// Class PICDocument:

class PICDocument : public Document 
{
public:
    PICDocument(const Object&);
};
  
  
/////////////////////////////////////////////////////////////////
// Class ImageDocument:

class ImageDocument : public Document 
{
public:
    ImageDocument(const Object&);
};
  
  
/////////////////////////////////////////////////////////////////
// Class SoundDocument:

class SoundDocument : public Document 
{
public:
    SoundDocument(const Object&);
};
  
  
/////////////////////////////////////////////////////////////////
// Class MovieDocument:

class MovieDocument : public Document 
{
public:
    MovieDocument(const Object&);
};
  
  
/////////////////////////////////////////////////////////////////
// Class PostScriptDocument:

class PostScriptDocument : public Document 
{
public:
    PostScriptDocument(const Object&);
};
  
  
/////////////////////////////////////////////////////////////////
// Class SceneDocument (3D Scene Descriptions)

class SceneDocument : public Document 
{
public:
    SceneDocument(const Object&);
};
  
  
/////////////////////////////////////////////////////////////////
// Class GenericDocument (User-definable)

class GenericDocument : public Document 
{
public:
    GenericDocument(const Object&);

    virtual boolean update();           // updates object attributes to current state in database

//  const RString& SubType() const           { return _subtype; }
    const RString& Arguments() const         { return _arguments; }

    virtual RString Info() const;
    virtual RString ClientAttributes() const;

    boolean loadDocument(BaseFD* fd, RString& cmdLine, RString& tempName, boolean forceLoad);

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

private:
//  RString _subtype;
    RString _arguments;
};
  
  
/////////////////////////////////////////////////////////////////
// Class RemoteDocument; abstract

class RemoteDocument : public Document 
{
public:
    RemoteDocument(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual ObjectList* MyChildren(HgStub*)             { return nil; }
//    virtual ObjectList* LinksTo(HgStub*)                { return nil; }

// These are the NEW functions, please use these instead of the one above !
    virtual ObjectList* myChildren(ObjectList* oldList = 0, 
                                   BaseFD* fd = 0)      { return nil; }
    virtual ObjectList* links(ObjectList* oldList = 0, 
                              BaseFD* fd = 0)         { return nil; }
};

/////////////////////////////////////////////////////////////////
// Class WWWDocument:
class WWWViewer;
class WWWDocument : public RemoteDocument 
{
public:
    WWWDocument(const Object&);
    ~WWWDocument();

    virtual boolean update();           // updates object attributes to current state in database

    const RString& WWWType() const        { return _wwwtype; }

    virtual RString Info() const;

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr& next, Anchor* dest = nil);

    virtual boolean getData(boolean forceload, HgStub*, ObjectPtr& next, RString tempName = "", RString cmdLine = "");

    virtual ObjectList* LinksTo(HgStub*);
*/
// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr& next, Anchor* dest = nil);

    virtual boolean getData(boolean forceload, RString tempName = "", RString cmdLine = "");

    virtual ObjectList* links(ObjectList* oldList = 0, BaseFD* fd = 0);

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

private:
    RString _wwwtype;
    WWWViewer* _viewer;
};

/////////////////////////////////////////////////////////////////
// Class GopherDocument; abstract
class GopherViewer;
class GopherDocument : public RemoteDocument 
{
public:
    GopherDocument(const Object&);
    ~GopherDocument();

    virtual boolean update();           // updates object attributes to current state in database

    const RString& GopherType() const        { return _gophertype; }

    virtual RString Info() const;
    virtual RString ClientAttributes() const;

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr& next, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr& next, Anchor* dest = nil);

protected:
    ObjectList* Gophers(HgStub*, const RString&) const;
    
    ObjectList* gophers(const RString&) const;
    GopherViewer* _viewer;

    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

private:
    RString _gophertype;
};

/////////////////////////////////////////////////////////////////
// Class GopherText:

class GopherText : public GopherDocument 
{
public:
    GopherText(const Object&);
};

/////////////////////////////////////////////////////////////////
// Class GopherImage:

class GopherImage : public GopherDocument 
{
public:
    GopherImage(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);

};

/////////////////////////////////////////////////////////////////
// Class GopherSound:

class GopherSound : public GopherDocument 
{
public:
    GopherSound(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);

};

/////////////////////////////////////////////////////////////////
// Class GopherCollection:

class GopherCollection : public GopherDocument 
{
public:
    GopherCollection(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil) { return '\0'; }
    virtual ObjectList* MyChildren(HgStub*);
*/
// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil)          { return '\0'; }
    virtual ObjectList* myChildren(ObjectList* oldList = 0, BaseFD* fd = 0);

};

/////////////////////////////////////////////////////////////////
// Class GopherIndex:


class GopherIndex : public GopherDocument 
{
public:
    GopherIndex(const Object&);

//    virtual void activate(boolean);

//    virtual boolean Active() const;
//    virtual void Activate(boolean);

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil) { return '\0'; }
    virtual ObjectList* KeyQuery(HgStub*, RString&, boolean wholeDB = false, boolean score = false) const;
*/
// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil)          { return '\0'; }
/*
    virtual ObjectList* keyQuery(RString&, 
                                 boolean wholeDB = false, 
                                 boolean score = false,
                                 ObjectList* oldList = 0, BaseFD* fd = 0) const;
*/
};

/////////////////////////////////////////////////////////////////
// Class WaisDocument; abstract
class WaisViewer;
class WaisDocument : public RemoteDocument 
{
public:
    WaisDocument(const Object&);

    virtual boolean update();           // updates object attributes to current state in database

    const RString& Size() const { return _size; }
    const RString& WaisType() const { return _waistype; }
    const RString& DocId() const { return _docid; }  
    void DocId (const RString& id) { _docid = id; }

    virtual RString IDString() const { return _docid; }
    virtual RString Info() const;
    virtual RString ClientAttributes() const;

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);
    virtual ObjectList* FindSimilar(HgStub*) const;
*/
// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);
//    virtual ObjectList* findSimilar() const;

protected:
    // update object attributes by object information string
    virtual boolean updateAttributes(Object o);

    WaisViewer* _viewer;

private:
    RString _size;
    RString _waistype;
    RString _docid;

};

/////////////////////////////////////////////////////////////////
// Class WaisSource:

class WaisSource : public WaisDocument 
{
public:
    WaisSource(const Object&);

    virtual RString IDString() const { return RemoteDocument::IDString(); }

//    virtual void activate(boolean);

//    virtual boolean Active() const;
//    virtual void Activate(boolean);

// These are the OLD functions, please use the new protocol-independent below !
/*
    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);
    virtual ObjectList* KeyQuery(HgStub*, RString&, boolean wholeDB = false, boolean score = false) const;
*/
// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);
/*
    virtual ObjectList* keyQuery(RString&, 
                                 boolean wholeDB = false, 
                                 boolean score = false,
                                 ObjectList* oldList = 0, BaseFD* fd = 0) const;
*/
};

/////////////////////////////////////////////////////////////////
// Class WaisText:

class WaisText : public WaisDocument 
{
public:
    WaisText(const Object&);
};

/////////////////////////////////////////////////////////////////
// Class WaisImage:

class WaisImage : public WaisDocument 
{
public:
    WaisImage(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);
};

/////////////////////////////////////////////////////////////////
// Class TelnetSession:

class TelnetSession : public RemoteDocument 
{
public:
    TelnetSession(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);
};

/////////////////////////////////////////////////////////////////
// Class FTPDocument:

class FTPDocument : public RemoteDocument 
{
public:
    FTPDocument(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil);

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil);
};

/////////////////////////////////////////////////////////////////
// warning, kludge ahead:
//
// Class FakeObject (e.g. for displaying Command List with same routine
// as ordinary Hyper-G Objects)

class FakeObject : public HGObject 
{ 
public:
    FakeObject(const Object&);

// These are the OLD functions, please use the new protocol-independent below !
//    virtual char Browse(HgStub*, ObjectPtr&, Anchor* dest = nil) { return '\0'; }

// These are the NEW functions, please use these instead of the one above !
    virtual char browse(ObjectPtr&, Anchor* dest = nil) { return '\0'; }
};

#endif
