//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        baseprot.cpp
//
// Purpose:     class HyperGProt (Hyper-G protocol class) 
//
// Created:     10 Oct 94    Thomas Dietinger
//
// Modified:    18 Aug 95    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>


#include <stdafx.h>

#ifndef _CONSOLE
#include <hgapp.h>
#include <mainfrm.h>
#endif

#include <resource.h>

#include <apps/conf.h>
#include <apps/hgobject.h>
#include "baseprot.h"
#include "hgprot.h"
#include <apps/objectlist.h>

#include <Dispatch/rpcbuf.h>
#include <Dispatch/rpcservice.h>
#include <apps/bin/servconn.h>      // for the Hyper-G protocol stack
#include <Dispatch/dispatcher.h>    // for the Hyper-G protocol stack
#include <hyperg/hgheader.h>

#include <apps/utils.h>
#include <utils/environ.h>

#include <stdlib.h>

#if defined(_MSC_VER) & defined(_DEBUG)
#define new DEBUG_NEW
#endif


const RString hgProtStr = "hyperg";
const RString defaultPath = "~anonymous";


//////////////////////////////////////////////////////////////////////////////
// Class DocumentService: Helper class for pipeDocument/pipeRemote

class DocumentService : public RpcService
{
public:
    DocumentService() : fd_(INVALID_SOCKET), RpcService(0) { };
    virtual void createReader(SOCKET fd);
    SOCKET getSockId() const { return fd_; }
    int port() const { return _port; }

private:
    SOCKET fd_;
};


void DocumentService::createReader(SOCKET fd) 
{ 
    fd_ = fd; 
    TRACE("DocumentService::createReader, socket assigned (%d)\n", fd_);
    ASSERT(fd_ != INVALID_SOCKET);

}



HGDataFD::~HGDataFD()
{
    if (fdData_)
    {
        fdData_->linger();
        fdData_->Close();
        delete fdData_;
        fdData_ = 0;
    }
}

/////////////////////////////////////////////////////////////////
// class HyperGProt: Hyper-G protocol class

HyperGProt::HyperGProt(RString connectStr, boolean connect) : BaseProt(connectStr, connect)
{
    // specify features this protocol has !
    features_ = identFunc     | 
                readFunc      |
                writeFunc     |
                moveFunc      |
                copyFunc      |
                eraseFunc     |
                unlinkFunc    |
                keyQueryFunc  |
                fTQueryFunc   |
                childrenFunc  |
                parentsFunc   |
                linksToFunc   |
                linksFromFunc |
                lockFunc      |
                unlockFunc    |
                chgPasswdFunc;

    hgConnPtr_ = 0;
    binary_      = TRUE;
    distributed_ = TRUE;
    compressed_  = TRUE;

    directDC_ = TRUE;

    if (protType_ != "hyperg" && protType_.length())
    {
        setErrorCode(IDS_PROTERRWRNGPROT);  // wrong protocol !
        return;
    }

    if (!portNr_)
        portNr_ = 418;                      // default Hyper-G protocol port number

    // hyperg://{<user>{:<password>}@}<host>{:<port>{;<binary>{;<distributed>{;<compressed>}}}}/<path>    
    int pos = connectStr_.index('@');
    if (pos == -1)
        pos = connectStr_.index(':');   // no user etc. specified, skip first ':' 

    RString param;
    pos = connectStr_.index(pos, ':');  // search for port specification
    int pathPos = connectStr_.index(pos, '/');
    if (pos != -1)
    {
        pos = connectStr_.index(pos, ';');      // binary ?
        if (pos != -1 && pos < pathPos && connectStr_.gWord(pos, ":/", param))
        {
            binary_ = atoi(param);
            if (pos != -1 && pos < pathPos && connectStr_.gWord(pos, ":/", param))
            {
                distributed_ = atoi(param);     // distributed ?
                if (pos != -1 && pos < pathPos && connectStr_.gWord(pos, ":/", param))
                    compressed_ = atoi(param);  // compressed ?
            }
        }
    }

    TRACE("baseprot.cpp: HyperGProt::HyperGProt: trying to connect to %s with port %d and user %s\n", host_.string(), portNr_, user_.string());

    // connect if possible !
    hgConnPtr_ = new ServerConnection(user_, passWord_, host_, portNr_, binary_, distributed_, compressed_);
    if (!hgConnPtr_)
    {
        setErrorCode(IDS_ERRNOTCONNECT);  // could not connect !
        return;
    }

    if (hgConnPtr_->error()) 
    {
        setErrorCode(IDS_PROTERRHGERR);  // error occured !
        errStr_ += HGError(hgConnPtr_->error());
        return;
    }

    identify(user_, passWord_, false);

    if (!homePath_.length())
        homePath_ = defaultPath;

    // set subType if any !
}

HyperGProt::~HyperGProt()
{
    // disconnect !
    if (hgConnPtr_)
    {
        delete hgConnPtr_;

        // Attention: ONLY works if connected to ONE Hyper-G server !!!!!
        delete &Dispatcher::instance(); // delete dispatcher !
    }
}

// get object information by destination ID/URL
Object HyperGProt::getObjectInfo(RString dest)
{
    Object obj;
    int pos = dest.index(':');      // test for protocol specifier (eg. "http:", "gopher:", ..)
    if (pos != -1 && dest.gSubstrIndex(0, pos - 1) != "hyperg")
    {
        // remote-link !!
        RString protocol = dest.GetSubstrIndex(0,pos-1);
        RString host = dest.GetSubstrDelim(pos+3,'/');
        pos += 2 + host.length();
        RString path = dest.GetSubstrDelim(pos+1,'#'); // no dest anchor!!

        RString port = host.GetRight(host.index(':')+1);
        if (port == host)   // no port number specified!
        {
            if (protocol == "gopher")
                port = DEFAULT_GOPHER_PORT;
            else
            if (protocol == "http")
                port = DEFAULT_WWW_PORT;
            else
                port = "";
        }
        host = host.GetSubstrDelim(0,':');

        obj  = "ObjectID=0x00000000\nType=Document\nDocumentType=Remote\n";
        obj += "Author=system\nProtocol=" + protocol + "\nHost=" + host + 
               "\nPort=" + port + "\nPath=" + path + "\nTitle=??:" + path + 
               "\nTimeCreated=" + TimeNow() +"\n";

        if (protocol == "gopher")   // especially for gophers: GopherType !
        {
            if (path.gRight(1) == "/")
                path = path.gRight(1);      // ignore leading '/'
            obj += "GopherType=" + path.gSubstrIndex(0, 0) + "\n";
        }
    }
    else
    {
        if (!hgConnPtr_)
        {
            ErrorMsg("Not connected to a Hyper-G database !");
            return obj;
        }

        if (pos != -1)
        {
/*
            RString protocol;
            RString user;
            RString passwd;
            RString host;
            int portNr;
            RString path;
            splitURL(dest, protocol, user, passwd, host, portNr, path);
*/
            // URL of form 'hyperg:0x????????_0x????????'
            if (dest.index('/') == -1)
            {
                dest = dest.gSubstrDelim(pos + 1, ';');
            }
            else
            {
                RString host = dest.gSubstrDelim(pos + 3, '/');
                pos += 2 + host.length();
                dest = dest.gSubstrDelim(pos + 2, ';');     // store path (beginning AFTER '/')
                host = host.gSubstrDelim(0,':');

                // make GOID out of host name and collection name/object id
                RString servAddr;
                RString hostAddr;

                if (!Environ::hostAddress(host, servAddr))
                    return "";
                
                if (!Environ::hostAddress(host_, hostAddr))
                    return "";

                if (!dest.length())
//                    dest = "~anonymous";    //"rootcollection";
                    dest = homePath_;

                if (hostAddr != servAddr)   // foreign server !
                {
                    ObjectID ooid((long)ntohl(inet_addr( servAddr.string())));
                    RString goid = ooid.IDString();

                    // dest holds now the collection name or an object id
                    if (dest.GetSubstrIndex(0,1) == "0x")   // passed ID
                    {
                        dest = ooid.IDString() + " " + dest;
                    }
                    else                                    // passed collection name
                    {
                        // find out oid of server
                        goid += " 0x00000000";
                        ooid = hgConnPtr_->mapId(goid);
                        hgConnPtr_->GetObject(ooid, obj);

                        ObjIdArray olist, activeColls;
                        activeColls.insert(ooid);
                        RString query = "Name=" + dest;
                        if (hgConnPtr_->idsByQueryColl(query, activeColls, olist))
                        {
                            if (olist.count() &&
                                hgConnPtr_->GetObject(olist[0], obj))
                                return obj;
                        }

                        return "";  // error !
                    }
                }
//                dest = buf + dest;
            }
        }
        
        boolean success;

        if (dest.GetSubstrIndex(0,1) == "0x")   // passed ID
        {
            int ndx = 0;
            RString word2;
            ObjectID helpid;
            dest.gWord(ndx, " _", word2);  // set ndx after the first word
            if (dest.gWord(ndx, " _", word2))
            {
                if (word2.GetSubstrIndex(0,1) == "0x")
                {
                    RString tmp = dest;
                    tmp.subst('_', ' '); // in the url _ seperates the two parts of the GOID
                    helpid = hgConnPtr_->mapId(tmp);
                }
                else // grumml because of sorry not implemented with ATNT
                    helpid = dest;
            }
            else 
                helpid = dest;

            success = hgConnPtr_->GetObject(helpid, obj);
        }
        else                                    // passed collection name
        {
            ObjectID oId;
            oId = hgConnPtr_->GetFirstCollID(dest);
            if (!oId.id())
                success = false;    // didn't find collection !
            else
                success = hgConnPtr_->GetObject(oId, obj);
        }
    }

    return obj;
} // getObjectInfo

//
// evaluate whether we can use the Hyper-G protocol with this URL
//
boolean HyperGProt::useConnection(RString url)
{
    // <protocol>
    RString protocol = "hyperg";    // defaults to 'hyperg' of no URL specified !
    int pos = url.index(':');
    if (pos > 0)
        protocol = url.gSubstrIndex(0, pos - 1);
//    else
//        protocol = url;

    // protocols currently supported by Hyper-G: hyperg, http, gopher
    if (protocol == "hyperg" ||
        protocol == "http" ||
        protocol == "gopher")
        return true;
    else
        return false;
}


boolean HyperGProt::identify(const RString& name, 
                             const RString& passwd, 
                             boolean encrypt)
{ 
    if (!passwd.length())
        hgConnPtr_->Identify(name, "NONE", false);
    else
        hgConnPtr_->Identify(name, passwd, encrypt);

    RString newUser = hgConnPtr_->user().gSubstrDelim(0, '(');  // till end or "(system)", if super-user !
    if (newUser != name)
        return false; 

    user_ = name;
    passWord_ = passwd;

    hgConnPtr_->GetObject(hgConnPtr_->userId(), userObj_);

    homePath_ = userObj_.field("Home=");
    if (homePath_ == RString()) // no home specified or anonymous; use default
        homePath_ = defaultPath;
    else 
    {
        ObjectID collid;
        if (homePath_.GetSubstrIndex(0,1) == "0x")  // passed ID
            collid = (const char*)homePath_;
        else                                        // passed name
            collid = hgConnPtr_->GetFirstCollID(homePath_);

        if (collid == 0 || !hgConnPtr_->IsCollection(collid))
            homePath_ = defaultPath;                // home specified but not existing
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////
//
// encrypts a password using random salt:
//
RString HyperGProt::enCrypt(const RString& clear) 
{
#ifdef LOCAL_DB_20
  return RString(clear);
#else
  static const char *saltletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789./";

  srand((int)::time(nil));

  char salt[3];
  salt[0] = saltletters[(int)rand()%strlen(saltletters)];
  salt[1] = saltletters[(int)rand()%strlen(saltletters)];
  salt[2] = '\0';

  const char* tmp = ::crypt((const char*)clear, salt);
  return RString(tmp);
#endif
}

//////////////////////////////////////////////////////////////////////////////
//
// checks whether the clear text version corresponds to the encrypted version
//
boolean HyperGProt::validatePassword(const RString& clear, const RString& encr) 
{
#ifdef LOCAL_DB_20
  return true; // local DB dosn't support identification
#else
    const char* tmp = ::crypt((const char*)clear, (char*)(const char*)encr);
    return (strcmp(tmp, (const char*) encr) == 0);
#endif
}

boolean HyperGProt::changePassword(const RString& oldPassword, 
                                   const RString& newPassword)
{
    RString newpw = enCrypt(newpw);
    RStringArray pwds;

    Object userObj;
    hgConnPtr_->GetObject(hgConnPtr_->userId(), userObj);
    userObj.fields("Passwd=", pwds);

    for (int i=0; i < pwds.count(); i++) 
    {
        if (validatePassword(oldPassword, pwds[i]))     // this is the pwd we want to change
        { 
            RString command = "rem Passwd=" + pwds[i] + "\\add Passwd=" + newpw;
            return hgConnPtr_->ChangeObject(hgConnPtr_->userId(), command);
        }
    }
    return false;         // old password was wrong
}


BaseFD* HyperGProt::open(ObjectPtr objPtr,              // pointer to object of interest
                         OpenMode mode,                 // opening mode
                         ObjectList* objList,           // pointer to additional list (e.g. list of activated collections)
                         RString helpStr)               // additional string (e.g. query string)
{
    if (!hgConnPtr_)
        return nil;

    BaseFD* fd;
    if (mode == getDataMode ||
        mode == putDataMode)
    {
        fd = new HGDataFD;
    }
    else
    if (mode == fullTextQueryMode)
    {
        fd = new HGFTListFD;
        ((HGFTListFD*)fd)->helpStr_ = helpStr;
    }
    else
    {
        fd = new HGListFD;
        ((HGListFD*)fd)->helpStr_ = helpStr;
    }

    fd->mode_ = mode;
    fd->Type_ = "hyperg";
    fd->objPtr_ = objPtr;

    boolean okFlag = false;

    ObjIdArray* pObjIdList = 0;
    if (objList && objList->count())
    {
        pObjIdList = new ObjIdArray(objList->count());
        for (ObjectNode* oNode = objList->getFirst(); 
             oNode; 
             oNode = objList->getNext(oNode))
        {
            if (oNode->ObjPtr().Obj()->ID() > 0)
                pObjIdList->insert(oNode->ObjPtr().Obj()->ID());
        }
    }

    switch (mode)
    {
        case getDataMode :
            {
                sockaddr_in sain ;
                int len = sizeof (sockaddr_in) ;

                RString hostAddr;
/*
                RString hostName = Environ::hostName();
                if (!hostName.length() || !Environ::hostAddress( hostName, hostAddr))
                {
                    ErrorMsg(IDS_CANTGETHOSTADDR);
                    delete fd;
                    return 0;
                }
                TRACE("hgprot.cpp:HyperGProt::open hostname:'%s', hostaddr:'%s'\n", hostName.string(), hostAddr.string());
*/

                if (::getsockname (hgConnPtr_->server().rdbuf()->getSockId(), (sockaddr*)&sain, &len) >= 0)
                {
                    char* inaddr;
                    if (!(inaddr = inet_ntoa(sain.sin_addr))) 
                    {
                        ErrorMsg(IDS_CANTGETHOSTADDR);
                        delete fd;
                        return 0;
                    }
                    else 
                        hostAddr = inaddr;
                }

                DocumentService service;
                boolean dcConnectType = true;
                int dcPort = -1;

                CSocket* dcConnection = 0;
                if (directDC_)
                {
                    dcConnection = new CSocket(host_, 4712);
                    if (dcConnection->getSockId() != SOCKET_ERROR)
                    {
                        len = sizeof (sockaddr_in) ;
                        if (::getsockname (dcConnection->getSockId(), (sockaddr*)&sain, &len) >= 0)
                        {
                            dcPort = sain.sin_port;

                            char buf[2];
                            buf[0] = 0;
                            dcConnection->Write(buf, 1);
                            while (dcConnection->Read(buf, 1) > 0 && buf[0]);
                        }
                    }
                    else
                        directDC_ = false;
                }

                if (dcPort == -1)
                {
                    dcPort = service.port();
                    dcConnectType = false;
                }
                else
                    dcPort = ntohs(dcPort);

                boolean pipeResult = false;

                Object o(objPtr.Obj()->Info());
                if (objPtr.Obj()->Type() == "Document" &&
                    ((Document*)objPtr.Obj())->DocType() == "Remote")    // remote documents
                    pipeResult = hgConnPtr_->pipeRemote(o, hostAddr, dcPort, 1);
                else
                    pipeResult = hgConnPtr_->pipeDocument(hostAddr, dcPort, objPtr.Obj()->ID(), 1);

                if (!pipeResult)
                {
//                    HGError(*((HgStub*)hyperG));
                    if (hgConnPtr_->error() == USR_BREAK)
                    {
                        InfoMsg(IDS_MSGOPINTERRUPT);
                    }
                    else
                    {
                        RString errStr = "Error: " + HGError(hgConnPtr_->error());
                        ErrorMsg(errStr);
                    }
                    delete fd;
                    return 0;
                }

                if (dcConnectType)
                    ((HGDataFD*)fd)->fdData_ = dcConnection;
                else
                {
                    ((HGDataFD*)fd)->fdData_ = new CSocket(service.getSockId());
                    delete dcConnection;
                }

                int retryCount = 3;
                while (((HGDataFD*)fd)->fdData_->getSockId() == INVALID_SOCKET && retryCount-- > 0)
                {
                    long pollTime = 0L;
                    Dispatcher::instance().dispatch(pollTime, pollTime);  // just poll !
                }

                if (((HGDataFD*)fd)->fdData_->getSockId() == INVALID_SOCKET)
                {
                    delete fd;
                    return 0;
                }

                okFlag = true;
            }
            break;
        case putDataMode :
            break;

        case getLinksMode :
            if (fd->objPtr_.Obj()->ID().id() >= 0)
                okFlag = hgConnPtr_->anchors(fd->objPtr_.Obj()->ID(), 
                                              ((HGListFD*)fd)->fdData_);
            if (fd)
                fd->dataLen_ = ((HGListFD*)fd)->fdData_.count();
            break;
        case getReferencesMode :
            if (fd->objPtr_.Obj()->ID().id() >= 0)
                okFlag = hgConnPtr_->srcsOfDst(fd->objPtr_.Obj()->ID(), 
                                              ((HGListFD*)fd)->fdData_);
            if (fd)
                fd->dataLen_ = ((HGListFD*)fd)->fdData_.count();
            break;
        case getChildrenMode :
            okFlag = hgConnPtr_->children(fd->objPtr_.Obj()->ID(), 
                                          ((HGListFD*)fd)->fdData_);
            if (fd)
                fd->dataLen_ = ((HGListFD*)fd)->fdData_.count();
            break;
        case getParentsMode :
            okFlag = hgConnPtr_->HgStub::parents(fd->objPtr_.Obj()->ID(), 
                                          ((HGListFD*)fd)->fdData_);
            if (fd)
                fd->dataLen_ = ((HGListFD*)fd)->fdData_.count();
            break;
        case keyQueryMode :
            if (pObjIdList)
                okFlag = hgConnPtr_->idsByQueryColl(helpStr, *pObjIdList, 
                                                    ((HGListFD*)fd)->fdData_);
            else
                okFlag = hgConnPtr_->idsByQuery(helpStr, ((HGListFD*)fd)->fdData_);

            if (fd)
                fd->dataLen_ = ((HGListFD*)fd)->fdData_.count();
            break;
        case fullTextQueryMode :
            if (pObjIdList)
                okFlag = hgConnPtr_->GetObjByFtQueryColl(helpStr, *pObjIdList,
                                                         ((HGFTListFD*)fd)->fdData_);
            else
                okFlag = hgConnPtr_->GetObjByFtQuery(helpStr, ((HGFTListFD*)fd)->fdData_);

            if (fd)
                fd->dataLen_ = ((HGFTListFD*)fd)->fdData_.count();
            break;
    }

    if (!okFlag)
    {
        ASSERT(0);
        delete fd;
        fd = 0;
    }

    delete pObjIdList;
    
    return fd;
}

boolean HyperGProt::close(BaseFD* fd)
{
    //
    // close fd .. or do something appropriate
    //

    if (fd)
        delete fd;

    return false;
}


long HyperGProt::getData(BaseFD* fd, 
                         char* dataBuf, 
                         long buflen)
{
    if (!fd || fd->mode_ != getDataMode)
        return -1;

    int bytesRead = -1;

    if (!fd->pos_)  // have to look for header!
    {
        int totalBytes = 0;
        int headerLen = 0; 

        bytesRead = ((HGDataFD*)fd)->fdData_->Read(dataBuf, buflen);     // get header or first <buflen> bytes to evaluate file type
        if (bytesRead == -1 || !bytesRead)
        {
            if (bytesRead == SOCKET_ERROR)
                ErrorMsg("Socket error:", WSAGetLastError());
            ErrorMsg(IDS_ERRNOSOCKREAD);
            return -1;
        }

        HgHeader fileHeader;
        int headerRet = fileHeader.read(dataBuf, bytesRead);
        int bytesRead2 = 0;
        if (headerRet != -1 && headerRet == bytesRead)
        {
            headerLen = headerRet;
            bytesRead = ((HGDataFD*)fd)->fdData_->Read(dataBuf, buflen);
            ASSERT(bytesRead>0);
        }
        else
        if (headerRet != -1 && headerRet < bytesRead)
        {
            memmove(dataBuf, dataBuf + headerRet, bytesRead - headerRet);
            bytesRead2 = ((HGDataFD*)fd)->fdData_->Read(dataBuf+bytesRead-headerRet, buflen-bytesRead + headerRet);
            if (bytesRead2 != -1)
                bytesRead += bytesRead2-headerRet;
            else
            {
                if (bytesRead2 == SOCKET_ERROR)
                    ErrorMsg("Socket error:", WSAGetLastError());
                ErrorMsg(IDS_ERRNOSOCKREAD);
                return -1;
            }
        }

        if (headerRet)
            fd->dataLen_ = fileHeader.size();
    }
    else
    {
        bytesRead = ((HGDataFD*)fd)->fdData_->Read(dataBuf, buflen);
    }

    fd->pos_ += bytesRead;

    return bytesRead;
}

long HyperGProt::putData(BaseFD* fd, 
                         char* dataBuf, 
                         long buflen)
{
    return -1;
}

//
// Attention ! The resulting 'ObjectList' has to be freed by the caller !
//
ObjectList* HyperGProt::getList(BaseFD* fd, 
                                long count,
                                ObjectList* oldList)
{
    if (!hgConnPtr_ || fd->Type_ != "hyperg")
        return nil;

    ObjectList* result = oldList;
    if (!result)
        result = new ObjectList();

    switch (fd->mode_)
    {
        case getDataMode :
            break;
        case putDataMode :
            break;

        case getChildrenMode :
        case getParentsMode :
        case getLinksMode :
        case getReferencesMode :
            receiveObjects((HGListFD*)fd, result, false, count);
            break;
        case keyQueryMode :
            receiveObjects((HGListFD*)fd, result, true, count);
            break;
        case fullTextQueryMode :
            receiveObjects((HGFTListFD*)fd, result, false, count);
            break;
    }

    return result;
}


//
// This will receive a number of Objects based on their IDs and create an
// ObjectList holding them. An additional query can be specified.
//
void HyperGProt::receiveObjects(HGListFD* fd,
                                ObjectList* result,
                                boolean parents,
                                long count) const
{
    if (!hgConnPtr_ || !result || !fd || 
        !fd->fdData_.count() || fd->pos_ < 0)
    {
//        ASSERT(0);  // not initialized yet or strange parameters !!
        return;
    }

    if (count < 0)
    {
        count = fd->fdData_.count();
        fd->pos_ = 0;
    }

    if (fd->fdData_.count() < count)
        count = fd->fdData_.count();

//    result->free();

    ObjIdArray part(fd->fdData_, fd->pos_, count);
    ObjectField ofield;
    if (!hgConnPtr_->objectByIdQuery(part, "", ofield, parents)) 
    {
        ASSERT(0);  // error occured !
        return;
    }

    for (int i = 0; i < part.count(); i++) 
    {
        Object o = ofield[i];
        o += "\nScore=1000";
        HGObject* obj = makeObject(o, ProtPtr((BaseProt*)this));
        if (obj)
        {
            if (fd->mode_ == getChildrenMode)
//                obj->parentObj_ = ObjectPtr(fd->objPtr_);    // this is the parent of all children !
                obj->parentObjKey_ = fd->objPtr_.Obj()->IDString();    // this is the parent of all children !
            ObjectPtr oPtr(obj);
            result->addTail(new ObjectNode(oPtr));
        }
    }

    fd->pos_ += result->count();
}

// 
// same as above except that a list of previously found fulltext document IDs
// is the base. Thus the resulting ObjectList contains scores.
// 
void HyperGProt::receiveObjects(HGFTListFD* fd,
                                ObjectList* result,
                                boolean parents,
                                long count) const
{
    if (!hgConnPtr_ || !result || !fd || 
        !fd->fdData_.count() || fd->pos_ < 0)
    {
//        ASSERT(0);  // not initialized yet or strange parameters !!
        return;
    }

    if (count < 0)
    {
        count = fd->fdData_.count();
        fd->pos_ = 0;
    }

    if (fd->fdData_.count() < count)
        count = fd->fdData_.count();

//    result->free();
    FtDocTreap dt ;
    float highest = 0.0f;

    for (const DocListNode* n = fd->fdData_.getFirst()  ;  n  ;  n = fd->fdData_.getNext(n)) 
    {
        dt.insert (n->document()->id(), n->document()->features()) ;
        if (n->document()->features().weight() > highest)
            highest = n->document()->features().weight() ;
    }

    // get the objects in parts from the database
    ObjIdArray part(fd->fdData_.idArray(), fd->pos_, count);
    ObjectField ofield;
    if (!hgConnPtr_->objectByIdQuery(part, "" /*fd->helpStr_*/, ofield, parents)) 
    {
        ASSERT(0);  // error occured !
        return;
    }

    for (int i = 0; i < part.count(); i++) 
    {
        Object o = ofield[i];
        // merge with the scores and positions from the treap
        FtDocTreapData f ; 
        if (dt.remove (o.ID(), f)) 
        {
            int score = int (f.weight() * 1000 / highest) ; 
            char scorestr[10];
            sprintf (scorestr, "%d", score) ;
            o += RString("\nScore=") + scorestr ;
            if (f.positions()) 
            {
                RString s;
                f.positionsStr(s);
                o += RString("\nPositions=") + s;
            }
        }

        HGObject* obj = makeObject(o, ProtPtr((BaseProt*)this));
        if (obj)
        {
//            if (fd->mode_ == fullTextQueryMode)
//                obj->parentObj_ = ObjectPtr(fd->objPtr_);    // this is the parent of all children !
//                obj->parentObjKey_ = fd->objPtr_.Obj()->IDString();    // this is the parent of all children !
            ObjectPtr oPtr(obj);
            result->addTail(new ObjectNode(oPtr));
        }
    }

    fd->pos_ += result->count();
}


boolean HyperGProt::insert(ObjectPtr object, 
                           ObjectPtr toColl)
{
    HGObject* hgObj = object.Obj();
    if (hgObj->Type() == "Document")
    {
        if (((Document*)hgObj)->DocType() == "collection")   // collection
            return hgConnPtr_->InsColl(hgObj->Info().GetRight(hgObj->Info().index('\n')+1)+"\n", 
                                       toColl.Obj()->IDString(), hgObj->_id);
        else                                    // simple document
            return hgConnPtr_->InsDoc(hgObj->Info().GetRight(hgObj->Info().index('\n')+1)+"\n", 
                                      toColl.Obj()->IDString(), "", hgObj->_id);
    }
    else
    if (hgObj->Type() == "Anchor")    // do something with an anchor ...
    {
        if (((Anchor*)hgObj)->IsSrc())
            return hgConnPtr_->insTypeSrcAnch( toColl.Obj()->ID(), 
                                               hgObj->Author(), 
                                               hgObj->Rights(), 
                                               "", "", "", 
                                               ((Anchor*)hgObj)->Position(), 
                                               ObjectID(((Anchor*)hgObj)->Dest()), 
                                               ((Anchor*)hgObj)->LinkType(), 
                                               hgObj->_id);
        else
            return hgConnPtr_->insDestAnch( toColl.Obj()->ID(), 
                                            hgObj->Author(), 
                                            hgObj->Rights(), 
                                            "", "", "", 
                                            hgObj->Title(), 
                                            "", 
                                            ((Anchor*)hgObj)->Position(), 
                                            hgObj->_id);
    }        

    return false;
}

boolean HyperGProt::erase(ObjectPtr object)
{ 
    return hgConnPtr_->deleteObject(object.Obj()->ID());
}

boolean HyperGProt::moveTo(ObjectPtr object, 
                           ObjectPtr fromColl, 
                           ObjectPtr toColl)
{ 
    HGObject* hgObj = object.Obj();
    ObjIdArray docs(1);
    docs.insert(hgObj->ID());

    if (hgObj->Type() == "Document")
    {
        if (((Document*)hgObj)->DocType() == "collection")   // collection
            return hgConnPtr_->MvCpCollsColl(true, fromColl.Obj()->ID(), toColl.Obj()->ID(), docs);
        else                                    // simple document
            return hgConnPtr_->MvCpDocsColl(true, fromColl.Obj()->ID(), toColl.Obj()->ID(), docs);
    }
    else
    if (hgObj->Type() == "Anchor")    // do something with an anchor ...
        //
        // ...
        //
        return false;

    return false;
}

boolean HyperGProt::copyTo(ObjectPtr object,
                           ObjectPtr toColl)
{ 
    HGObject* hgObj = object.Obj();
    ObjIdArray docs(1);
    docs.insert(hgObj->ID());

    if (hgObj->Type() == "Document")
    {
        if (((Document*)hgObj)->DocType() == "collection")   // collection
            return hgConnPtr_->MvCpCollsColl(false, 0, toColl.Obj()->ID(), docs);
        else                                    // simple document
            return hgConnPtr_->MvCpDocsColl(false, 0, toColl.Obj()->ID(), docs);
    }
    else
    if (hgObj->Type() == "Anchor")    // do something with an anchor ...
        //
        // ...
        //
        return false;

    return false;
}

boolean HyperGProt::unlinkFrom(ObjectPtr object, 
                                  ObjectPtr fromColl)
{ 
    HGObject* hgObj = object.Obj();
    ObjIdArray docs(1);
    docs.insert(hgObj->ID());

    if (hgObj->Type() == "Document")
    {
        if (((Document*)hgObj)->DocType() == "collection")   // collection
            return hgConnPtr_->MvCpCollsColl(true, fromColl.Obj()->ID(), 0, docs);
        else                                    // simple document
            return hgConnPtr_->MvCpDocsColl(true, fromColl.Obj()->ID(), 0, docs);
    }
    else
    if (hgObj->Type() == "Anchor")    // do something with an anchor ...
        //
        // ...
        //
        return false;

    return false;
}

HGObject* HyperGProt::lock(ObjectPtr objPtr)
{ 
    Object o;
    if (hgConnPtr_->GetAndLock(objPtr.Obj()->ID(), o))
    {
        HGObject* hgObj = makeObject(o, ProtPtr((BaseProt*)this));
//        hgObj->parentObj_ = objPtr.Obj()->parentObj_;   // copy parent from prev. object if any !
        hgObj->parentObjKey_ = objPtr.Obj()->parentObjKey_;   // copy parent from prev. object if any !
        return hgObj;
    }
    else
        return nil;
}

boolean HyperGProt::unlock(ObjectPtr objPtr)
{ 
    hgConnPtr_->Unlock(objPtr.Obj()->ID());
    return true;
}


//
//
boolean HyperGProt::modifyObjAttribute(ObjectPtr objPtr,
                                       RString oldAttrLine,
                                       RString newAttrLine)
{
    boolean success = false;
    if (oldAttrLine.length())
    {
        RString type = oldAttrLine.gSubstrDelim(0, '=');
        if (type.length() == oldAttrLine.length())
        {
            RString oldAttr = objPtr.Obj()->Info();
            int pos = oldAttr.indexa(type);
            oldAttrLine = oldAttr.gSubstrDelim(pos, '\n');
        }
        success = hgConnPtr_->ChangeObject(objPtr.Obj()->ID(), "rem " + oldAttrLine);
        if (!success)
            return success;
    }

    if (newAttrLine.length())
    {
        RString type = newAttrLine.gSubstrDelim(0, '=');
        if (type.length() == newAttrLine.length())
        {
            if (success)
                objPtr.Obj()->update();                   // update object information !
            return false;   // nonsense !
        }
        success = hgConnPtr_->ChangeObject(objPtr.Obj()->ID(), "add " + newAttrLine);
    }

    objPtr.Obj()->update();                   // update object information !

    return success;
}
