// <copyright>
// 
// Copyright (c) 1994,1995
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
// </copyright>


//<file>
//
// Name:       LangUtl.cpp
//
// Purpose:    using laguage dependencies
//
// Created:    21 Feb 1995   Gnter Nagler
//
// Modified:   17 Mar 1995   Thomas Dietinger
//
//
//</file>

#include <stdafx.h>
#include "langutl.h"
#include "language.h"
#include <ctype.h>

static HGLanguages LangTable;

#define MAXORDER (MAXLANGUAGES+MAXLANGUAGES)

HGLanguages* HGLanguages::DefTable()
{
  return &LangTable;
}

// default languages are english and german
HGLanguages::HGLanguages()
{
  Clear();

  Register("en", "English");
  Register("ge", "German");
  Register("fr", "Franais");
  Register("st", "Steirisch");
  Register("it", "Italiano");
  Register("sp", "Espaol");
}

HGLanguages::~HGLanguages()
{}

// remove registered languages
void HGLanguages::Clear()
{
  mLangCodes = 0;
}

// number of registered languages
int HGLanguages::Count()
{
  return mLangCodes;
}

int HGLanguages::ActiveCount()
{
int count = 0;

  for (int i = 0; i < mLangCodes; i++)
	if (mLangActive[i])
	  count++;
  return count;
}

void HGLanguages::Sort()
{
int i;

  for (i = 0; i < mLangCodes; i++)	// mark unused 
    mLangSort[i] = -1;
  for (i = 0; i < mLangCodes; i++)
     if (mLangActive[i])
       mLangSort[mLangOrder[i]] = i;
}

// returns 0 if too many languages registered
// returns number of registered languages otherwise
int HGLanguages::Register(RString langcode, RString langdesc, int active)
{
  if (FindCode(langcode) >= 0)  // already registered
    return mLangCodes;

  if (mLangCodes >= MAXLANGUAGES)
    return 0;
  mLangOrder[mLangCodes] = mLangCodes;  // lowest priority for newcomers
  mLangDescr[mLangCodes] = langdesc;  // lowest priority for newcomers
  mLangActive[mLangCodes] = active;
  mLangCode[mLangCodes++] = langcode;
  return mLangCodes;
}

// returns 0 if langcode is not registered
// returns 1 otherwise
int HGLanguages::Prefer(RString langcode)
{
   int i = FindCode(langcode);
   if (i < 0)  
     return 0;
   // mLangCode[i] should get a new priority
   for (int j = 0; j < mLangCodes; j++)
   if (mLangOrder[j] < mLangOrder[i]) // all currently better languages become a bad point
     mLangOrder[j] = mLangOrder[j]+1;
   mLangOrder[i] = 0;  // this will be the best language
   Sort();
   return 1;
}

// prefer a language that has now this order
int HGLanguages::Prefer(int order)
{
  for (int i = 0; i < mLangCodes; i++)
    if (mLangOrder[i] == order)
	  return Prefer(mLangCode[i]);
  return 0;	 // invalid language order
}

int HGLanguages::FindCode(RString langcode)
{
   // language codes must always be lower case
   langcode.tolower();
   // check if langcode is registered
   for (int i = 0; i < mLangCodes; i++)
     if (langcode == mLangCode[i])
	   return i;
   return -1;
}
 
RString HGLanguages::GetDescription(RString langcode)
{
   int i = FindCode(langcode);
   return i >= 0 ? mLangDescr[i] : "";
}
 
int HGLanguages::GetOrder(RString langcode)
{
   int i = FindCode(langcode);
   return (i >= 0 && mLangActive[i]) ? mLangOrder[i] : MAXORDER;
}

// get a language code that has now that order
RString HGLanguages::GetLangByOrder(int order)
{
   for (int i = 0; i < mLangCodes; i++)
   if (mLangActive[i] && order == mLangOrder[i])
	 return mLangCode[i];
   return RString::lambda();
}


// helper functions for compatibility reasons with older code
//
// return a Hyper-G language id (default: get highest ordered language)
int HGLanguages::getHgLangId(RString langcode)
{
    if (!langcode.length())
        langcode = GetLangByOrder(0);
    int langId = ENGLISH;

    if (langcode == "en")
        langId = ENGLISH;
    else
    if (langcode == "ge")
        langId = GERMAN;
    else
    if (langcode == "fr")
        langId = FRENCH;
    else
    if (langcode == "it")
        langId = ITALIAN;
    else
    if (langcode == "sp")
        langId = SPANISH;
    else
    if (langcode == "st")
        langId = STYRIAN;

    return langId;
}


// GetBestTitle: A Hyper-G Object has 0 or more title fields 
// get the one that seems to fit best into current language options
RString HGLanguages::GetBestTitle(RStringArray& titles)
{
int BestLangOrder, IsQuest, order;
RString langcode, DefaultTitle;

  
  BestLangOrder = MAXORDER;
  BestTitle = "";
  for (int i = 0; i < titles.count() && BestLangOrder > 0; i++)
  {
	if (titles[i][2] != ':')    // no language specified !
    {
      if (!DefaultTitle.length())
      {
        DefaultTitle = titles[i];
      }
	  continue;
    }
  	langcode = titles[i].GetSubstrIndex(0, 1);
	IsQuest = (langcode == "??");
	order = GetOrder(langcode);
	if (order < BestLangOrder || IsQuest)
	{
	  if (IsQuest)
        DefaultTitle = titles[i].GetRight(3);
	  else
	  {
		BestLangOrder = order;
		BestTitle = titles[i].GetRight(3);
		BestLanguage = langcode;
	  }
	}
  }
  // second best language is always ??
  if (BestLangOrder > 0 && DefaultTitle.length() > 0) 
  {
    BestTitle = DefaultTitle;
	BestLanguage = "??";
  }
  BestTitle.SGMLtoISO(); 
  return BestTitle;
}

RString HGLanguages::GetBestTitle(Object& hgobj)
{
RStringArray titles;

  hgobj.titles(titles);
  return GetBestTitle(titles);
}

RString HGLanguages::GetBestTitle(LangMsg& msg)
{
  return GetBestTitle(msg.msgs_);
}

// get the languagecodes in correct order
RString HGLanguages::GetOrder()
{
RString order;

  for (int ord = 0; ord < mLangCodes; ord++)
  {
    for (int i = 0; i < mLangCodes; i++)
    {
      if (mLangActive[i] && mLangOrder[i] == ord)
	  {
	    order += mLangCode[i];
	    break;
	  }
    }
  }
  return order;
}

// arrange the orders of registered languages
int HGLanguages::SetOrder(RString order)
{
int i, n;

  n = order.length()/2;

  // all languagecodes must be registered
  for (i = 0; i < n; i++)
    if (FindCode(order.GetSubstrIndex(i*2, i*2 + 1)) < 0)
	  return 0;

  // change order and activity
  for (i = 0; i < mLangCodes; i++)
    mLangActive[i] = 0;

  for (i = 0; i < n; i++)
  {
    mLangOrder[FindCode(order.GetSubstrIndex(i*2, i*2 + 1))] = i;
	mLangActive[i] = 1;
  }
  
  Sort();
  return 1;
}

// set active state of a language and get old state
int HGLanguages::SetActive(RString langcode, int isactive)
{
int oldstate;
int i;

  // is it a valid language code?
  i = FindCode(langcode);
  if (i < 0)
    return 0;  // no - it cannot be active
  oldstate = mLangActive[i];
  mLangActive[i] = (isactive ? 1 : 0);
  if (mLangActive[i] != oldstate)
  {
	if (mLangActive[i])
	{
	  // all actives that were less good become better
	  for (int j = 0; j != i && mLangActive[j] && mLangOrder[j] > mLangOrder[i]; j++)
		mLangOrder[j]--;
	  mLangOrder[i] = MAXORDER;
	}
	else // becomes worst of active languages
	  mLangOrder[i] = ActiveCount()-1;
  }
  Sort();
  return oldstate;
}
 
// get inactive language codes
RString HGLanguages::GetInactive()
{
RString inactive;

  for (int i = 0; i < mLangCodes; i++)
  if (!mLangActive[i])
	inactive += mLangCode[i];
  return inactive;
} 



/////////////////////////////////////////////////////////////////////////////
// LangMsg	for simple language dependent text
/////////////////////////////////////////////////////////////////////////////

LangMsg::LangMsg(RString msg1, RString msg2, RString msg3, RString msg4, RString msg5, RString msg6)
{
  Add(msg1);
  Add(msg2);
  Add(msg3);
  Add(msg4);
  Add(msg5);
  Add(msg6);
}

// get msg in best language
RString LangMsg::Msg(HGLanguages* lang)
{
  if (!lang)
    lang = &LangTable;

  if (!lang)
  {
	if (msgs_.count() > 0)
      return msgs_[0].GetRight(3); // without language code!
	else
	  return "";
  }
  return lang->GetBestTitle(*this);
}

void LangMsg::Add(RString msgs)
{
int starti = 0, endi, len = msgs.length();
RString s;

  while (starti < len)  
  {
	// skip spaces
	while (starti < len && isspace(msgs[starti]))
	  starti++;


	// search end of string
	endi = starti;
	while (endi < len && msgs[endi] != SEP[0])
	  endi++;
	if (endi - starti < 3) // not a valid language dependent message?
	{
	  starti = endi;
	  continue;
	}
	s = msgs.GetSubstrIndex(starti, endi-1);
    if (msgs[starti+2] != ':')	// author forgot to specify a language -> use it as default language
       s = "??:" + s;
	msgs_.insert(s);
	starti = endi;
  }
}

/*
inline LangMsg::operator RString() const {
     return RString LangMsg::Msg();
}
*/
