#ifndef __LANGUTL_H_  // g. nagler 
#define __LANGUTL_H_

#include "utils\str.h"
#include "hyperg\object.h"

class HGLanguages;
class LangMsg;

#define MAXLANGUAGES 10
#define SEP "|"

#define LANG_ACTIVE   1
#define LANG_INACTIVE 0

// a message has the form "ge: text"
class LangMsg
{
friend class HGLanguages;

public:
  LangMsg(RString msg1 = "", RString msg2 = "", RString msg3 = "", RString msg4 = "", RString msg5 = "", RString msg6 = "");
  
  RString Msg(HGLanguages* = 0);  // get msg in best language
  operator const char*() { return Msg().string(); }

  void Add(RString msgs);	// add new msgs of the form  "en:text"

protected:
  RStringArray msgs_;
};

class HGLanguages
{
public:
// default languages are english and german
  HGLanguages();
  ~HGLanguages();

  static HGLanguages* DefTable();

// remove registered languages
  void Clear();

  int Count();	// number of registered languages
  int ActiveCount(); // number of active languages

  int Register(RString langcode, RString langdescription, int active = LANG_ACTIVE);  // e.g. "en", "ge", ...
	// returns 0 if too many languages registered
	// returns number of registered languages otherwise

  int Prefer(RString langcode); // prefer a special language
  int Prefer(int order); // prefer a language that has now this order
    // returns 0 if langcode is not registered
	// returns 1 otherwise
  
  RString GetBestTitle(Object& hgobj);
  RString GetBestTitle(LangMsg& msg);
  RString GetBestTitle(RStringArray& arr);
  RString BestTitle;
  RString BestLanguage; // result of recent GetBestTitle()

  RString GetDescription(RString langcode); // description of language code en is english
  
  int GetOrder(RString langcode); // get order of special language
  RString GetLangByOrder(int order); // get a language code that has now that order

  int getHgLangId(RString langcode = "");   // return a Hyper-G language id (default: get highest ordered language)

  RString GetOrder(); // get active language codes sorted by current order
  int SetOrder(RString order); // arrange the orders of registered languages

  int SetActive(RString langcode, int isactive = LANG_ACTIVE); // set active state of a language and get old state
  RString GetInactive(); // get inactive language codes

protected:
  int mLangCodes;
  RString mLangCode[MAXLANGUAGES]; // registered languages
  RString mLangDescr[MAXLANGUAGES]; // description of registered language
  int mLangOrder[MAXLANGUAGES];  // 0 is highest priority
  int mLangSort[MAXLANGUAGES];   // sorted indices by order
  int mLangActive[MAXLANGUAGES]; // only active languages are being searched

  int FindCode(RString langcode); // returns index in register table
  void Sort(); // sort by order
};

/////////////////////////////////////////////////////////////////////////////

#endif
