// This may look like C code, but it is really -*- C++ -*-

// **********************************************************
// File objectlist.C
// ----------------------------------------------------------
// Created: F. Kappe
// 
// An ObjectList stores a list of Hyper-G Objects
// **********************************************************

#include "objectlist.h"
#include <stdio.h>

// **********************************************************
// class IdNode
// ----------------------------------------------------------
// **********************************************************

extern RString SortOrder;

boolean ObjectNode::operator < (DLListNode& n) 
{
    const char *p1 = MainKey();
    const char *p2 = ((ObjectNode&)n).MainKey();
    for (; *p1 && *p1==*p2; p1++, p2++);
    if (!*p1) 
        return true;   // p1 is prefix of p2
    else 
        return *p1 < *p2;
}


RString ObjectNode::MainKey() 
{
    if (!_cached) 
    {
        boolean invert=false;
        RString author;
        for (int i=0; i<SortOrder.length(); i++)
            switch (SortOrder[i]) 
            {
                case '-': invert=true; break;
                case '#': 
                { 
                    RString seq = _object.Obj()->Sequence();
                    if (!seq.length()) 
                        seq = "0";

                    long sequence= atol(seq.string());
                    char tmp[20];
                    if (sequence < 0) 
                    {
                        sequence=-sequence;
                        sprintf(tmp,"%08d",sequence);
                        seq=tmp;
                        seq.invert();
                        seq = "-" + seq;
                    }
                    else 
                    {
                        sprintf(tmp,"%08d",sequence);
                        seq=tmp;
                    }

                    if (invert) 
                        _keycache += seq.invert() + "\255";
                    else
                        _keycache += seq + "\1";
                    invert=false;
                    break;
                }
                case 'A': 
                { 
                    int pos = 0;
                    _object.Obj()->Author().GetWord(pos,"@ ",author);
                    if(invert) 
                        _keycache += author.tolower().invert()+ "\255";
                    else 
                        _keycache += author.tolower() + "\1";
                    invert=false;
                    break;
                }
                case 'C': 
                { 
                    RString ct = _object.Obj()->Created();
                    if(invert) 
                        _keycache += ct.invert() + "\255";
                    else
                        _keycache += ct + "\1";
                    invert=false;
                    break;
                }
                case 'E': 
                { 
                    RString ex = _object.Obj()->Expires();
                    if (invert)
                        _keycache += ex.invert() + "\255";
                    else
                        _keycache += ex + "\1";
                    invert=false;
                    break;
                }
                case 'I':   // sort after unique ID
                { 
                    int pos = 0;
                    author = _object.Obj()->IDString();
                    if(invert) 
                        _keycache += author.tolower().invert()+ "\255";
                    else 
                        _keycache += author.tolower() + "\1";
                    invert=false;
                    break;
                }
                case 'O': 
                { 
                    RString op = _object.Obj()->Opens();
                    if (invert)
                        _keycache += op.invert() + "\255";
                    else
                        _keycache += op + "\1";
                    invert=false;
                    break;
                }
                case 'P': 
                { 
                    RString p = _object.Obj()->Parent();
                    p.tolower();
                    if (invert)
                        _keycache += p.invert() + "\255";
                    else
                        _keycache += p + "\1";
                    invert=false;
                    break;
                }
                case 'S': 
                { 
                    HGObject *d =  _object.Obj() ;
                    RString s;
                    if (d->Type() == "Document") 
                        s = ((Document*)d)->Score();
                    if (invert)
                        _keycache += s.invert() + "\255";
                    else
                        _keycache += s + "\1";
                    invert=false;
                    break;
                }
                case 'T': 
                { 
                    RString title = _object.Obj()->Title();
                    if (invert)
                        _keycache += title.SGMLISOtolowerASCII().invert() + "\255";
                    else
                        _keycache += title.SGMLISOtolowerASCII() + "\1";
                    invert=false;
                    break;
                }
                case 't': 
                { 
                    // make a unique key out of the type
                    char tmpChar[2];
                    tmpChar[0] = (int)_object.Obj()->getType() + 'A';
                    tmpChar[1] = 0;
                    RString t = tmpChar;

                    if (invert)
                        _keycache += t.invert() + "\255";
                    else
                        _keycache += t + "\1";
                    invert=false;
                }

                default: 
                { 
                    if (invert)
                        _keycache += "\255";
                    else
                        _keycache += "\1";
                    invert=0;
                }
            }
        _cached = true;
    }
    return _keycache;
}

void ObjectNode::FlushCache() 
{
    _cached = false;
    _keycache = RString();
}

/////////////////////////////////////////////////////////////
// class ObjectList
// ----------------------------------------------------------
//

ObjectNode* ObjectList::find(RString idKey)
{
    for (ObjectNode* n = getFirst(); 
        n && n->obj() && n->obj()->IDString() != idKey; 
        n = (ObjectNode*)getNext(n));

    return n;
}

void ObjectList::sortAndTrashDuplicates() 
{
    RString oldSortOrder = SortOrder;
    SortOrder = oldSortOrder + "I";
    ObjectNode* n = 0;

    // need to empty list cache
    for (n = getFirst(); n; n = getNext(n))
        n->FlushCache(); 

    sort();

    RString lastKey = "$";
    for (n = getFirst(); 
        n && n->obj() && n->obj()->IDString(); 
        n = (ObjectNode*)getNext(n))
    {
        while (lastKey==n->obj()->IDString())
        {
            ObjectNode* n2 = (ObjectNode*)getNext(n);
            remove(n);
            delete(n);
            n = n2;
        }

        lastKey = n->obj()->IDString();
    }

    SortOrder = oldSortOrder;
}

/////////////////////////////////////////////////////////////
// class RStringList
// ----------------------------------------------------------
//

RStringNode* RStringList::find(RString n2)
{
    for (RStringNode* n = getFirst(); 
        n && n->str_ != n2; 
        n = (RStringNode*)getNext(n));

    return n;
}
