// This may look like C code, but it is really -*- C++ -*-

//<copyright>
// 
// Copyright (c) 1993,94,95
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        objectlist.h
//
// Purpose:     Basic Hyper-G object classes: class HGObject and derived Classes
//
// Created:                  Frank Kappe
//                           Thomas Dietinger
//
// Modified:    22 Jun 95    Thomas Dietinger
//
//
//
// Description: an RStringList stores a list of RStrings
//              an ObjectList stores a list of Hyper-G Objects
//
//
//</file>

#ifndef objectlist_h
#define objectlist_h

#include "hgobject.h"
#include <utils/list.h>

/////////////////////////////////////////////////////////////
// class ObjectNode
// ----------------------------------------------------------
//
class ObjectNode : public DLListNode 
{
public:
    ObjectNode (const ObjectPtr&);

    ObjectPtr ObjPtr() const        { return _object; }
    HGObject* obj() const           { return _object.Obj(); }
    RString MainKey();
    void FlushCache();
    boolean operator < (DLListNode& n); 

private:
    ObjectPtr _object;
    RString _keycache;
    boolean _cached;
};

inline ObjectNode::ObjectNode(const ObjectPtr& o) : DLListNode() 
{
    _object = o;
    _cached = false;
}

/////////////////////////////////////////////////////////////
// class ObjectList
// ----------------------------------------------------------
//

DLListdeclare(TmpObjectList, ObjectNode)

class ObjectList : public TmpObjectList
{
public:
    ObjectNode* find(RString idKey);
    ObjectNode* find(ObjectPtr n)       { return find(n.Obj()->IDString()); }
    void sortAndTrashDuplicates();      // sorts list and trashes all duplicate objects (identified by IDString)
};


///////////////////////////////////////////////////////////////////
// class ObjListNode: list node including an object list
// ----------------------------------------------------------
//
class ObjectListNode : public DLListNode 
{
public:
    ObjectList* list()      { return &objList_; }
    HGObject* firstObj();  // return first object in list if any
private:
    ObjectList objList_;
};

inline HGObject* ObjectListNode::firstObj()
{ 
    if (objList_.getFirst())
        return ((ObjectNode*)objList_.getFirst())->obj(); 
    else
        return 0;
}


///////////////////////////////////////////////////////////////////
// class ObjListList: list of object lists
// ----------------------------------------------------------
//
DLListdeclare(ObjectListList, ObjectListNode)


/////////////////////////////////////////////////////////////
// class RStringNode
// ----------------------------------------------------------
//

class RStringNode : public DLListNode
{
public:
    RStringNode(RString str) : DLListNode() { str_ = str; }
    RString str_;
};


/////////////////////////////////////////////////////////////
// class RStringList
// ----------------------------------------------------------
//

DLListdeclare(TmpRStringList, RStringNode)

class RStringList : public TmpRStringList
{
public:
    RStringNode* find(RString n);
};

#endif
