// This may look like C code, but it is really -*- C++ -*-

// **********************************************************
// File offsetlist.h
// An OffsetList maintains an ordered list of offsets.
// ----------------------------------------------------------
// Created: F. Kappe
// **********************************************************

#ifndef offsetlist_h
#define offsetlist_h

#include <utils/list.h>

#define OFFSETTYPE ObjectID

// **********************************************************
// class OffsetNode
// ----------------------------------------------------------
// **********************************************************

class OffsetNode : public DLListNode {
 public:
  OffsetNode (OFFSETTYPE o, ObjectID i) : DLListNode() { _id = i; _offset = o; }

  OffsetNode (OFFSETTYPE o, ObjectID i, const RString& ltype) : DLListNode() { 
                                                         _id = i; 
							 _offset = o;
						         _linktype = ltype; }

  ObjectID ID() const       { return _id; }
  OFFSETTYPE offset() const { return _offset; }
  RString LinkType() const  { return _linktype; }

 private:
  OFFSETTYPE _offset;
  ObjectID _id;
  RString _linktype;
};

// **********************************************************
// class OffsetList
// ----------------------------------------------------------
// **********************************************************

DLListdeclare(TmpOffsetList,OffsetNode)

class OffsetList : public TmpOffsetList {
 public:
  OffsetNode* Find(OFFSETTYPE offset);
  void Insert(OffsetNode* n);
};



inline OffsetNode* OffsetList::Find(OFFSETTYPE offset) {
  // find first node with offset >= this offset
  for (OffsetNode* n = getFirst(); n && n->offset() < offset; n = getNext(n)) ;
  return n;
}

inline void OffsetList::Insert(OffsetNode* n) {
  // insert at currect position so that ordering remains intact
  OffsetNode* t = Find(n->offset());
  if (t) 
       insertBefore( n, t);
  else
       addTail( n);
}

#endif
