//<copyright>
// 
// Copyright (c) 1993
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:       userprof.C
//
// Purpose:    UserProfile - manages the user record
//
// Created:    07 Nov 93    Frank Kappe
//
// Modified:   07 Nov 93    Frank Kappe
//
//
//
// Description:
//
// Contains class UserProfile.
//
//</file>

#include "userprof.h"
#include "utils.h"

#if defined(sparc) && OSMajorVersion >= 5
#include <crypt.h>
#endif

#ifdef SYSV
#include <unistd.h>
#include <time.h>
#define srandom srand
#define random rand
#endif

UserProfile* UserProfile::instance_;

UserProfile::UserProfile(HgStub* db) 
{
    db_ = db;
    initProfile();
}

UserProfile::~UserProfile()
{
}

void UserProfile::initProfile() 
{
    if (db_)
        db_->GetObject(db_->userId(), o_);
}

ObjectPtr UserProfile::homeCollection(const RString& defaultname) 
{
    if (!db_)
        return ObjectPtr();

    RString newhome = o_.field("Home=");
    if (newhome == RString()) // no home specified or anonymous; use default
        newhome = defaultname;
    else 
        if (CheckCollection(*db_, newhome) == 0) // home specified but not existing
            newhome = defaultname;

    if (oldHome_ != newhome)   // otherwise we dont generate extra transactions
    { 
        oldHome_ = newhome;

        ObjIdArray olist;
        db_->getObjByQuery("Name=" + newhome, olist);
        ObjectList* homes = new ObjectList();
        ReceiveObjects(*db_, olist, "DocumentType=collection", homes);  

        if (homes->count() == 1)  // there should be at most one (names are unique)
            homecollection_ = (*homes)[0]->ObjPtr();
        else
            homecollection_ = nil;

        delete homes;
    }
    return homecollection_;
}

boolean UserProfile::changePassword(const RString& oldpw, const RString& newpw) {
    return changePasswordCrypt(oldpw, enCrypt(newpw));
}

boolean UserProfile::changePasswordCrypt(const RString& oldpw, const RString& newpw) {
    RStringArray pwds;
    o_.fields("Passwd=", pwds);

    for (int i=0; i<pwds.count(); i++) {
        if (validatePassword(oldpw, pwds[i])) { // this is the pwd we want to change
            RString command = "rem Passwd=" + pwds[i] + "\\add Passwd=" + newpw;
            return db_->ChangeObject(db_->userId(), command);
        }
    }
    return false;         // old password was wrong
}

///////////////////////////////// private methods ////////////////////////


//////////////////////////////////////////////////////////////////////////////
//
// encrypts a password using random salt:
//
RString UserProfile::enCrypt(const RString& clear) {
#ifdef LOCAL_DB_20
    return RString(clear);
#else
    static const char *saltletters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789./";

    srand((int)::time(nil));

    char salt[3];
    salt[0] = saltletters[(int)rand()%strlen(saltletters)];
    salt[1] = saltletters[(int)rand()%strlen(saltletters)];
    salt[2] = '\0';

    const char* tmp = ::crypt((const char*)clear, salt);
    return RString(tmp);
#endif
}

//////////////////////////////////////////////////////////////////////////////
//
// checks whether the clear text version corresponds to the encrypted version
//
boolean UserProfile::validatePassword(const RString& clear, const RString& encr) {
#ifdef LOCAL_DB_20
    return true; // local DB dosn't support identification
#else
    const char* tmp = ::crypt((const char*)clear, (char*)(const char*)encr);
    return (strcmp(tmp, (const char*) encr) == 0);
#endif
}
