//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        utils.cpp
//
// Purpose:     Utility routines used by some modules
//
// Created:     14 Jan 92    Frank Kappe (utils.C for HGC-T)
//                           Thomas Dietinger (modified for use with Amadeus)
//
// Modified:    15 Sep 95    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>



#include <stdafx.h>

#define MAXRECEIVE 50

#ifndef _CONSOLE
#include <hgapp.h>
#include <mainfrm.h>
#else
#include <resource.h>
#include <pcutil/consutil.h>
#include "language.h"
#endif

#include <pcutil/socket.h>

#include "conf.h"

#include "utils.h"
#include "treaps.h"
#include "baseprot.h"
#include <l_server/s_ldbprt.hpp>

#include <Dispatch/rpcbuf.h>
#include <server/hgstub.h>
#include "hgprot.h"

#include <hyperg/dcerror.h>

#include <time.h>
#include "hglinks.h"

#if defined(_MSC_VER) & defined(_DEBUG)
#define new DEBUG_NEW
#endif

// global variables

//extern HgStub* HG;
extern RString verString;

static RString WWW_Protocol = "http";       // current WWW protocol
static RString WWW_Host = "";               // current WWW host
static RString WWW_Port = DEFAULT_WWW_PORT; // current WWW port
static RString WWW_Path = "";               // current WWW path


///////////////////////// Internal Routines ///////////////////////


/////////////////////////////////////////////////////////////////////////////
// this function constructs an unique temporary file name using the info it gets
// from the object id structure (stored in a text file)
RString getTempName(Document* o)
{                      
    RString tempName = getenv("TEMP");
    if (tempName.length() && tempName[tempName.length()-1] != '\\')
        tempName += RString("\\");
    
    tempName += o->ID().IDString().gRight(2) + ".*";

    TRACE("utils.cpp:getTempName: %s\n", (const char*)tempName);
    return tempName;
} // getTempName

  
/////////////////////////////////////////////////////////////////////////////
// 

RString getTempNameExt( const char *testFileBuf, int bufLen )
{                                                                  
    char buf[128];

    buf [0] = '\0';

// this function is NOT supported for concolse applications !
#ifndef _CONSOLE
    CString magicFileName = theApp.GetPrivateProfileString("DocManager", "MagicFile", "magic");
    
    findFileType(magicFileName, testFileBuf, bufLen, buf, 128);
#endif
    
    TRACE("utils.cpp:getTempNameExt: %s\n", (const char*)buf);
//    ASSERT(strlen(buf));
    return buf;
} // getTempNameExt

               
/////////////////////////////////////////////////////////////////////////////
// This function looks for an existing file, where the filename may be specified
// without extension. (If it's a non-generic type)
// If successful, it returns TRUE, else FALSE; the extension is added at the end of
// the input filename (if possible).
BOOL getCompleteFileName( RString& fileName, boolean& forceLoad)
{
    WIN32_FIND_DATA fileInfo;
    HANDLE hFind;
    CString tempName = fileName;
       
    if ( (hFind = FindFirstFile(tempName, &fileInfo)) == INVALID_HANDLE_VALUE)
    {
        forceLoad = false;
        return FALSE;                   // file not found !
    }
                                           
    if (fileName[fileName.length()-1] == '*') // no generic object !
    {
        tempName.Empty();
        tempName = CString(fileInfo.cFileName);
        fileName = fileName.gSubstrIndex(0, fileName.length()-2);
        if (tempName.ReverseFind('.') > -1)
             fileName += RString(tempName.Mid( tempName.ReverseFind('.') + 1));  // add file name extension
    }

    FindClose(hFind);

    TRACE("utils.cpp:getCompleteFileName: %s\n", (const char*)fileName);

    if (forceLoad)         // must load new file, old one is not valid any more !
    {
        unlink(fileName);   // delete old file
        forceLoad = false; // ok, for next load attempt
        return FALSE;
    }

    return TRUE;
} // getCompleteFileName


/////////////////////////////////////////////////////////////////////////////

enum WWWDocType { unknowWWWDocType = 0, textWWWDocType = 1, imageWWWDocType = 2,
                  soundWWWDocType = 3, movieWWWDocType = 4, postscriptWWWDocType = 5 };

WWWDocType getWWWDocType(Object& o)
{
    TRACE("utils.cpp:getWWWDocType:DocType=%s, RemoteDocType=%s, SubType=%s\n", 
          o.documenttype().string(), o.field("RemoteDocType=").string(), o.field("SubType=").string());
    if (o.field("SubType=") == "inline" || o.field("RemoteDocType=") == "Image")
    {
        if (!o.field("RemoteDocType=").length())
            o += "\nRemoteDocType=Image";                          
        return imageWWWDocType;
    }
    else
    if (o.field("RemoteDocType=") == "Sound")
        return soundWWWDocType;
    if (o.field("RemoteDocType=") == "Movie")
        return movieWWWDocType;
    if (o.field("RemoteDocType=") == "PostScript")
        return postscriptWWWDocType;
    else
    {
        if (!o.field("RemoteDocType=").length())
            o += "\nRemoteDocType=text";                          
        return textWWWDocType;
    }
}

/////////////////////////////////////////////////////////////////////////////
//
// This will make a viewer Object out of the object's text description
//

HGObject* MakeObject(Object& o) 
{
//    HgStub* hgConnPtr = HG;
    HGObject* obj = nil;
    RString type = o.type();
    if (type == "Document") 
    {
        RString doctype = o.documenttype();
        if (doctype == "text")
            obj = new TextDocument(o);
        else 
        if (doctype == "PIC")
            obj = new PICDocument(o);
        else 
        if (doctype == "Image")
            obj = new ImageDocument(o);
        else 
        if (doctype == "Movie")
            obj = new MovieDocument(o);
        else 
        if (doctype == "PostScript")
            obj = new PostScriptDocument(o);
        else 
        if (doctype == "Sound")
            obj = new SoundDocument(o);
        else 
        if (doctype == "Scene")
            obj = new SceneDocument(o);
        else 
        if (doctype == "Generic")
            obj = new GenericDocument(o);
        else 
        if (doctype == "collection") 
        {
            RString colltype = o.collectiontype();
            if (colltype == "Cluster")
                obj = new Cluster(o);
            else
                obj = new Collection(o);
        }
        else 
        if (doctype == "Remote") 
        {
            RString protocol = o.protocol();
            if (protocol == "http")
            {           
                if (!o.field("Path=").length())
                    o += "Path=/\n";

                WWWDocType wwwType = getWWWDocType(o);
/*
                if (hgConnPtr)
                {
*/
                    switch (wwwType)
                    {
                        default :
                        case textWWWDocType  : 
                            obj = new TextDocument(o); break;
                        case imageWWWDocType : 
                            obj = new ImageDocument(o); break;
                    }
/*
                }                 
                else                            // direct connection (without the help of
                {                               // a Hyper-G server)
                    obj = new WWWDocument(o);   
                    switch (wwwType)
                    {
                        default :
                        case textWWWDocType  : 
                            obj = new WWWDocument(o); break;
                        case imageWWWDocType : 
                            obj = new ImageDocument(o); break;
                    }
                }
*/
            }
            else 
            if (protocol == "telnet" ||
                protocol == "tn3270")
                obj = new TelnetSession(o);
            else 
            if (protocol == "gopher") 
            {
                char gtype = o.gophertype()[0];
                switch(gtype) 
                {
                    case '1' : 
                        obj = new GopherCollection(o);
                        break;
                    case '7' : 
                        obj = new GopherIndex(o);
                        break;
                    case 'I' : //old
                    case 'g' : //old
                    case ':' : 
                        obj = new GopherImage(o);
                        break;
                    case 'S' : //old
                    case 's' : //old
                    case '<' : 
                        obj = new GopherSound(o);
                        break;
                    default : 
                        obj = new GopherText(o);
                        break;
                }
            }
            else 
            if (protocol == "wais") 
            {
                RString wtype = o.field("WaisType=");
                if (wtype == "WSRC")
                    obj = new WaisSource(o);
                else 
                if (wtype == "GIF" || wtype == "TIFF" || wtype == "PICT")
                    obj = new WaisImage(o);
                else 
                    obj = new WaisText(o);
            }
            else 
            if (protocol == "ftp")
                obj = new FTPDocument(o);
        }
    }
    else 
    if (type == "Anchor")
        obj = new Anchor(o);

    return obj;
} // MakeObject


HGObject* makeObject(Object& o, ProtPtr pPtr) 
{
    HGObject* tmpObj = MakeObject(o);

    if (tmpObj)
    {
//        if (pPtr.protocol()->useConnection(tmpObj->IDString()))
            tmpObj->setProt(pPtr);
//        else
//            tmpObj->setProt(makeProt(pPtr));         
    }
    return tmpObj;
}
/*
HGObject* makeObjectById(HgStub* db, RString dest)
{
    HGObject* hgo = 0;

    int pos = dest.index(':');      // test for protocol specifier (eg. "http:", "gopher:", ..)
    if (pos != -1)
    {
        // remote-link !!
        RString protocol = dest.GetSubstrIndex(0,pos-1);
        RString host = dest.GetSubstrDelim(pos+3,'/');
        pos += 2 + host.length();
        RString path = dest.GetSubstrDelim(pos+1,'#'); // no dest anchor!!

        RString port = host.GetRight(host.index(':')+1);
        host = host.GetSubstrDelim(0,':');

        Object o = "ObjectID=0x00000000\nType=Document\nDocumentType=Remote\n";
        o += "Author=system\nProtocol=" + protocol + "\nHost=" + host + 
             "\nPort=" + port + "\nPath=" + path + "\nTitle=??:" + path + 
             "\nTimeCreated=" + TimeNow() +"\n";

        if (protocol == "gopher")   // especially for gophers: GopherType !
        {
            if (path.gRight(1) == "/")
                path = path.gRight(1);      // ignore leading '/'
            o += "GopherType=" + path.gSubstrIndex(0, 0) + "\n";
        }

        hgo = MakeObject(o);
    }
    else
    {
        if (!db)
        {
            ErrorMsg("Not connected to a Hyper-G database !");
            return hgo;
        }
        
        Object o;
        boolean success;

        if (dest.GetSubstrIndex(0,1) == "0x")   // passed ID
        {
//            success = db->GetObject(ObjectID(dest), o);
//###
            int ndx = 0;
            RString word2;
            ObjectID helpid;
            dest.gWord(ndx, " _", word2);  // set ndx after the first word
            if (dest.gWord(ndx, " _", word2))
            {
                if (word2.GetSubstrIndex(0,1) == "0x")
                {
                    RString tmp = dest;
                    tmp.subst('_', ' '); // in the url _ seperates the two parts of the GOID
                    helpid = db->mapId(tmp);
                }
                else // grumml because of sorry not implemented with ATNT
                    helpid = dest;
            }
            else 
                helpid = dest;

            success = db->GetObject(helpid, o);
//###
        }
        else                                    // passed collection name
        {
            ObjectID oId;
            oId = db->GetFirstCollID(dest);
            if (!oId.id())
                success = false;    // didn't find collection !
            else
                success = db->GetObject(oId, o);
        }
        if (success) 
        {
            hgo = MakeObject (o);
        }
    }

    return hgo;
}
*/

//splitURL(RString url, RString& protocol, RString& user, 
//              RString& passwd, RString& host, int portNr, RString& path)

ProtPtr makeProtocolById(RString dest, ProtPtr pPtr = ProtPtr(0))
{
    //
    // calculate correct protocol (out of dest)
    //

    RString protocol;
    int pos = dest.index(':');
    if (pos > 0)
        protocol = dest.gSubstrIndex(0, pos - 1);

    // steve
    ProtocolNode* protNodeTest = HGObject::conOrProtList_.getFirst();
    if (protNodeTest  && 
        protNodeTest->protocol()  &&
        protNodeTest->protocol()->getProtType() == "hglocal")
            protocol = "hglocal";


    // protocols our Hyper-G server is capable of
    if (!protocol.length()   || 
        protocol == "hyperg" ||
        protocol == "http"   ||
        protocol == "gopher")
    {
        if ((pPtr.protocol() && pPtr.protocol()->getProtType() != "hyperg") || !pPtr.protocol())
        {
            // look for Hyper-G protocol in protocol list, or create a new protocol
            // steve
            for (ProtocolNode* protNode = HGObject::conOrProtList_.getFirst();
                 protNode && protNode->protocol() && protNode->protocol()->getProtType() != "hyperg";
                 protNode = HGObject::conOrProtList_.getNext(protNode));

            if (protNode)
                pPtr = protNode->protPtr();
            else
            {
                // create a new Hyper-G protocol connection
                if (!protocol.length() || protocol == "hyperg")
                {
                    BaseProt* newProt = new HyperGProt(dest);
                    if (newProt)
                    {
                        pPtr = ProtPtr(newProt);
                        HGObject::conOrProtList_.addHead(new ProtocolNode(pPtr));
                    }
                }
                // else, we have to use native http and gopher protocol !
            }
        } 
    }
    else 
        if (protocol == "hglocal") {
            // look for Local-Hyper-G-Database protocol in protocol list, or create a new protocol
            for (ProtocolNode* protNode = HGObject::conOrProtList_.getFirst();
                 protNode && protNode->protocol() && protNode->protocol()->getProtType() != "hglocal";
                 protNode = HGObject::conOrProtList_.getNext(protNode));

            if (protNode)
                pPtr = protNode->protPtr();
            else
            {
                BaseProt* newProt = new LocalDatabaseProt(dest);
                if (newProt)
                {
                    pPtr = ProtPtr(newProt);
                    HGObject::conOrProtList_.addTail(new ProtocolNode(pPtr));
                }
            }
        }

    if (!pPtr.protocol())   // we don't have a protocol so far ...
    {
        //
        // other protocols ...
        //
/*
        if (protocol == "http")
        {
        }
        else
        if (protocol == "gopher")
        {
        }
        if (protocol == "ftp")
        {
        }
        if (protocol == "telnet")
        {
        }
*/
        return pPtr;    // currently not supported !
    }

    return pPtr;
}

HGObject* makeObjectById(RString dest, ProtPtr pPtr)
{
    HGObject* tmpObj = 0;

    pPtr = makeProtocolById(dest, pPtr);

    Object obj = pPtr.protocol()->getObjectInfo(dest);
    if (obj.length() && (tmpObj = MakeObject(obj)))
        tmpObj->setProt(pPtr);

    return tmpObj;
}

/*
// get object information by destination ID/URL
Object getObjectInfo(RString dest)
{
    Object obj;
    int pos = dest.index(':');      // test for protocol specifier (eg. "http:", "gopher:", ..)
    if (pos != -1)
    {
        // remote-link !!
        RString protocol = dest.GetSubstrIndex(0,pos-1);
        RString host = dest.GetSubstrDelim(pos+3,'/');
        pos += 2 + host.length();
        RString path = dest.GetSubstrDelim(pos+1,'#'); // no dest anchor!!

        RString port = host.GetRight(host.index(':')+1);
        host = host.GetSubstrDelim(0,':');

        obj  = "ObjectID=0x00000000\nType=Document\nDocumentType=Remote\n";
        obj += "Author=system\nProtocol=" + protocol + "\nHost=" + host + 
               "\nPort=" + port + "\nPath=" + path + "\nTitle=??:" + path + 
               "\nTimeCreated=" + TimeNow() +"\n";

        if (protocol == "gopher")   // especially for gophers: GopherType !
        {
            if (path.gRight(1) == "/")
                path = path.gRight(1);      // ignore leading '/'
            obj += "GopherType=" + path.gSubstrIndex(0, 0) + "\n";
        }
    }
    else
    {
        if (!HG)
        {
            ErrorMsg("Not connected to a Hyper-G database !");
            return obj;
        }
        
        boolean success;

        if (dest.GetSubstrIndex(0,1) == "0x")   // passed ID
        {
            int ndx = 0;
            RString word2;
            ObjectID helpid;
            dest.gWord(ndx, " _", word2);  // set ndx after the first word
            if (dest.gWord(ndx, " _", word2))
            {
                if (word2.GetSubstrIndex(0,1) == "0x")
                {
                    RString tmp = dest;
                    tmp.subst('_', ' '); // in the url _ seperates the two parts of the GOID
                    helpid = HG->mapId(tmp);
                }
                else // grumml because of sorry not implemented with ATNT
                    helpid = dest;
            }
            else 
                helpid = dest;

            success = HG->GetObject(helpid, obj);
        }
        else                                    // passed collection name
        {
            ObjectID oId;
            oId = HG->GetFirstCollID(dest);
            if (!oId.id())
                success = false;    // didn't find collection !
            else
                success = HG->GetObject(oId, obj);
        }
    }

    return obj;
}
*/
//
// vId .. supply remote objects with virtual temporary link ids !
// - should once be replaced by a universal, global object id ! (stored in 'dest')
//
//
/*
HGObject* makeInlineObjectById(HgStub* db, RString dest, RString vId)
{
*/
/*
    HGObject* hgo = 0;
    int pos = dest.index(':');      // test for protocol specifier (eg. "http:", "gopher:", ..)
    if (pos != -1)
    {
        // remote-link !!
        RString protocol = dest.GetSubstrIndex(0,pos-1);
        RString host = dest.GetSubstrDelim(pos+3,'/');
        pos += 2 + host.length();
        RString path = dest.GetSubstrDelim(pos+1,'#'); // no dest anchor!!

        RString port = host.GetRight(host.index(':')+1);
        host = host.GetSubstrDelim(0,':');

        Object o = "ObjectID=" + vId + "\nType=Document\nDocumentType=Remote\n";
        o += "Author=system\nProtocol=" + protocol + "\nHost=" + host + 
             "\nPort=" + port + "\nPath=" + path + "\nTitle=??:" + path + 
             "\nTimeCreated=" + TimeNow() +"\nSubType=inline\n";

        if (protocol == "gopher")   // especially for gophers: GopherType !
        {
            if (path.gRight(1) == "/")
                path = path.gRight(1);      // ignore leading '/'
            o += "GopherType=" + path.gSubstrIndex(0, 0) + "\n";
        }

        o += "RemoteDocType=Image\n";
        o += "InlineID=" + dest + "\n";

        hgo = MakeObject(o);
    }
    else
    {
        Object o;
        ObjectID oId(dest);
        if (db->GetObject(oId, o)) 
        {
//            oId = o.ID();       // work-around for replica-bug !
            o += "SubType=inline\n";
            o += "InlineID=" + oId.IDString() + "\n";
            hgo = MakeObject (o) ;
        }
        else
            ASSERT(0);
    }

    return hgo;
*/
/*
    HGObject* tmpObj = 0;

    ProtPtr pPtr = makeProtocolById(dest);

    Object obj = pPtr.protocol()->getObjectInfo(dest);
    if (obj.length())
    {
        ObjectID oId(dest);
        obj += "SubType=inline\n";
        obj += "InlineID=" + oId.IDString() + "\n";
        if (tmpObj = MakeObject(obj))
            tmpObj->setProt(pPtr);
    }

    return tmpObj;
}
*/

//
// vId .. supply remote objects with virtual temporary link ids !
// - should once be replaced by a universal, global object id ! (stored in 'dest')
//
HGObject* makeInlineObjectById(RString dest, RString vId)
{
    HGObject* tmpObj = 0;

    ProtPtr pPtr = makeProtocolById(dest);

    Object obj = pPtr.protocol()->getObjectInfo(dest);
    if (obj.length())
    {
        ObjectID oId(dest);
        obj += "SubType=inline\n";
        obj += "InlineID=" + oId.IDString() + "\n";
        if (tmpObj = MakeObject(obj))
            tmpObj->setProt(pPtr);
    }

    return tmpObj;
}

//
// This will get an object by its anchor and makes an inline object out of it
//
/*
ObjectPtr getInlineObjectByAnchor(HgStub& db, ObjectID linkId)
{

    Object o;
    ObjectPtr tmpPtr(0);
    if (HG->GetObject(linkId, o)) 
    {
//        linkId = o.ID();
        o += "SubType=inline\n";
        o += "InlineID=" + linkId.IDString() + "\n";
        HGObject* hgo = MakeObject (o) ;
        tmpPtr = ObjectPtr(hgo);
    }
*/
/*
    ObjectPtr tmpPtr(0);
    ObjectList tmp;

    GetDestByAnchor(db, &tmp, linkId, "");
    if (tmp.getFirst()) 
    {
        Object o;
        if (HG->GetObject(tmp.getFirst()->ObjPtr().Obj()->ID(), o)) 
        {
            o += "SubType=inline\n";
            o += "InlineID=" + linkId.IDString() + "\n";
            HGObject* hgo = MakeObject (o) ;
            tmpPtr = ObjectPtr(hgo);
        }
    }
*/
/*
    return tmpPtr;
}
*/
//
// This will receive a number of Objects based on their IDs and create an
// ObjectList holding them. An additional query can be specified.
//

void ReceiveObjects(HgStub& db,
                    ObjIdArray& oids,
                    const char* query,
                    ObjectList* result,
                    boolean parents,
                    boolean score) 
{
    result->free();
    for (int pos = 0; !guiControls()->interrupted() && pos < oids.count(); 
         pos += MAXRECEIVE) 
    {
        int count1 = (oids.count()-pos) < MAXRECEIVE ? oids.count()-pos : MAXRECEIVE; 
        ObjIdArray part(oids, pos, count1);

        ObjectField ofield;
        if (!db.objectByIdQuery(part, query, ofield, parents)) 
        {
            // gpani: error occured
            ASSERT(0);
            return;
        }
        int count = part.count();

        guiControls()->setProgIndicator(pos+count, oids.count());

        for (int i = 0; i < count; i++) 
        {
            Object o = ofield[i];
            if (score)
                o += "\nScore=1000";

            HGObject* obj = MakeObject(o);
            if (obj)
            {
                ObjectPtr oPtr(obj);
                result->addTail(new ObjectNode(oPtr));
            }
        }
    }
    guiControls()->clearProgIndicator();
}

// 
// same as above except that a list of previously found fulltext document IDs
// is the base. Thus the resulting ObjectList contains scores.
// 
/*
void ReceiveScoredObjects(HgStub& db,
                          FtDocList& dl,
                          const char* query,
                          ObjectList* result,
                          boolean parents) 
{
    result->free();

#ifdef LOCAL_DB_20
    ObjIdArray oids = dl.idArray() ;
    ReceiveObjects(db, oids, query, result, parents);
#else

    FtDocTreap dt ;
    float highest = 0.0f;

    for (const DocListNode* n = dl.getFirst()  ;  n  ;  n = dl.getNext(n)) 
    {
        dt.insert (n->document()->id(), n->document()->features()) ;
        if (n->document()->features().weight() > highest)
            highest = n->document()->features().weight() ;
    }

    // get the objects in parts from the database
    ObjIdArray oids = dl.idArray() ;

    for (int pos = 0; !guiControls()->interrupted() && pos < oids.count(); pos += MAXRECEIVE) 
    {
        int count1 = (oids.count()-pos) < MAXRECEIVE ? oids.count()-pos : MAXRECEIVE; 
        ObjIdArray part(oids, pos, count1);
        ObjectField ofield;
        if (!db.objectByIdQuery(part, query, ofield, parents)) 
        {
            // gpani: error occured
            ASSERT(0);
            return;
        }
        int count = part.count();
        guiControls()->setProgIndicator(pos+count, oids.count());

        for (int i = 0; i < count; i++) 
        {
            Object o = ofield[i];

            // merge with the scores and positions from the treap
            FtDocTreapData f ; 
            if (dt.remove (o.ID(), f)) 
            {
                int score = int (f.weight() * 1000 / highest) ; 
                char scorestr[10];
                sprintf (scorestr, "%d", score) ;
                o += RString("\nScore=") + scorestr ;
                if (f.positions()) 
                {
                    RString s;
                    f.positionsStr(s);
                    o += RString("\nPositions=") + s;
                }
            }

            // add to the ObjectList
            HGObject* obj = MakeObject(o);
            if (obj)
                result->addTail(new ObjectNode(obj));
        }
    }
    guiControls()->clearProgIndicator();

#endif
}
*/
//
// This will find all children (i.e. documents accessible via a link) of
// a given document. If reverse is true, it will find the parents.
//

//void GetChildren(HgStub& db,            // connection to database
//                 ObjectList* children,  // return value (doc-list)
//                 ObjectID doc,          // the given document
//                 char* query,           // target objects conform to this query
//                 boolean reverse) 
//{
//  
//    // Step 1: get all anchors of doc 
//    ObjIdArray anchors;
//    int acount; // for the anchors
//    if (!db.anchors(doc, anchors)) 
//    {
//        // gpani: error occured
//    }
//    acount = anchors.count(); 
//
//    if (reverse)                // add document ID as default destination anchor
//    {
//        anchors.insert(doc);
//        acount++;
//    }
//       
//    if (acount)                 // if anchors found
//    {
//    
//        // Step 2: for each anchor get all link-anchor-relations:
//        //         where anchor is source/destination
//        ObjIdArray oids((acount+1)); // for the documents
//    
//
//        for (int j = 0; j < acount; j++) 
//        {
//            if (!reverse) 
//            {
//                ObjectID dst;
//                if (db.destOfSrc( anchors[j], dst) && dst.id()) 
//                {
//                    ObjectID doc;
//                    if (!db.getDocByAnchor(dst, doc)) 
//                    {
//                        // gpani: error occured
//                    }
//                    oids.insert(doc);
//                }
//            }
//            else 
//            {
//                ObjIdArray srcs;
//                if (db.srcsOfDst( anchors[j], srcs)) 
//                {
//                    for (int i = 0; i < srcs.count(); i++) 
//                    {
//                        ObjectID doc;
//                        if (!db.getDocByAnchor(srcs[i], doc)) 
//                        {
//                            // gpani: error occured
//                        }
//                        oids.insert(doc);
//                    }
//                }
//                else 
//                {
//                    // gpani: error occured
//                }
//            }
//        }
//
//        // Step 4: Get the whole destination document (only text documents)
//        ReceiveObjects(db, oids, query, children);
//    }
//} // GetChildren


//
// This will find all parents (i.e. documents accessible via a link) of
// a given document.
//
/*
void GetParents(HgStub& db,          // connection to database
                 ObjectList* children,// return value (doc-list)
                 ObjectID doc,        // the given document
                 char* query) {         // target objects conform to this query
  
    // Step 1: get all anchors of doc 
    ObjIdArray anchors;
    int acount; // for the anchors
    if (!db.anchors(doc, anchors)) 
    {
        // gpani: error occured
        ASSERT(0);
        return;
    }
    acount = anchors.count(); 

    // add document ID as default destination anchor
    anchors.insert(doc);
    acount++;
  
    // Step 2: for each anchor get all link-anchor-relations:
    //         where anchor is destination
    ObjIdArray oids((acount+1)); // for the documents
    
    for (int j = 0; j < acount; j++) 
    {
        ObjIdArray srcs;
        if (db.srcsOfDst( anchors[j], srcs)) 
        {
            for (int i = 0; i < srcs.count(); i++) 
            {
                ObjectID doc;
                if (!db.getDocByAnchor(srcs[i], doc)) 
                {
                    // gpani: error occured
                    ASSERT(0);
                    return;
                }
                oids.insert(doc);
            }
        }
        else 
        {
            // gpani: error occured
            ASSERT(0);
            return;
        }
    }

    // Step 4: Get the whole destination document (only text documents)
    ReceiveObjects(db, oids, query, children);
}
*/

//
// This will retrieve all (valid) anchors of a given document. In addition,
// the correponding destination anchor ID (which is in most cases the 
// destination document ID) for each anchor is stored in an OidUStringTree
// (for source anchors only, of course). If allanchors is true, open source
// anchors are also returned.
//
/*
void GetAnchors(HgStub& db,
                ObjectID doc,
                ObjectList* srcanchors,
                ObjectList* destanchors,
                boolean allanchors) 
{
    if (doc.id() < 0)
        return;         // skip remote documents !

    ObjIdArray anchIDs;
    int acount;
    if (!db.anchors(doc, anchIDs))  // source plus destination!
    {
        // gpani: error occured
        ASSERT(0);
        return;
    }
    acount = anchIDs.count();

    if (acount) 
    {
        ObjectList tmplist;
        ReceiveObjects(db, anchIDs, nil, &tmplist);
        for (ObjectNode* n = tmplist.getFirst(); n ; n = tmplist.getNext(n)) 
        {
            Anchor* a = (Anchor*)n->ObjPtr().Obj(); // we are sure this is an anchor
            if (a->IsSrc()) 
            {
                if (allanchors || a->Dest() != 0)     // valid source anchors or all?
                    srcanchors->addTail(new ObjectNode(ObjectPtr(a)));
            }
            else          // destination anchor
                destanchors->addTail(new ObjectNode(ObjectPtr(a)));
        }
    }
} // GetAnchors
*/

//
// This will sort the anchors of a TextDocument
// according to start position and end position in two OffsetLists. 
//
/*
void SortTextAnchors(ObjectList* anchors,  
                     OffsetList* start,
                     OffsetList* end) 
{

    for (ObjectNode* n = anchors->getFirst(); n ; n = anchors->getNext(n)) 
    {
        // decompose position attribute (assumes that start/end positions are
        // split be exactly one blank!!):
        Anchor* a = (Anchor*)n->ObjPtr().Obj(); // we are sure this is an anchor
        RString pos = a->Position();
        ObjectID startpos(pos);
        ObjectID endpos(pos.GetRight(pos.index(' ')+1));

        // Insert into output lists
        start->Insert(new OffsetNode(startpos, a->ID(), a->LinkType()));
        end->Insert(new OffsetNode(endpos, a->ID(), a->LinkType()));
    }
} // SortTextAnchors
*/

//
// This takes a source anchor ID and returns the corresponding destination document
//
/*
void GetDestByAnchor(HgStub& db,
                     ObjectList* dest,  // contains only one (historic reasons)
                     ObjectID aid,      // the given anchor id
                     char* query)       // target objects conform to this query
{

    // gpani: is it necessary to fetch this anchor ???
    Object obj;
    if (db.GetObject( aid, obj)) 
    {
        ObjIdArray oids(1);            // for the destination document(s)
        ObjectID dst = obj.field("Dest=");
        if (dst.id()) 
            oids.insert(dst);
       
        ReceiveObjects(db, oids, query, dest);
    }

//    ObjectID dst;
//    if (db.destOfSrc(aid, dst)) 
//    {
//        ObjIdArray oids(1);            // for the destination document(s)
//
//        if (dst.id()) 
//            oids.insert(dst);
//       
//        ReceiveObjects(db, oids, query, dest);
//    }
} // GetDestByAnchor
*/


//
// Read HTML text and convert anchors to HTF
//

// some helper routines:

void getHeaderParam(RString header, RString token, RString defParam, RString& param)
{
    RString capsBuf(header);        
    capsBuf = capsBuf.toupper();    // make it case-insensitive !
    token = token.toupper();

    int tokenPos = capsBuf.indexa(token);
    if (tokenPos > -1)
    {
//        tokenPos += token.length();
        header.gWord(tokenPos, " \t\n\r", param);
        if (!param.length())
            param = defParam;
    }
    else
        param = defParam;
}

#define TOBUFFER(s) buffer[bufptr] = '\0'; strcat(buffer,s); bufptr+=strlen(s);
#define TAGTOBUFFER(t) buffer[bufptr++] = '<'; TOBUFFER(t) buffer[bufptr++] = '>';


RString getNextTag(RString& text, int& pos)
{
    int oldPos = pos;
    RString tag;

    pos = text.index(oldPos, '<') + 1;      // look for start of tag
    if (pos)
    {
        oldPos = pos;
        pos = text.index(oldPos, '>') + 1;  // look for end of tag
        if (!pos)
            return tag; // Huh ??, nonsens !, not a valid tag - skip

        tag = text.gSubstrIndex(oldPos, pos - 2);
    }

    return tag;
} // getNextTag


RString getNextAnchor(RString& tag, RString anchorTag)
{
    RString tagu = tag;       // upper case version of tag
    tagu.toupper();
    int pos = tagu.indexa(anchorTag);
    
    while (tag[pos] != '=') 
        pos++;
    pos++;

    while (tag[pos] == ' ' || tag[pos] == '\t' || 
        tag[pos] == '\n' || tag[pos] == '\r') pos++;

    if (tag[pos] == '"') 
        pos++;

    RString tempTarget = tag.GetSubstrDelim(pos,'"');
    pos = tempTarget.index(':');
    RString target = WWW_Protocol; 
    if (pos >= 0)
        target = tempTarget.GetSubstrDelim(0,':');
    RString host = WWW_Host;
    if (tempTarget[pos+1] == '/' && tempTarget[pos+2] == '/')   // absolute host 
    {
        host = tempTarget.GetSubstrDelim(pos+3,'/');
        pos += 2 + host.length();
    }
    RString port = WWW_Port;
    if (host.index(':')>=0) 
    {
        port = host.GetRight(host.index(':')+1);
        host = host.GetSubstrDelim(0,':');
    }
    target += "://" + host + ":" + port;
    RString path = tempTarget.GetRight(pos+1);
    if (path[0] == '/') // absolute path
        target += path;
    else 
    {
        for (int i=WWW_Path.length()-1; WWW_Path[i]!='/'; i--); // last '/'
        while (path.indexa("../") == 3)     // starts with "../"
        {
            for (i--; WWW_Path[i]!='/'; i--);  // eat one more
            path = path.GetRight(3);
        }
        target += WWW_Path.GetSubstrIndex(0,i) + path;
    }

    return target;
} // getNextAnchor


boolean getHTMLLinks(Document* current,
                     RString& title,
                     RString& html, 
                     LinkList* linkList)
{
    int lenHtml = html.length();
    int posHtml = 0;
    int oldPosHtml = 0;
    RString tag;
    RString tagu;       // upper case version of tag
    RString target;     // destination anchor of current link
    int anchorno = -10;
    int inlineno = 0;
    boolean anchor_open = false;

    TextLinkData tmpLinkData;

    WWW_Protocol = current->Protocol();
    WWW_Host = current->Host();
    WWW_Port = current->Port();
    WWW_Path = current->Path();

    if (!linkList)
    {
        ASSERT(linkList);
        return false;
    }

    while (posHtml < lenHtml)
    {
        tag = getNextTag(html, posHtml);
        if (!posHtml)
            return true;

        tagu = tag;
        tagu.toupper();

        oldPosHtml = posHtml;

        if (posHtml < lenHtml && tag.length() && tagu == "TITLE")
        {
            tag = getNextTag(html, posHtml);
            if (!posHtml)
                return true;

            tagu = tag;
            tagu.toupper();

            title += html.gSubstrIndex(oldPosHtml, posHtml - tag.length() - 3);

            current->Title("??:"+title);        // overwrite Title
            oldPosHtml = posHtml;
        }

        if (posHtml < lenHtml && tag.length() && 
            tagu[0] == 'A' && (tag[1] == ' ' || tag[1] == '\n')) // make an anchor
        {
            oldPosHtml = posHtml;
            target = getNextAnchor(tag, "HREF");
            tmpLinkData.startOffset_ = oldPosHtml;
            anchor_open = true;
            --anchorno;             // new virtual anchor id ...
            oldPosHtml = posHtml;
        }

        if (posHtml < lenHtml && tag.length() && tagu == "/A")
        {

            oldPosHtml = posHtml;
            if (anchor_open) 
            {
                anchorno += inlineno;
                tmpLinkData.endOffset_ = oldPosHtml;
                Anchor a("ObjectID=" + ObjectID(anchorno).IDString() +
                         "\nType=Anchor\nTAnchor=Src\n" +
                         "\nDest=" + target +
                         "\nPosition=" + ObjectID(tmpLinkData.startOffset_).IDString() + " " + 
                                         ObjectID(tmpLinkData.endOffset_).IDString() + "\n");
                linkList->insert(new TextLinkNode(&a));
                anchor_open = false;      
                anchorno -= inlineno;
                inlineno = 0;
            }
            oldPosHtml = posHtml;
        }

        if (posHtml < lenHtml && tag.length() && !strncmp(tagu,"IMG",3)) // make an inline image
        {
            TextLinkData tmpInlLinkData;
            oldPosHtml = posHtml;
            RString inlTarget = getNextAnchor(tag, "SRC");

            tmpInlLinkData.startOffset_ = posHtml - tag.length() - 4;
            tmpInlLinkData.endOffset_ = posHtml;

            --anchorno;
           
            Anchor a("ObjectID=" + ObjectID(anchorno).IDString() +
                     "\nType=Anchor\nTAnchor=Src\n" +
                     "\nDest=" + inlTarget +
                     "\nLinkType=" + RString("inline") +
                     "\nPosition=" + ObjectID(tmpInlLinkData.startOffset_).IDString() + " " + 
                                     ObjectID(tmpInlLinkData.endOffset_).IDString() + "\n");
            linkList->insert(new TextLinkNode(&a));
            if (anchor_open)
                inlineno++;
            oldPosHtml = posHtml;
        }
    }

    return true;
}


boolean ReadHTMLtoHTF(WWWDocument* current,
                      RString& title,
                      RString& text,
                      LinkList* linkList)
{
#ifdef LOCAL_DB_20
#else

    const int bufsize = 1024;
    char buffer[bufsize];
    RString Target;
    int bufptr = 0;
    int offset = 0;
    int anchorno = 0;
    char receivestr[80];
    char connectstr[256];
    boolean anchor_open = false;

    RString method = "GET ";
    RString protVersion = " HTTP/1.0 \r\n";
    RString hTRQHeader = "Accept: */* \r\n";
    RString userAgent = "User-Agent: Amadeus/" + verString + " Beta \r\n";

    WWW_Protocol = current->Protocol();
    WWW_Host = current->Host();
    WWW_Port = current->Port();
    WWW_Path = current->Path();

    sprintf(connectstr, Connectstr, (const char*)WWW_Host, atoi(WWW_Port));
    guiControls()->setProgIndicator(connectstr);

    CSocket connection;
    if (connection.Connect(WWW_Host, atoi(WWW_Port)) == 0) 
    {
        RString header = method + WWW_Path + protVersion;
        header += hTRQHeader + "\r\n" + userAgent + "\r\n\r\n";
        connection.Write(header.string(), header.length()+1);

        sprintf(receivestr, Bytesstr,0);     
    
        guiControls()->setProgIndicator(receivestr);
        BeginBusyCursor(); 

        // evaluate header (if any)
        header = "";
        boolean headerEnd = false;  // end of header found ?
        int bufLen;
        do
        {
            buffer[0] = 0;
            bufLen = connection.Read(buffer, sizeof(buffer)-1);
            if (bufLen < -1)
                headerEnd = true;
            else
            {
                buffer[bufLen+1] = 0;
                int headerLen = header.length();
                header += buffer;

                // check for empty line = header end !
                int hEndPos;
                if ((hEndPos = (header.indexa("\n\n")-2)) < 0)
                    hEndPos = header.indexa("\r\n\r\n")-4;
                if (hEndPos > -1)
                {
                    header = header.gSubstrIndex(0, hEndPos);
                    bufLen = bufLen - hEndPos - 2;
                    memmove(buffer, buffer + hEndPos + 2, bufLen);
                    buffer[bufLen] = 0;
                    headerEnd = true;
                }
            }
        } while (bufLen > 0 && !headerEnd);

        RString contentType;
        RString location;
        RString lengthStr;
        RString encoding;
        RString language;

        getHeaderParam(header, "Content-Type:", "text/html", contentType);
        getHeaderParam(header, "Content-Length:", "", lengthStr);
        getHeaderParam(header, "Content-Endcoding:", "", encoding);
        getHeaderParam(header, "Content-Language:", "", language);
        getHeaderParam(header, "Location:", "", location);

        if (contentType == "text/html")
        {
            // read in HTML document
            text = buffer;  // part of data we alread read when reading the header
            int len;
            do
            {
                buffer[0] = 0;
                len = connection.Read(buffer, sizeof(buffer)-1);
                if (len > -1)
                    buffer[len] = 0;
                text += buffer;

                sprintf(receivestr, Bytesstr, text.length());     
                guiControls()->setProgIndicator(receivestr);

            } while (len > 0 && !guiControls()->interrupted());

            getHTMLLinks(current, title, text, linkList);
        }
        else
        {
            // binary, non-text document !
            guiControls()->clearProgIndicator();
            EndBusyCursor(); 
            ErrorMsg("Document type not supported yet !");
            return false;

            //
            int len;
            do
            {
                len = connection.Read(buffer, sizeof(buffer));

                // do something with 'buffer'

                sprintf(receivestr, Bytesstr, text.length());     
                guiControls()->setProgIndicator(receivestr);
             } while (len > 0 && !guiControls()->interrupted());
        }

        guiControls()->clearProgIndicator();
        EndBusyCursor(); 
        return true;
    }

    guiControls()->clearProgIndicator();
#endif

    return false; // something went wrong
} // ReadHTMLtoHTF


//
// Load a Gopher Document into an RString
//

boolean LoadGopher(const GopherDocument* current, RString& out, 
                   const char* query) 
{
#ifdef LOCAL_DB_20
#else

    char receivestr[80];
    char connectstr[80];
    sprintf(connectstr,Connectstr,(const char*)current->Host(),
            atoi(current->Port()));
    guiControls()->setProgIndicator(connectstr);
    rpcbuf connection;
    if (connection.connect(current->Host(), atoi(current->Port()))) 
    { 
        iostream s(&connection);
        if (query)
            if (current->Path() != "") 
                s << current->Path() + "\t" + query << "\r" << endl;
            else 
                s << query << "\r" << endl;
        else
            s << current->Path() << "\r" << endl;

        boolean MakeSGML = (current->GopherType() == "0"); // text
        if (MakeSGML)
            out = "<TITLE>" + ((HGObject*)current)->Title() + "<PLAIN>\n";

        int headerlen = out.length();
        char linebuf[512];
        BOOL aborted;
        while (!s.eof() && !(aborted = guiControls()->interrupted())) 
        {
            sprintf(receivestr,Bytesstr,out.length()-headerlen);     
            guiControls()->setProgIndicator(receivestr);
            if (MakeSGML) 
            {
                s.getline(linebuf, sizeof(linebuf));
                if (linebuf[strlen(linebuf)-1] == '\r')
                    linebuf[strlen(linebuf)-1] = '\0';
                if (*linebuf != '.' || *(linebuf+1) != '\0') 
                {
                    out += linebuf;
                    out += "\n"; 
                }
            }
            else    // !MakeSGML
            {
                s.read(linebuf, sizeof(linebuf));
                out += RString(linebuf, s.gcount());
            }
        }
        guiControls()->clearProgIndicator();
        if (MakeSGML)
            out += "</PLAIN>\n";

        if (aborted)
            return false;   // try to get whole document next time !

        return true;
    }
#endif

    return false; // something went wrong
} // LoadGopher


//
// This will check if the named collection exists. Returns ObjectID of collection,
// zero if not exists. One can also pass an RString containing the ID of the
// collection.
//
/*
ObjectID CheckCollection(HgStub& hg, const RString& name) 
{
    ObjectID collid;
    if (name.GetSubstrIndex(0,1) == "0x") // passed ID
        collid = (const char*)name;
    else                                  // passed name
        collid = hg.GetFirstCollID(name);

    if (collid != 0 && hg.IsCollection(collid))
        return collid;
    else
        return 0;
} // CheckCollection
*/

//
// Output the last error code from HgStub
//

RString HGError(errorcode code) 
{
    if (code <= UNDEF) 
        switch (code) 
        {
            case 0 /* OK */:        return "No error";
            case NOACCESS:          return "Access denied";
            case NODOCS:            return "No documents?";
            case NONAME:            return "No collection name";
            case NODOC:             return "Object is not a document";
            case NOOBJ:             return "No object received";
            case NOCOLLS:           return "No collections received";
            case DBSTUBNG:          return "Connection to database failed";
            case NOTFOUND:          return "Object not found or access denied";
            case EXIST:             return "Collection already exists";
            case FATHERDEL:         return "Father collection disappeared";
            case FATHNOCOLL:        return "Father collection not a collection";
            case NOTEMPTY:          return "Collection not empty";
            case DESTNOCOLL:        return "Destination not a collection";
            case SRCEQDEST:         return "Source equals destination";
            case REQPEND:           return "Request pending";
            case TIMEOUT:           return "Timeout";
            case NAMENOTUNIQUE:     return "Name not unique";
            case WRITESTOPPED:      return "Database now read-only; try again later";

#ifdef LOCAL_DB_20
            case xxxLOCKEDxxx:      return "Object locked; try again later";
#else
            case LOCKED:            return "Object locked; try again later";
#endif
            case CHGE_BASE_FLD:     return "Change of base-attribute";
            case NOT_RMD:           return "Attribute not removed";
            case FLD_EXIST:         return "Attribute exists";
            case CMD_SYNTAX:        return "Syntax error in command";
            case NOLANGUAGE:        return "No or unknown language specified";
            case WRGTYPE:           return "Wrong type in object (or no path)";
            case WRGVERSION:        return "Client version too old";
            case CONNECTION:        return "No connection to other server";
            case SYNC:              return "Synchronisation error";
            case NOPATH:            return "No path entry";
            case WRGPATH:           return "Wrong path entry";
            case PASSWD:            return "Wrong password";
            case LC_NO_MORE_USERS:  return "License expired (too many users); try again later";
            case LC_NO_MORE_DOCS:   return "License violation: Too many documents retrieved";
            case WRG_VALUE:         return "Wrong value";
            case INSUFF_FUNDS:	    return "Insufficient funds";
            case REORG: 	        return "Reorganization in progress, try again later";
            case UNDEF:             return "Undefined error";
            default:                return "No description for this error availiable";
        }
    else 
    {
        char ibuf[20];
        itoa(code, ibuf, 10);

        if (code <= FTSERV+UNDEF)
            return "Fulltext server error ! Code: " + RString(ibuf);
        else 
        if (code <= DOCSERV+UNDEF)
        {
            DcError dcError(code);
            return RString("Document Cache error: ") + dcError.description(0);
//            return "Document Cache error ! Code: " + RString(ibuf);
        }
        else
            return "Unknown error ! Code: " + RString(ibuf);
    }
}


void HGError(HgStub& hg) 
{
    if (hg.error() == USR_BREAK)
    {
        InfoMsg(IDS_MSGOPINTERRUPT);
    }
    else
    {
        RString errStr = "*** Error: " + HGError(hg.error()) + " ***";
        ErrorMsg(errStr);
    }
}


void hgError() 
{
    // look for Hyper-G protocol in protocol list, or create a new protocol
    for (ProtocolNode* protNode = HGObject::conOrProtList_.getFirst();
         protNode && protNode->protocol() && protNode->protocol()->getProtType() != "hyperg";
         protNode = HGObject::conOrProtList_.getNext(protNode));

    if (!protNode || 
        !protNode->protocol() || 
        protNode->protocol()->getProtType() != "hyperg")
    {
        ErrorMsg(IDS_ONLYHG);
        return;
    }

    if (((HyperGProt*)protNode->protocol())->serverPtr()->error() == USR_BREAK)
    {
        InfoMsg(IDS_MSGOPINTERRUPT);
    }
    else
    {
        RString errStr = "*** Error: " + HGError(((HyperGProt*)protNode->protocol())->serverPtr()->error()) + " ***";
        ErrorMsg(errStr);
    }
}


//
// Delivers current time in GMT:
//

RString TimeNow() 
{
    char timestr[18];
    time_t time = ::time(nil);
    struct tm* tm = gmtime(&time);
    sprintf(timestr, "%02d/%02d/%02d %02d:%02d:%02d", 
        tm->tm_year, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);
    return timestr;
}


//
// Find out my IP address
//

RString MyIPaddress() 
{
#ifdef LOCAL_DB_20
    return RString("0.0.0.0");
#else

    char hostname[64];
    gethostname(hostname, 63);

    struct hostent* host;
    host = gethostbyname(hostname);
    in_addr *ia = (in_addr*)host->h_addr;

    char* inaddr;
    inaddr = inet_ntoa(*ia);
    return RString(inaddr);
#endif
}


//
// Runs process in a shell. Tries to keep terminal clean.
//

int RunProcess(const char* command, boolean quiet) 
{
    extern int LineMode;

    int status = -1;
    if (command) 
    {
        status = 0;
        WaitAppExec(command);
    }
    return status;
}


//
// Open a pipe to a subprocess.
//

FILE* PipeOpen(const char* command) 
{
    return 0;             // not supported on PCs
}


//
// Close the pipe;
//

int PipeClose(FILE* f) 
{
    return -1;
}


//
// Formats special characters in RString s to UDI format (as used by WWW).
// Could be more efficient.
//

RString toUDI(const RString& s) 
{
    RString out;
    for (int i = 0; i<s.length(); i++) 
    {
        switch(s[i]) 
        {
            case ' ' : out += "%20"; break;          

            default : out += s.gSubstrIndex(i, i);
        }
    }
    return out;
}


//
// split 'command' into individual arguments (with parameter replacement)
//
RString parseCmdLine(const RString& cmd, RString inName, RString outName) 
{
    RString newCmdString;

    // now we have to split the line into individual arguments
    int pos = 0;
    RString arg;
    boolean noSpace = true;
    while (cmd.GetWord(pos, " \t$", arg)) 
    {
        if ( ((pos - arg.length() - 1) >= 0) && (cmd[pos - arg.length() - 1] == '$') )
        {
            arg = "$" + arg;
            if ( ((pos - arg.length() - 1) >= 0) && (cmd[pos - arg.length() - 1] != ' ') )
                noSpace = true;
            else
                noSpace = false;
        }
        else
            noSpace = false;

        if (arg == "$in")
            arg = inName;
        else
        if (arg == "$out")
            arg = outName;
    
        if (!noSpace && newCmdString.length())
            newCmdString += " ";

        newCmdString += arg;
    }

    return newCmdString;
}
