//<copyright>
// 
// Copyright (c) 1993,94
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        utils.h
//
// Purpose:     Utility routines used by some modules
//
// Created:     14 Jan 92    Frank Kappe
//
// Modified:    10 Oct 94    Thomas Dietinger
//
//
//
// Description:
//
//
//</file>

#ifndef HG_APPS_UTILS_H
#define HG_APPS_UTILS_H

#ifdef LOCAL_DB_20
#include <l_server/hgstub.h>
#else
#include <server/hgstub.h>
#endif

#include "offsetlist.h"
#include "objectlist.h"
#include "cursesviewer.h"


/////////////////////////////////////////////////////////////////////////////
// this function constructs an unique temporary file name using the info it gets
// from the object id structure (stored in a text file)

extern RString getTempName(Document* o);


/////////////////////////////////////////////////////////////////////////////
// 
extern RString getTempNameExt( const char *testFileBuf, int bufLen );

/////////////////////////////////////////////////////////////////////////////
// This function looks for an existing file, where the filename may be specified
// without extension. (If it's a non-generic type)
// If successful, it returns TRUE, else FALSE; the extension is added at the end of
// the input filename (if possible).
extern BOOL getCompleteFileName( RString& fileName, boolean& forceLoad);

//
// This will make a viewer Object out of the object's text description
//

extern HGObject* MakeObject(Object& o);

extern HGObject* makeObject(Object& o, ProtPtr pPtr);

//extern HGObject* makeObjectById(HgStub* db, RString dest);

//extern Object getObjectInfo(RString dest);

extern HGObject* makeObjectById(RString dest, ProtPtr pPtr = ProtPtr(0));

//extern HGObject* makeInlineObjectById(HgStub* db, RString dest, RString vId = "");

extern HGObject* makeInlineObjectById(RString dest, RString vId = "");

//
// This will get an object by its anchor and makes an inline object out of it
//
/*
extern ObjectPtr getInlineObjectByAnchor(HgStub& db, ObjectID linkId);
*/

//
// This will receive a number of Objects based on their IDs and create an
// ObjectList holding them. An additional query can be specified.
//
extern void ReceiveObjects(HgStub& db,
                           ObjIdArray& oids,
                           const char* query,
                           ObjectList* result,
                           boolean parents = false,
                           boolean score = false);

// 
// same as above except that a list of previously found fulltext document IDs
// is the base. Thus the resulting ObjectList contains scores.
// 
/*
extern void ReceiveScoredObjects (HgStub& db,
                                  FtDocList& dl,
                                  const char* query,
                                  ObjectList* result,
                                  boolean parents = true);
*/

//
// This will find all children (i.e. documents accessible via a link) of
// a given document. If reverse is true, it will find the parents.
//
/*
extern void GetChildren(HgStub& db,          // connection to database
                        ObjectList* children,// return value (doc-list)
                        ObjectID doc,        // the given document
                        char* query,         // target objects conform to this query
                        boolean reverse=false);
*/
//
// This will find all parents (i.e. documents accessible via a link) of
// a given document. 
//
/*
extern void GetParents(HgStub& db,              // connection to database
                       ObjectList* parents,     // return value (doc-list)
                       ObjectID doc,            // the given document
                       char* query);            // target objects conform to this query
*/
//
// This will retrieve all (valid) anchors of a given document. In addition,
// the correponding destination anchor ID (which is in most cases the 
// destination document ID) for each anchor is stored in an OidUStringTree
// (for source anchors only, of course).
//
/*
void GetAnchors(HgStub& db,
                ObjectID doc,
                ObjectList* srcanchors,
                ObjectList* destanchors,
                boolean allanchors = false);
*/
//
// This will sort the anchors of a TextDocument
// according to start position and end position in two OffsetLists. 
//
/*
void SortTextAnchors(ObjectList* anchors,  
                     OffsetList* start,
                     OffsetList* end);
*/
//
// This takes a source anchor ID and returns the corresponding destination document
//
/*
extern void GetDestByAnchor(HgStub& db,
                            ObjectList* dest,// could be more than one
                            ObjectID aid,    // the given anchor id
                            char* query);    // target objects conform to this query
*/
/////////////////////////////////////////////////////////////////////////////
// 
extern boolean getHTMLLinks(Document* current,
                            RString& title,
                            RString& html, 
                            LinkList* linkList);

/////////////////////////////////////////////////////////////////////////////
// 
/*
extern boolean getHTMLLinks( WWWDocument* current,
                             RString& title,
                             RString& html, 
                             OffsetList& start,
                             OffsetList& end,
                             OidUStringTree& dest);

*/
//
// Read HTML text and convert to HTF, filter out links, etc.
//

extern boolean ReadHTMLtoHTF(WWWDocument* current,
                             RString& title,
                             RString& text,
                             OffsetList& start,
                             OffsetList& end,
                             OidUStringTree& dest);

//
// Load a Gopher Document into an RString
//
boolean LoadGopher(const GopherDocument* current, 
                   RString& out,
                   const char* query = nil);

//
// This will check id the named collection exists. Returns ObjectID of collection,
// zero if not exists.
//
extern ObjectID CheckCollection(HgStub& hg, const RString& name);

//
// Output the last error code from HgStub
//
extern RString HGError(errorcode code);
extern void HGError(HgStub& hg);
extern void hgError();

//
// Delivers current time in GMT:
//
RString TimeNow();

//
// Runs process in a shell. Tries to keep terminal clean.
//
int RunProcess(const char* command, boolean quiet = false);

//
// Open a pipe to a subprocess.
//
/*
FILE* PipeOpen(const char* command);
*/
//
// Close the pipe;
//
/*
int PipeClose(FILE* f);
*/
//
// Formats special characters in RString s to UDI format (as used by WWW).
//
RString toUDI(const RString& s);

//
// split 'command' into individual arguments (with parameter replacement)
//
RString parseCmdLine(const RString& cmd, RString inName, RString outName) ;

#endif
