// This may look like C code, but it is really -*- C++ -*-

// **********************************************************
// File viewerlist.C
// ----------------------------------------------------------
// Created: T. Dietinger
// 
// The viewerlist is used by the session manager
// to keep track of its opened viewers.
//
// **********************************************************

#include <stdafx.h>
#include <hgapp.h>
#include <mainfrm.h>

#include "conf.h"
#include "viewerlist.h"
#include "hgobject.h"
#include "utils.h"

#include <fstream.h>
#include <direct.h>
#include <dos.h>
#include <fcntl.h>
#include <afxdlgs.h>
#include <pcutil/pcstuff.h>

#include <utils/environ.h>

#include <viewer/text/viewdoc.h>
#include <viewer/text/viewvw.h>


#if defined(_MSC_VER) & defined(_DEBUG)
#define new DEBUG_NEW
#endif


// **********************************************************
// class ViewerNode
// ----------------------------------------------------------
// **********************************************************

boolean ViewerNode::viewDocument(Document* o)
{
    viewerApp_ = o->viewerApp();
    tempName_ = o->tempName();

//    TRACE("\nViewerNode::viewDocument:DocType:%s, SubType: %s, TempName: %s, ViewerApp: %s\n",
//          (const char*)o->DocType(), (const char*)o->SubType(), (const char*)o->tempName(), (const char*)viewerApp_);

    // internal viewer ??
    if (viewerApp_.gSubstrIndex(0, 7) == "internal")
    {
        if ( !(o->DocType() == "Image" && o->SubType() == "inline") )
            // first check whether a viewer with this document is already visible
            for (CInfoNode* iNode = theApp.openViews_.getFirst(); iNode; iNode = theApp.openViews_.getNext(iNode))
            {
                if (iNode->info_.m_pDocument_ && 
                   ((CViewerDocument*)iNode->info_.m_pDocument_)->hgObj_.Obj() &&
                   ((CViewerDocument*)iNode->info_.m_pDocument_)->hgObj_.Obj()->ID() == o->ID() &&
                   iNode->type_ != collBrowser)
                {
                    iNode->info_.m_pFrameWnd_->ShowWindow(SW_MINIMIZE);
                    iNode->info_.m_pFrameWnd_->ShowWindow(SW_SHOWNORMAL);
                    return true;    // document already visible !
                }
            }

        if (o->DocType() == "Image" || (o->DocType() == "Remote" && o->remoteDocType() == "Image"))
        {
            int showStyle;
            if (o->SubType() == "inline")
            {
//                            showStyle = SW_SHOW;
                CInlineObjectList tmpInObjList;
                tmpInObjList.addTail(new CInlineObjectNode(o->getInlineID(), o->tempName()));
                theApp.getTextView()->addInlineObject(&tmpInObjList);
                return true;
            }
            else
            {
                showStyle = theApp.GetPrivateProfileInt("MainFrame", "MMViewerStyle", SW_SHOWMINIMIZED);
                viewerInfo_ = theApp.m_pImageTemplate_->OpenNewObject(o->tempName(), o->Title(), showStyle, o);
                if (viewerInfo_ && viewerInfo_->m_pDocument_)
                {
                    viewerInfo_->m_pDocument_->SetTitle(o->Title());
                    theApp.openViews_.addTail(viewerInfo_, imageViewer);
                    theApp.tileWindows();
                }
            }
        }
        else
        if (o->DocType() == "Movie" || (o->DocType() == "Remote" && o->remoteDocType() == "Movie"))
        {
            int showStyle = theApp.GetPrivateProfileInt("MainFrame", "MMViewerStyle", SW_SHOWMINIMIZED);
            viewerInfo_ = theApp.m_pMovieTemplate_->OpenNewObject(o->tempName(), o->Title(), showStyle, o);
            if (viewerInfo_ && viewerInfo_->m_pDocument_)
            {
                viewerInfo_->m_pDocument_->SetTitle(o->Title());
                theApp.openViews_.addTail(viewerInfo_, movieViewer);
                theApp.tileWindows();
            }
        }
        else
        if (o->DocType() == "PostScript" || (o->DocType() == "Remote" && o->remoteDocType() == "PostScript"))
        {
            int showStyle = theApp.GetPrivateProfileInt("MainFrame", "MMViewerStyle", SW_SHOWMINIMIZED);
            viewerInfo_ = theApp.m_pPScriptTemplate_->OpenNewObject(o->tempName(), o->Title(), showStyle, o);
            if (viewerInfo_ && viewerInfo_->m_pDocument_)
            {
                viewerInfo_->m_pDocument_->SetTitle(o->Title());
                theApp.openViews_.addTail(viewerInfo_, pScriptViewer);
                theApp.tileWindows();
            }
        }
        else
        {
            ErrorMsg("No internal viewer for this document type:", o->DocType());
            return false;
        }

        if (!viewerInfo_)
        {
        }

//        if (viewerInfo_)
//            ((CViewerDocument*)viewerInfo_->m_pDocument_)->hgObj_ = o;
        
    }
    else    // external viewer !
    {
        BOOL retVal;
        pId_ = new EXECAPPINFO;
        if (o->DocType() == "Generic")
            retVal = AppExec(viewerApp_, pId_, SW_SHOW);
        else
        {
            int showStyle = theApp.GetPrivateProfileInt("MainFrame", "MMViewerStyle", SW_SHOWMINIMIZED);
            retVal = AppExec(viewerApp_, pId_, showStyle /*SW_SHOWMINNOACTIVE*/);
        }
        if (retVal == FALSE)
        {
            char buffer[100];
            sprintf(buffer, "Sorry, but I couldn't launch '%s' for some reason ! (Error:%d)", (const char*)viewerApp_, retVal);
            ErrorMsg(buffer);
            delete pId_;
            pId_ = 0;
            return false;
        }
    } // external viewer !

    return true;
}




ViewerNode::ViewerNode(Document* o) 
{
    pId_ = 0;
    viewerInfo_ = 0;
    docPtr_ = o;
    viewDocument(o);
    tempName_ = o->tempName();
//    CDocument* pTmpDoc = theApp.m_pDummyTemplate_->OpenDocumentFile(NULL);
//    pTmpDoc->SetTitle(o->Title());
//    viewerInfo_ = theApp.m_pDummyTemplate_->OpenNewObject(NULL, o->Title(), SW_SHOWMINIMIZED);
//    if (viewerInfo_ && viewerInfo_->m_pDocument_)
//    {
//        viewerInfo_->m_pDocument_->SetTitle(o->Title());
//        theApp.openViews_.addTail(viewerInfo_, imageViewer);
//    }
//    theApp.tileWindows();
}

ViewerNode::~ViewerNode() 
{
    if (pId_) 
    {
        if (pId_->pi.hProcess)
            KillApp(pId_);
        delete pId_;
        pId_ = 0;
    }

    if (viewerInfo_)
    {
        // check, whether document is still valid (then it must also be in the viewer list !)
        for (CInfoNode* iNode = theApp.openViews_.getFirst(); iNode; iNode = theApp.openViews_.getNext(iNode))
        {
            if (iNode->info_.m_pDocument_ == viewerInfo_->m_pDocument_)
                viewerInfo_->m_pDocument_->OnCloseDocument();   // delete viewer, if it is still open
        }
    }

    // also delete related files if they still exist !
    if (tempName_.length())
    {

        LPOFSTRUCT pTmpOf = new OFSTRUCT;
        HFILE hTest;
        if ((hTest = OpenFile(tempName_, pTmpOf, OF_READ)) != HFILE_ERROR)
        {
            _lclose(hTest);
            TRACE("ViewerNode::~ViewerNode: delete file: %s\n", (const char*)tempName_);
            unlink(tempName_);
        }
        delete pTmpOf;
    }

}


