#include <ctype.h>
#include <sys/timeb.h>
#include <stdafx.h>
#include <pcutil\pcstuff.h>
#include "socket.h"
#include "mfcutil.h"


#ifdef LOCAL_DB_20
#include <l_server/s_mml_db.hpp>
#include <stdio.h>
#include <string.h>
extern HighLevelDataBase db;
long CSocket::lRefNo = 0L;
extern int nHGLexId;
#endif


#if defined(_MSC_VER) & defined(_DEBUG) & defined(_VERBOSE)
#define new DEBUG_NEW
#endif

const int MAXSOCKBUF = 2048; //16384;



struct Linger 
{
    Linger(int onoff, int timeout) { onoff_ = onoff; timeout_ = timeout; }
    int onoff_;
    int timeout_;
};


/////////////////////////////////////////////////////////////////////////////
// CSockIdNode:
//
CSockIdNode::CSockIdNode(SOCKET sId): DLListNode() 
{ 
    sId_ = sId; 
}


/////////////////////////////////////////////////////////////////////////////
// CSockIdList:
//

// remove socket from id list and close it 
void CSockIdList::remove(SOCKET sId)
{                                               
#ifdef LOCAL_DB_20
    // ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP remove");
	return;
#else
    if (sId == INVALID_SOCKET)
        return;

    for (CSockIdNode* idNode = getFirst(); idNode ; idNode = getNext(idNode))
    {
        if (idNode->getSockId() == sId)
        {
            CTmpSockIdList::remove(idNode);
            delete idNode;
            idNode = 0;
            
            if (closesocket(sId))
            {
                int err = WSAGetLastError();
                TRACE("Warning ! Couldn't close socket successfully: %d", err);
            }
            TRACE("CSockIdList::remove socket removed, id: %d\n", sId);

            break;
        }
    }
#endif
}


// remove ALL sockets from id list and close it !!
void CSockIdList::removeAll()
{                                               
#ifdef LOCAL_DB_20
    // ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP removeAll");
	return;
#else
    CSockIdNode* n;

    for(; n = (CSockIdNode*)removeHead();) 
    {
        int true = 1;
        Linger linger(1, 0);    // timeout to 0 ! -> close immediately !
        setsockopt(n->getSockId(), SOL_SOCKET, SO_LINGER, (const char*)&linger, sizeof(Linger));
            
        closesocket(n->getSockId());

        TRACE("CSockIdList::removeAll socket removed, id: %d\n", n->getSockId());

        delete n;
    }
#endif
}


void CSockIdList::addTail(SOCKET sId)
{
#ifdef LOCAL_DB_20
    // ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP addTail");
	return;
#else
    if (sId != INVALID_SOCKET)
        CTmpSockIdList::addTail(new CSockIdNode(sId));      // add new socket id to list

    TRACE("CSockIdList::addTail new socket added, id: %d\n", sId);
#endif
}


                               
/////////////////////////////////////////////////////////////////////////////
// CSocket:
//

long CSocket::winsockRef_ = 0;
CSockIdList CSocket::usedSockets_;


void CSocket::initLib()
{
#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP initLib");
	return;
#else
    // try to load 'winsock.dll'
    WSADATA myWSAData;
    WORD    VersionReqd = 0x0101;
                                            
    if (!winsockRef_ && !usedSockets_.count())               // first instance of CSocket !
    {
        BeginBusyCursor();
        int sockRet = WSAStartup(VersionReqd, &myWSAData);
        EndBusyCursor();
        if ( (sockRet == SOCKET_ERROR) || 
             (LOBYTE(myWSAData.wVersion) != 1) || (HIBYTE(myWSAData.wVersion) != 1) )
        {
            ErrorMsg("Couldn't find a usable 'WINSOCK.DLL V1.1'");
            if (sockRet != SOCKET_ERROR) WSACleanup();
            exit(-1);
        }
    }
    winsockRef_ ++;
    sId_ = INVALID_SOCKET;
    port_ = 0;
#endif
}



void CSocket::Open()
{
#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP Open");
	return;
#else
    if (sId_ != INVALID_SOCKET)
        Close();

    sId_ = socket(PF_INET, SOCK_STREAM, 0);
    if ( !sockIsValid() )
    {
        ErrorMsg("Couldn't create a valid socket !"); 
        WSACleanup();
        exit(-1);
    }
    usedSockets_.addTail(sId_);     // add new socket id to list
#endif
}


CSocket::CSocket()
{
#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP CSocket 1");
	return;
#else
    initLib();
    Open();
#endif
}


CSocket::CSocket(SOCKET sId)
{
#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP CSocket 2");
	return;
#else
    initLib();
    sId_ = sId;
    usedSockets_.addTail(sId_);     // add new socket id to list
#endif
}


CSocket::CSocket(const char* host, int port)
{
#ifdef LOCAL_DB_20
	sId_ = new S_result;

	if (sId_ == NULL) 
		return;

	lRefNo++;

	sscanf(host+2, "%lx %lx", &(sId_->key_data.entry_point), &lSize);
	sId_->key_data.lex_id = db.GetLexId(sId_->key_data.entry_point);
	sId_->key_data.flags = HG_DOCUMENT;

   	sId_->lDataPos = 0L;
	sId_->data_status = 0;
	sId_->lBOF = 0L;
	sId_->pOffsetsArray = NULL;
	sId_->nOffsetsCnt = 0;
	sId_->nOffsetsPos = 0;

	lRead = -1L;
#else
    initLib();
    Connect(host, port);
#endif
}

CSocket::~CSocket()   
{
#ifdef LOCAL_DB_20
	if (sId_ != NULL)
		delete sId_;
#else
    Close();
    winsockRef_ --; 
    if (!winsockRef_ && !usedSockets_.count()) 
        WSACleanup();
    if (winsockRef_ < 0) 
        winsockRef_ = 0;   // Hm ???, this should never happen !!
#endif
}


int CSocket::Close()
{   
    int ret = 1;
              
#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP Close");
	return ret;
#else
    port_ = 0;
    BeginBusyCursor();

    usedSockets_.remove(sId_);
    sId_ = INVALID_SOCKET;  // socket is now closed or was already closed !

    EndBusyCursor();

    return ret;
#endif
}


int CSocket::Connect(const struct sockaddr* name, int nameLen)
{
#ifdef LOCAL_DB_20
    int ret = 1;
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP Connect 1");
	return ret;
#else
    BeginBusyCursor();
    int ret = ::connect(sId_, name, nameLen);
    EndBusyCursor();

    return ret;
#endif
}


int CSocket::Connect(const char* host, int port)
{
#ifdef LOCAL_DB_20
    int ret = 1;
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP Connect 2");
	return ret;
#else
    int retVal;
    struct sockaddr_in far name;
    memset(&name, 0, sizeof(name));
    name.sin_family = PF_INET;
    name.sin_port = htons(port);

    const unsigned long INVALIDADDR = (unsigned long) -1;
    unsigned long hostinetaddr = INVALIDADDR;
    
    if (isascii(host[0]) && isdigit(host[0])) 
    {
        hostinetaddr = inet_addr((char*)host);
        name.sin_addr.s_addr = hostinetaddr;
    }
    if (hostinetaddr == INVALIDADDR) 
    {
        BeginBusyCursor();
        struct hostent far* hp = gethostbyname((char*)host);
        EndBusyCursor();
        if (hp == 0)
        {
            char buffer[100];
            sprintf(buffer, "CSocket::Connect: gethostbyname: no such host %s", host);
            ErrorMsg(buffer);
            return SOCKET_ERROR;
        }
        if (hp->h_addrtype != PF_INET)
        {
            ErrorMsg("CSocket::Connect: gethostbyname: not an Internet host");
            return SOCKET_ERROR;
        }
        memcpy(&name.sin_addr, hp->h_addr, sizeof(name.sin_addr));
    }

    int retries = 2;
    do
    {
        if (sId_ == INVALID_SOCKET)
            sId_ = socket(PF_INET, SOCK_STREAM, 0);     // try again
        if (sId_ == INVALID_SOCKET)
        {
            ErrorMsg("CSocket::Connect: couldn't create a socket !");
            return SOCKET_ERROR;
        }

        BeginBusyCursor();
        if ( (retVal = ::connect(sId_, (struct sockaddr*)&name, sizeof(name))) < 0)
        {
            if (h_errno == WSAECONNREFUSED && retries > 0)
            {
                // try again since peer's backlog may just be full
                Close();
                Sleep(1);
                continue;
            }
            else
            {
//                char buffer[100];
//                sprintf(buffer, "CSocket::Connect: couldn't connect to %s", host);
//                ErrorMsg(buffer);
                Close();
                EndBusyCursor();
                return SOCKET_ERROR;
            }
        }
        EndBusyCursor();
        break;          // the connection succeeded
    } while (retries-- > 0);

    return retVal;
#endif
} // Connect


int CSocket::Read(void *buf, int len)
{
#ifdef LOCAL_DB_20
	if (sId_ == NULL) 
		return -1;

    int ret = 0;

	if (lRead >= 0  &&  lRead >= lSize)
		return 0;

	if (lRead == -1) {
		sprintf((char*)buf, "%s\nver=%d\ntype=%d\nsz=%d\nref=%d\nZ=%d\n",
				"HGHDR", 1000, 1000, lSize, lRefNo, 0);
		ret = strlen((char*)buf) + 1;
		lRead = 0;
	}
	else {
		if (len > (lSize-lRead))
			len = lSize - lRead;
		ret = db.read(sId_, (char*)buf, len, absolute, lRead);

		if (ret < len)
			lRead = lSize;
		else
			lRead += ret;
	}

//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP Read");
	return ret;
#else
    if (sId_ == INVALID_SOCKET) 
        return 0;

    BeginBusyCursor();

    int ret = 0;;

    if ((int)GetVersion() > 0)      // Windows NT or Win32s ??
    {                               // => Windows NT !
        if (DoNothing() == -1)      // enable Windows 3.1 to do kind of multitasking ...
        {                           // WM_QUIT occured ! 
            linger();
            EndBusyCursor();
            WSASetLastError(WSAECONNABORTED);
            return -1L;             // ^ prevent hang at closesocket
        }

        ret = recv(sId_, (char*)buf, len, 0);
    }
    else
    {                               // => Win32s !
        int len16;

        if (DoNothing() == -1)      // enable Windows 3.1 to do kind of multitasking ...
        {                           // WM_QUIT occured ! 
            linger();
            EndBusyCursor();
            WSASetLastError(WSAECONNABORTED);
            return -1L;             // ^ prevent hang at closesocket
        }
        
        if (len > MAXSOCKBUF)
            len16 = MAXSOCKBUF;
        else
            len16 = (int)len;

        ret = recv(sId_, (char*)buf, len16, 0);
    } // Win32s !

    EndBusyCursor();
    return ret;
#endif
} // Read


int CSocket::Write(const void *buf, int len)
{
#ifdef LOCAL_DB_20
//  ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP Write");
	return -1;
#else
    if (sId_ == INVALID_SOCKET) 
        return 0;

    int ret = 0;
    BeginBusyCursor();

    if ((int)GetVersion() > 0)      // Windows NT or Win32s ??
    {                               // => Windows NT !
        if (DoNothing() == -1)      // enable Windows 3.1 to do kind of multitasking ...
        {                           // WM_QUIT occured ! 
            linger();
            EndBusyCursor();        // ^ prevent hang at closesocket
            WSASetLastError(WSAECONNABORTED);
            return -1L;             
        }
        
        ret = send(sId_, (char*)buf, len, 0);
    }
    else
    {
        int len16 = 0;
        int ret16 = -1;
        int sErr = 0;

        if (len < 0)
        {
            ErrorMsg("CSocket::Write: FATAL BUG: buffer with negative length !!");
            linger();
            Close();
            MyExit();
        }

        do
        {
            if (DoNothing() == -1)      // enable Windows 3.1 to do kind of multitasking ...
            {                           // WM_QUIT occured !
                linger();
                EndBusyCursor();
                WSASetLastError(WSAECONNABORTED);
                return -1L;             // ^ prevent hang at closesocket
            }
        
            if (len > MAXSOCKBUF)
                len16 = MAXSOCKBUF;
            else
                len16 = (int)len;

            ret16 = send(sId_, ((char*)buf)+ret, len16, 0);

            if (ret16>0)
            {
                ret += ret16;
                len -= ret16;
            }
            else
            {
                if ((sErr=WSAGetLastError()) != WSAEWOULDBLOCK)  // true error occured !
                {                                   // (WSAEWOULDBLOCK would only signal 
                    ret = ret16;                    //  that destination is busy and thus
                    len = 0;                        //  send did not work !)
#if defined(_DEBUG) & defined(_VERBOSE)
                    if (ret)
                        afxDump << "ERROR: Socket:write:" << sErr << "\n";
#endif
                }
                else
                {
                    Sleep(1000);
                    if (!len && !ret)
                        ret = ret16;
                }
            }

        } while (len>0);
    } // Win32s !

    EndBusyCursor();
    return ret;
#endif
} // write
  
                 
// This routine prevents a hang at closesocket if we couldn't finish a Read or Write
void CSocket::dontLinger()
{
#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP dontLinger");
#else
    int true = 1;
    
    if (sId_ != INVALID_SOCKET)
        setsockopt(sId_, SOL_SOCKET, SO_DONTLINGER, (const char*)&true, sizeof(int));
#endif
}

// This routine prevents a hang at closesocket if we couldn't finish a Read or Write
void CSocket::linger(int timeout)
{
#ifdef LOCAL_DB_20
//    ErrorMsg("Steve, das mut du noch implementierten:  SOCKET.CPP Linger");
#else
    if (sId_ != INVALID_SOCKET)
    {
        int true = 1;
        Linger linger(1, timeout);    // timeout to 0 ! -> close immediately !
        setsockopt(sId_, SOL_SOCKET, SO_LINGER, (const char*)&linger, sizeof(Linger));
    }
#endif
}
