//<copyright>
// 
// Copyright (c) 1995
// Institute for Information Processing and Computer Supported New Media (IICM),
// Graz University of Technology, Austria.
// 
//</copyright>

//<file>
//
// Name:        wrldraw.C
//
// Purpose:     drawing routines of VRML nodes
//
// Created:     24 Apr 95   Michael Pichler
//
// Changed:      5 Sep 95   Michael Pichler
//
//
//</file>


#if defined(PMAX) || defined (HPUX9)
enum Part { goofyPart };        // cfront confused about QvCone::Part and QvCylinder::Part
enum Binding { goofyBinding };  // cfront confused about QvMaterialBinding/QvNormalBinding::Binding
#endif


#include <vrml/QvElement.h>
#include <vrml/QvNodes.h>
#include <vrml/QvExtensions.h>
#include <vrml/QvUnknownNode.h>
#include <vrml/QvState.h>

#include "scene3d.h"
#include "camera.h"
#include "vrmlscene.h"
#include "vecutil.h"

#include <ge3d/ge3d.h>

#include <hyperg/hyperg/verbose.h>

#include <math.h>
#include <iostream.h>



/***** groups *****/

void QvGroup::draw (QvState* state)
{
  int n = getNumChildren ();
  for (int i = 0; i < n; i++)
    getChild (i)->draw (state);
}


void QvLOD::draw (QvState* state)
{
  Camera* cam = scene_->getCamera ();
  int numchildren = getNumChildren ();

  if (!cam || !numchildren)  // nothing to draw
    return;

  point3D campos, wcenter;
  cam->getposition (campos);

  // transform center_ into world coordinates
  ge3dTransformMcWc (center_, &wcenter);
  // cerr << "camera pos.: " << campos << ", world center: " << wcenter;
  dec3D (wcenter, *center_);
  float distance = norm3D (wcenter);
  // cerr << ", distance: " << distance << endl;

  int numranges = range.num;
  const float* ranges = range.values;

  for (int i = 0; i < numranges; i++)
  { if (distance < *ranges++)  // use child i
      break;
  } // to use all ranges there should be numranges+1 children

  if (i >= numchildren)  // default: last child
    i = numchildren-1;

  getChild (i)->draw (state);
}


void QvSeparator::draw (QvState* state)
{
  int numlights = vrmlscene_->numLights ();
  //cerr << "QvSeparator::draw: push ()" << endl;
  state->push ();  // push attribute stacks (TODO: does any node use them in draw???)
  ge3d_push_matrix ();

  int n = getNumChildren ();
  for (int i = 0; i < n; i++)
    getChild (i)->draw (state);

  //cerr << "QvSeparator::draw: pop ()" << endl;
  ge3d_pop_matrix ();
  state->pop ();  // pop stacks
  vrmlscene_->deactivateLights (numlights);
}


void QvSwitch::draw (QvState* state)
{
  int which = whichChild.value;

  if (which == QV_SWITCH_NONE)  // draw nothing
    return;
  if (which == QV_SWITCH_ALL)  // draw all children
    QvGroup::draw (state);
  else if (which < getNumChildren ())  // draw one child
    getChild (which)->draw (state);
}


void QvTransformSeparator::draw (QvState* state)
{
  ge3d_push_matrix ();

  int n = getNumChildren ();
  for (int i = 0; i < n; i++)
    getChild (i)->draw (state);

  ge3d_pop_matrix ();
}


/***** coordinates *****/


void QvCoordinate3::draw (QvState*)
{ // already handled in build
}

void QvNormal::draw (QvState*)
{ // already handled in build
}

void QvTextureCoordinate2::draw (QvState*)
{ // TODO: handle in build
}



/***** properties *****/

// #define DO_PROPERTY(className, stackIndex)                                    \
// void                                                                          \
// className::draw(QvState *state)                                           \
// {                                                                             \
//     ANNOUNCE(className);                                                      \
//     QvElement *elt = new QvElement;                                           \
//     elt->data = this;                                                         \
//     state->addElement(QvState::stackIndex, elt);                              \
// }

// #define DO_TYPED_PROPERTY(className, stackIndex, eltType)                     \
// void                                                                          \
// className::draw(QvState *state)                                           \
// {                                                                             \
//     ANNOUNCE(className);                                                      \
//     QvElement *elt = new QvElement;                                           \
//     elt->data = this;                                                         \
//     elt->type = QvElement::eltType;                                           \
//     state->addElement(QvState::stackIndex, elt);                              \
// }

#define DRAW(className)  \
void className::draw (QvState*)  { }

DRAW(QvFontStyle)

void QvMaterial::draw (QvState*)
{
//cerr << "QvMaterial::draw" << endl;
  // TODO: care for material binding (front/back)
  // TODO: do not need full material definition for wireframe/hidden line
/*
  if (curdrawmode_ == ge3d_wireframe)
    ge3dLineColor (linecolor_);
  else
*/
  // if wireframe ge3d_setlinecolor (rgb) else
  ge3dMaterial (mat_front_and_back, matdef_);
}

DRAW(QvMaterialBinding)
DRAW(QvNormalBinding)


void QvShapeHints::draw (QvState*)
{
  if (scene_->twosidedpolys () == Scene3D::twosided_auto)
    ge3dHint (hint_backfaceculling, backfaceculling_);
}


void QvTexture2::draw (QvState*)
{
  if (curdrawmode_ == ge3d_texturing)
  {
    if (handle_ > 0)
    { ge3dDoTexturing (1);
      ge3dApplyTexture (handle_);
    }
    else  // turn off texturing
      ge3dDoTexturing (0);
  }
}


DRAW(QvTexture2Transform)



/***** lights *****/


void QvDirectionalLight::draw (QvState*)
{
  if (curdrawmode_ < ge3d_flat_shading)
    return;

  if (on.value)
  { int index = vrmlscene_->nextLight ();
    ge3dSetLightSource (index, &color_, &direction_, 0 /* directional */, 0);
    ge3d_switchlight (index, 1);
  }
}


void QvPointLight::draw (QvState*)
{
  if (curdrawmode_ < ge3d_flat_shading)
    return;

  if (on.value)
  { int index = vrmlscene_->nextLight ();
    ge3dSetLightSource (index, &color_, position_, 1 /* positional */, 0);
    ge3d_switchlight (index, 1);
  }
}


void QvSpotLight::draw (QvState*)
{
  if (curdrawmode_ < ge3d_flat_shading)
    return;

  if (on.value)
  { int index = vrmlscene_->nextLight ();
    ge3dSpotLight (index, &color_, position_, direction_, dropOffRate.value, cutangle_);
    ge3d_switchlight (index, 1);
  }
}



/***** cameras *****/


void QvOrthographicCamera::draw (QvState*)
{
  Camera* cam = vrmlscene_->getCamera ();

  // see comments in perspective camera below
  // TODO: provide mechanism for zooming with ortho cameras
  ge3dCamera (cam_orthographic | cam_relative,
    pos_, rotangle_, rotaxis_, height_,
    scene_->getWinAspect (),
    0.0, cam->getyon ()
  );
}


void QvPerspectiveCamera::draw (QvState*)
{
  Camera* cam = vrmlscene_->getCamera ();

  // focal distance not relevant here (may be used for default overview)
  ge3dCamera (cam_perspective | cam_relative,
    pos_, rotangle_, rotaxis_, yangle_,
    scene_->getWinAspect (),
    cam->gethither (), cam->getyon ()
  );
}


/***** transformations *****/

// TODO: may calculate overall transformations right after loading

void QvTransform::draw (QvState*)
{
  // cast: GO 090795 for WIN32
  ge3dMultMatrix ((const float (*)[4]) mat_);
}

void QvRotation::draw (QvState*)
{
  ge3dRotate (axis_, angle_);
}

void QvMatrixTransform::draw (QvState*)
{
  ge3dMultMatrix (mat_);
}

void QvTranslation::draw (QvState*)
{
  ge3dTranslate (trans_);
}

void QvScale::draw (QvState*)
{
  ge3dScale (scale_);
}



/***** shapes *****/


void QvAsciiText::draw (QvState*)  // TODO
{
}


void QvCube::draw (QvState*)
{
  ge3dCube (&min_, &max_);
}


void QvCone::draw (QvState*)
{
  float h = height.value;
  ge3dCylinder (bottomRadius.value, 0 /* topradius */, - h / 2.0, h, parts_);
}


void QvCylinder::draw (QvState*)
{
  float r = radius.value;
  float h = height.value;
  ge3dCylinder (r, r, - h / 2.0, h, parts_);
}


void QvSphere::draw (QvState*)
{
  ge3dSphere (radius.value);
}


void QvIndexedFaceSet::draw (QvState*)
{
//cerr << "QvIndexedFaceSet::draw" << endl;

  ge3dFaceSet (
    vertexlist_, numvertinds_, vertindices_,
    materials_, matbinding_, nummatinds_, matindices_,
    normallist_, numnormalinds_, normalindices_, facenormals_,
    texvertlist_, numtextureinds_, textureindices_
  );
  // currently materials, textures ignored (TODO)

} // QvIndexedFaceSet


void QvIndexedLineSet::draw (QvState*)
{
//cerr << "QvIndexedLineSet::draw" << endl;

  // assert: int and long of QvMFLong are of same size - see <vrml/QvMFLong>
  // assert: arrays of point3D's are compatible to arrays of float triples

  ge3dLineSet (vertexlist_, numverts_, vertindices_);
  // currently materials ignored (TODO) - (normals and textures not relevant)
  // will do this with a function ge3dLineSet (replacing face_open_wireframe) (TODO)

} // QvIndexedLineSet


void QvPointSet::draw (QvState*)
{
  ge3dPointSet (points_, num_);  // TODO: materials (normals, textures not relevant)

} // QvPointSet



/***** WWW *****/  // TODO


void QvWWWAnchor::draw (QvState* state)  // draw differently on anchor highlighting
{
  // spec: WWWAnchor must behave like a separator (although it is not derived from it)
  int numlights = vrmlscene_->numLights ();
  state->push ();  // push attribute stacks (TODO: does any node use them in draw???)
  ge3d_push_matrix ();

  // draw children of group
  int n = getNumChildren ();
  for (int i = 0; i < n; i++)
    getChild (i)->draw (state);

  ge3d_pop_matrix ();
  state->pop ();  // pop stacks
  vrmlscene_->deactivateLights (numlights);
}


void QvWWWInline::draw (QvState* state)
{
  // fetch first time needed
  if (state_ == s_virgin)
  {
    DEBUGNL ("QvWWWInline first encountered. Sending request");
    state_ = s_requested;
    // the request of the inline URL is VRML specific, but was put into Scene3D to allow
    // separate treatment for different viewer types (Harmony, standalone, etc.)
    scene_->requestInlineURL (this, name.value.getString (), parentURL_.getString ());
  }

  // draw bounding box until scene fetched
  if (hasextent_ && state_ != s_completed)
  {
    ge3dWireCube (&min_, &max_);
  }

  QvGroup::draw (state);  // draw children (if already fetched)
}


/***** misc *****/


void QvInfo::draw (QvState*)  { }

void QvUnknownNode::draw (QvState* state)
{ QvGroup::draw (state);
}



/***** extensions *****/


void QvLabel::draw (QvState*)  { }


void QvLightModel::draw (QvState*)
{
  if (scene_->dolighting () == Scene3D::lighting_auto)
    ge3dHint (hint_lighting, dolighting_);
}
